//
// Functions that describe FlOpCount operations and computations.
//
// Author : Benoit Hamelin
// 2003 Universit de Sherbrooke
//


function [left,right,sleft,sright]=FlOpCtr_PrepareElemOp(left, right)
	// Prepares operands of an element-wise operation.
	//
	sleft = size(left)
	sright = size(right)
	if (or(sleft <> sright)) then
		if (prod(sleft) == 1) then
			left.value = ones(sright(1), 1) * left.value * ones(1, sright(2))
			sleft = size(left)
		elseif (prod(sright) == 1) then
			right.value = ones(sleft(1), 1) * right.value * ones(1, sleft(2))
			sright = size(right)
		else
			error("Size discrepancy between two operands of an element-wise operation.")
		end  // if
	end  // if
endfunction


function [r]=%FLOPCTR_a_FLOPCTR(left, right)
	// Addition.
	//
	global FLOPCTR_OPCOUNTS FLOPCTR_ADDS FLOPCTR_MULTS FLOPCTR_INVS FLOPCTR_FETCHES FLOPCTR_STORES FLOPCTR_CPLXPRIMS;
	[left,right,sleft,sright]=FlOpCtr_PrepareElemOp(left, right)
	r = flopctr(FlOpCtr_GetValue(left) + FlOpCtr_GetValue(right))
	nbElems = prod(sleft)
	FLOPCTR_OPCOUNTS(FLOPCTR_ADDS) = FLOPCTR_OPCOUNTS(FLOPCTR_ADDS) + nbElems
	FLOPCTR_OPCOUNTS(FLOPCTR_FETCHES) = FLOPCTR_OPCOUNTS(FLOPCTR_FETCHES) + 2 * nbElems
	FLOPCTR_OPCOUNTS(FLOPCTR_STORES) = FLOPCTR_OPCOUNTS(FLOPCTR_STORES) + nbElems
endfunction


function [r]=%FLOPCTR_a_s(left, right)
	r = left + flopctr(right)
endfunction


function [r]=%s_a_FLOPCTR(left, right)
	r = flopctr(left) + right
endfunction


function [r]=%FLOPCTR_s(fc)
	// Unary minus.  FlOp count isn't modified, as opposing a floating point number implies the mere changing of a bit for each
	// component of the value matrix.
	//
	r = flopctr(-FlOpCtr_GetValue(fc))
endfunction


function [r]=%FLOPCTR_s_FLOPCTR(left, right)
	// Subtraction.
	//
	r = left + (-right)
endfunction


function [r]=%FLOPCTR_s_s(left, right)
	r = left + (-right)
endfunction


function [r]=%s_s_FLOPCTR(left, right)
	r = left + (-right)
endfunction


function [r]=%FLOPCTR_sum(varargin)
	[lhs,rhs] = argn()
	if (rhs < 1) then
		error("The sum operator takes at least one argument.")
	end  // if
	operand = varargin(1)
	op = 0
	if (rhs > 1) then
		arg2 = varargin(2)
		if (arg2 == 1 | arg2 == "r") then
			op = 1
		elseif (arg2 == 2 | arg2 == "c") then
			op = 2
		else
			error("Invalid second argument; must be 1, 2, ''r'' or ''c''.");
		end  // if
	end  // if

	if (op == 0) then
		// Must add together all elements of the vectorized expression.
		nbElem = size(operand, "*")
		r = 0
		for i = 1:nbElem do
			r = r + operand(i)
		end  // for
	elseif (op == 1) then
		// Must add together all rows of the operandession.
		[nbRows,nbCols] = size(operand)
		r = zeros(1, nbCols)
		for i = 1:nbRows do
			r = r + operand(i, :)
		end  // for
	elseif (op == 2) then
		// Must add together all columns of the operandession.
		[nbRows,nbCols] = size(operand)
		r = zeros(nbRows, 1)
		for i = 1:nbCols do
			r = r + operand(:, i)
		end  // for
	else
		error("Bad programming of primitive -- email benoit@benoithamelin.com with insult.")
	end  // if
endfunction


function [r]=%FLOPCTR_m_FLOPCTR(left, right)
	// Matricial product.
	//
	global FLOPCTR_OPCOUNTS FLOPCTR_ADDS FLOPCTR_MULTS FLOPCTR_INVS FLOPCTR_FETCHES FLOPCTR_STORES FLOPCTR_CPLXPRIMS;
	sleft = size(left)
	sright = size(right)
	if (sleft(2) <> sright(1)) then
		if ((prod(sleft) == 1) | (prod(sright) == 1)) then
			r = left .* right
		else
			error("Ill-sized multiplicands.")
		end  // if
	else
		r = flopctr(FlOpCtr_GetValue(left) * FlOpCtr_GetValue(right))
		nbScalarProducts = sleft(1) * sright(2)
		FLOPCTR_OPCOUNTS(FLOPCTR_ADDS) = FLOPCTR_OPCOUNTS(FLOPCTR_ADDS) + nbScalarProducts * (sleft(2) - 1)
		FLOPCTR_OPCOUNTS(FLOPCTR_FETCHES) = FLOPCTR_OPCOUNTS(FLOPCTR_FETCHES) + 2 * nbScalarProducts * sleft(2)
		FLOPCTR_OPCOUNTS(FLOPCTR_MULTS) = FLOPCTR_OPCOUNTS(FLOPCTR_MULTS) + nbScalarProducts * sleft(2)
		FLOPCTR_OPCOUNTS(FLOPCTR_STORES) = FLOPCTR_OPCOUNTS(FLOPCTR_STORES) + nbScalarProducts
	end  // if
endfunction


function [r]=%FLOPCTR_m_s(left, right)
	r = left * flopctr(right)
endfunction


function [r]=%s_m_FLOPCTR(left, right)
	r = flopctr(left) * right
endfunction


function [r]=%FLOPCTR_inv(fc)
	// Multiplicative inverse computation.
	//
	global FLOPCTR_OPCOUNTS FLOPCTR_ADDS FLOPCTR_MULTS FLOPCTR_INVS FLOPCTR_FETCHES FLOPCTR_STORES FLOPCTR_CPLXPRIMS;
	sfc = size(fc)
	if (sfc(1) <> sfc(2)) then
		error("Cannot invert a non-square matrix.")
	end  // if
	if (sfc(1) <> 1) then
		error("Only inversion of scalars is implemented.")
	else
		r = fc
		r.value = inv(r.value)
		FLOPCTR_OPCOUNTS(FLOPCTR_INVS) = FLOPCTR_OPCOUNTS(FLOPCTR_INVS) + 1
		FLOPCTR_OPCOUNTS(FLOPCTR_FETCHES) = FLOPCTR_OPCOUNTS(FLOPCTR_FETCHES) + 1
		FLOPCTR_OPCOUNTS(FLOPCTR_STORES) = FLOPCTR_OPCOUNTS(FLOPCTR_STORES) + 1
	end  // if
endfunction


function [r]=%FLOPCTR_r_FLOPCTR(left, right)
	// Right division.
	//
	global FLOPCTR_OPCOUNTS FLOPCTR_ADDS FLOPCTR_MULTS FLOPCTR_INVS FLOPCTR_FETCHES FLOPCTR_STORES FLOPCTR_CPLXPRIMS;
	sright = size(right)
	if and(sright == [1,1]) then
		r = flopctr(FlOpCtr_GetValue(left) / FlOpCtr_GetValue(right))
		FLOPCTR_OPCOUNTS(FLOPCTR_INVS) = FLOPCTR_OPCOUNTS(FLOPCTR_INVS) + 1
		FLOPCTR_OPCOUNTS(FLOPCTR_FETCHES) = FLOPCTR_OPCOUNTS(FLOPCTR_FETCHES) + 2
		FLOPCTR_OPCOUNTS(FLOPCTR_STORES) = FLOPCTR_OPCOUNTS(FLOPCTR_STORES) + 1
	elseif (sright(1) == sright(2)) then
		r = left * inv(right)
	else
		r = left ./ right
	end  // if
endfunction


function [r]=%FLOPCTR_r_s(left, right)
	sright = size(right)
	if (sright(1) == sright(2)) then
		r = left * inv(flopctr(right))
	else
		r = left ./ flopctr(right)
	end  // if
endfunction


function [r]=%s_r_FLOPCTR(left, right)
	sright = size(right)
	if (sright(1) == sright(2)) then
		r = flopctr(left) * inv(right)
	else
		r = flopctr(left) ./ right
	end  // if
endfunction


function [r]=%FLOPCTR_l_FLOPCTR(left, right)
	// Left division.
	//
	sleft = size(left)
	if (sleft(1) == sleft(2)) then
		r = inv(left) * right
	else
		r = left .\ right
	end  // if
endfunction


function [r]=%FLOPCTR_l_s(left, right)
	r = inv(left) * flopctr(right)
endfunction


function [r]=%s_l_FLOPCTR(left, right)
	r = inv(flopctr(left)) * right
endfunction


function [r]=%FLOPCTR_x_FLOPCTR(left, right)
	// Element-wise product.
	//
	global FLOPCTR_OPCOUNTS FLOPCTR_ADDS FLOPCTR_MULTS FLOPCTR_INVS FLOPCTR_FETCHES FLOPCTR_STORES FLOPCTR_CPLXPRIMS;
	[left,right,sleft,sright]=FlOpCtr_PrepareElemOp(left, right)
	//r = combineFlOpCtrs(left, right, FlOpCtr_GetValue(left) .* FlOpCtr_GetValue(right))
	r = flopctr(FlOpCtr_GetValue(left) .* FlOpCtr_GetValue(right))
	nbElems = prod(sleft)
	FLOPCTR_OPCOUNTS(FLOPCTR_MULTS) = FLOPCTR_OPCOUNTS(FLOPCTR_MULTS) + nbElems
	FLOPCTR_OPCOUNTS(FLOPCTR_FETCHES) = FLOPCTR_OPCOUNTS(FLOPCTR_FETCHES) + 2 * nbElems
	FLOPCTR_OPCOUNTS(FLOPCTR_STORES) = FLOPCTR_OPCOUNTS(FLOPCTR_STORES) + nbElems
endfunction


function [r]=%FLOPCTR_x_s(left, right)
	r = left .* flopctr(right)
endfunction


function [r]=%s_x_FLOPCTR(left, right)
	r = flopctr(left) .* right
endfunction


function [r]=%FLOPCTR_d_FLOPCTR(left, right)
	// Element-wise right division.
	//
	global FLOPCTR_OPCOUNTS FLOPCTR_ADDS FLOPCTR_MULTS FLOPCTR_INVS FLOPCTR_FETCHES FLOPCTR_STORES FLOPCTR_CPLXPRIMS;
	[left,right,sleft,sright]=FlOpCtr_PrepareElemOp(left, right)
	//r = combineFlOpCtrs(left, right, FlOpCtr_GetValue(left) ./ FlOpCtr_GetValue(right))
	r = flopctr(FlOpCtr_GetValue(left) ./ FlOpCtr_GetValue(right))
	nbElems = prod(sleft)
	FLOPCTR_OPCOUNTS(FLOPCTR_INVS) = FLOPCTR_OPCOUNTS(FLOPCTR_INVS) + nbElems
	FLOPCTR_OPCOUNTS(FLOPCTR_MULTS) = FLOPCTR_OPCOUNTS(FLOPCTR_MULTS) + nbElems
	FLOPCTR_OPCOUNTS(FLOPCTR_FETCHES) = FLOPCTR_OPCOUNTS(FLOPCTR_FETCHES) + 3 * nbElems
	FLOPCTR_OPCOUNTS(FLOPCTR_STORES) = FLOPCTR_OPCOUNTS(FLOPCTR_STORES) + 2 * nbElems
endfunction


function [r]=%FLOPCTR_d_s(left, right)
	r = left ./ flopctr(right)
endfunction


function [r]=%s_d_FLOPCTR(left, right)
	r = flopctr(left) ./ right
endfunction


function [r]=%FLOPCTR_q_FLOPCTR(left, right)
	// Element-wise left division.
	//
	global FLOPCTR_OPCOUNTS FLOPCTR_ADDS FLOPCTR_MULTS FLOPCTR_INVS FLOPCTR_FETCHES FLOPCTR_STORES FLOPCTR_CPLXPRIMS;
	[left,right,sleft,sright]=FlOpCtr_PrepareElemOp(left, right)
	//r = combineFlOpCtrs(left, right, FlOpCtr_GetValue(left) .\ FlOpCtr_GetValue(right))
	r = flopctr(FlOpCtr_GetValue(left) .\ FlOpCtr_GetValue(right))
	nbElems = prod(sleft)
	FLOPCTR_OPCOUNTS(FLOPCTR_INVS) = FLOPCTR_OPCOUNTS(FLOPCTR_INVS) + nbElems
	FLOPCTR_OPCOUNTS(FLOPCTR_MULTS) = FLOPCTR_OPCOUNTS(FLOPCTR_MULTS) + nbElems
	FLOPCTR_OPCOUNTS(FLOPCTR_FETCHES) = FLOPCTR_OPCOUNTS(FLOPCTR_FETCHES) + 3 * nbElems
	FLOPCTR_OPCOUNTS(FLOPCTR_STORES) = FLOPCTR_OPCOUNTS(FLOPCTR_STORES) + 2 * nbElems
endfunction


function [r]=%FLOPCTR_q_s(left, right)
	r = left .\ flopctr(right)
endfunction


function [r]=%s_q_FLOPCTR(left, right)
	r = flopctr(left) .\ right
endfunction


function [r]=%s_p_FLOPCTR(left, right)
	// Regular exponential.
	//
	global FLOPCTR_OPCOUNTS FLOPCTR_ADDS FLOPCTR_MULTS FLOPCTR_INVS FLOPCTR_FETCHES FLOPCTR_STORES FLOPCTR_CPLXPRIMS;
	sleft = size(left)
	sright = size(right)
	if (sleft(1) <> sleft(2) | or(sright <> ones(sright))) then
		r = left .^ right
	else
		if (sleft(1) <> 1) then
			error("Exponentials with matricial base not yet implemented.")
		end  // if
		r = flopctr(left ^ FlOpCtr_GetValue(right))
		FLOPCTR_OPCOUNTS(FLOPCTR_CPLXPRIMS) = FLOPCTR_OPCOUNTS(FLOPCTR_CPLXPRIMS) + 1
		if (abs(left - %e) > %eps) then
			// Calculate exponentiation through natural exponential and logarithm.
			FLOPCTR_OPCOUNTS(FLOPCTR_CPLXPRIMS) = FLOPCTR_OPCOUNTS(FLOPCTR_CPLXPRIMS) + 1
		end  // if
		FLOPCTR_OPCOUNTS(FLOPCTR_FETCHES) = FLOPCTR_OPCOUNTS(FLOPCTR_FETCHES) + 2
		FLOPCTR_OPCOUNTS(FLOPCTR_STORES) = FLOPCTR_OPCOUNTS(FLOPCTR_STORES) + 1
	end  // if
endfunction


function [r]=%s_j_FLOPCTR(left, right)
	// Element-wise exponential.
	//
	global FLOPCTR_OPCOUNTS FLOPCTR_ADDS FLOPCTR_MULTS FLOPCTR_INVS FLOPCTR_FETCHES FLOPCTR_STORES FLOPCTR_CPLXPRIMS;
	[left,right,sleft,sright] = FlOpCtr_PrepareElemOp(flopctr(left), right)
	r = flopctr(FlOpCtr_GetValue(left) .^ FlOpCtr_GetValue(right))
	nbElems = prod(sleft)
	FLOPCTR_OPCOUNTS(FLOPCTR_CPLXPRIMS) = FLOPCTR_OPCOUNTS(FLOPCTR_CPLXPRIMS) + nbElems
	nbNonNaturalBases = size(find(abs(FlOpCtr_GetValue(left) - %e) > %eps), "*")
	FLOPCTR_OPCOUNTS(FLOPCTR_CPLXPRIMS) = FLOPCTR_OPCOUNTS(FLOPCTR_CPLXPRIMS) + nbNonNaturalBases
	FLOPCTR_OPCOUNTS(FLOPCTR_FETCHES) = FLOPCTR_OPCOUNTS(FLOPCTR_FETCHES) + 2 * nbElems
	FLOPCTR_OPCOUNTS(FLOPCTR_STORES) = FLOPCTR_OPCOUNTS(FLOPCTR_STORES) + nbElems
endfunction


function [r]=%FLOPCTR_p_s(base, exponent)
	// Regular pseudo-monomial.
	//
	global FLOPCTR_OPCOUNTS FLOPCTR_ADDS FLOPCTR_MULTS FLOPCTR_INVS FLOPCTR_FETCHES FLOPCTR_STORES FLOPCTR_CPLXPRIMS;
	sbase = size(base)
	if (sbase(1) <> 1) then
		error("Pseudo-monomials with a matricial base are not yet implemented.")
	end  // if

	if (exponent > 1) then
		[r,listPowers2,listIndex,listProds,rootExpo,rootFactor] = elevatePower(base, exponent)
	else
		r = flopctr(FlOpCtr_GetValue(base) ^ exponent)
		FLOPCTR_OPCOUNTS(FLOPCTR_CPLXPRIMS) = FLOPCTR_OPCOUNTS(FLOPCTR_CPLXPRIMS) + 1
		FLOPCTR_OPCOUNTS(FLOPCTR_FETCHES) = FLOPCTR_OPCOUNTS(FLOPCTR_FETCHES) + 2
		FLOPCTR_OPCOUNTS(FLOPCTR_STORES) = FLOPCTR_OPCOUNTS(FLOPCTR_STORES) + 1
	end  // if
endfunction


function [r]=%FLOPCTR_j_s(base, exponent)
	// Element-wise pseudo-monomial.
	//
	global FLOPCTR_OPCOUNTS FLOPCTR_ADDS FLOPCTR_MULTS FLOPCTR_INVS FLOPCTR_FETCHES FLOPCTR_STORES FLOPCTR_CPLXPRIMS;

	[base,exponent,sbase,sexponent] = FlOpCtr_PrepareElemOp(base, flopctr(exponent))
	if (or(FlOpCtr_GetValue(exponent) > 1)) then
		[r,listPowers2,listMasks,listProds,rootExpo,rootFactor] = elevatePowerElem(base, exponent)
	else
		r = flopctr(FlOpCtr_GetValue(base) .^ FlOpCtr_GetValue(exponent))
		nbElems = prod(sbase)
		FLOPCTR_OPCOUNTS(FLOPCTR_CPLXPRIMS) = FLOPCTR_OPCOUNTS(FLOPCTR_CPLXPRIMS) + nbElems
		FLOPCTR_OPCOUNTS(FLOPCTR_FETCHES) = FLOPCTR_OPCOUNTS(FLOPCTR_FETCHES) + 2 * nbElems
		FLOPCTR_OPCOUNTS(FLOPCTR_STORES) = FLOPCTR_OPCOUNTS(FLOPCTR_STORES) + nbElems
	end  // if
endfunction


function [r]=%FLOPCTR_exp(arg)
	// Element-wise natural exponential.
	//
	r = %e .^ arg
	//global FLOPCTR_OPCOUNTS FLOPCTR_ADDS FLOPCTR_MULTS FLOPCTR_INVS FLOPCTR_FETCHES FLOPCTR_STORES FLOPCTR_CPLXPRIMS;
	//r = flopctr(exp(FlOpCtr_GetValue(arg)))
	//nbElems = size(arg, '*')
	//FLOPCTR_OPCOUNTS(FLOPCTR_CPLXPRIMS) = FLOPCTR_OPCOUNTS(FLOPCTR_CPLXPRIMS) + nbElems
	//FLOPCTR_OPCOUNTS(FLOPCTR_FETCHES) = FLOPCTR_OPCOUNTS(FLOPCTR_FETCHES) + nbElems
	//FLOPCTR_OPCOUNTS(FLOPCTR_STORES) = FLOPCTR_OPCOUNTS(FLOPCTR_STORES) + nbElems
endfunction


function [r]=%FLOPCTR_sqrt(arg)
	// Element-wise square root.
	//
	r = arg .^ (0.5)
	//global FLOPCTR_OPCOUNTS FLOPCTR_ADDS FLOPCTR_MULTS FLOPCTR_INVS FLOPCTR_FETCHES FLOPCTR_STORES FLOPCTR_CPLXPRIMS;
	//r = flopctr(sqrt(FlOpCtr_GetValue(arg)))
	//nbElems = size(arg, '*')
	//FLOPCTR_OPCOUNTS(FLOPCTR_CPLXPRIMS) = FLOPCTR_OPCOUNTS(FLOPCTR_CPLXPRIMS) + nbElems
	//FLOPCTR_OPCOUNTS(FLOPCTR_FETCHES) = FLOPCTR_OPCOUNTS(FLOPCTR_FETCHES) + nbElems
	//FLOPCTR_OPCOUNTS(FLOPCTR_STORES) = FLOPCTR_OPCOUNTS(FLOPCTR_STORES) + nbElems
endfunction


function [r]=%FLOPCTR_o_FLOPCTR(left, right)
	// Equality comparison.
	//
	global FLOPCTR_OPCOUNTS FLOPCTR_ADDS FLOPCTR_MULTS FLOPCTR_INVS FLOPCTR_FETCHES FLOPCTR_STORES FLOPCTR_CPLXPRIMS;
	sleft = size(left)
	sright = size(right)
	if (or(sleft <> sright)) then
		r = %f
	else
		r = (FlOpCtr_GetValue(left) == FlOpCtr_GetValue(right))
		nbElems = prod(sleft)
		FLOPCTR_OPCOUNTS(FLOPCTR_ADDS) = FLOPCTR_OPCOUNTS(FLOPCTR_ADDS) + nbElems
		FLOPCTR_OPCOUNTS(FLOPCTR_FETCHES) = FLOPCTR_OPCOUNTS(FLOPCTR_FETCHES) + 2 * nbElems
		FLOPCTR_OPCOUNTS(FLOPCTR_STORES) = FLOPCTR_OPCOUNTS(FLOPCTR_STORES) + nbElems
	end  // if
endfunction


function [r]=%FLOPCTR_o_s(left, right)
	if (right == []) then
		r = (FlOpCtr_GetValue(left) == [])
	else
		r = (left == flopctr(right))
	end  // if
endfunction


function [r]=%s_o_FLOPCTR(left, right)
	if (left == []) then
		r = (FlOpCtr_GetValue(right) == [])
	else
		r = (flopctr(left) == right)
	end  // if
endfunction


function [r]=%FLOPCTR_n_FLOPCTR(left, right)
	// Difference comparison.
	//
	r = ~(left == right)
endfunction


function [r]=%FLOPCTR_n_s(left, right)
	r = ~(left == right)
endfunction


function [r]=%s_n_FLOPCTR(left, right)
	r = ~(left == right)
endfunction


function [r]=%FLOPCTR_1_FLOPCTR(left, right)
	// Strict order comparison.
	//
	global FLOPCTR_OPCOUNTS FLOPCTR_ADDS FLOPCTR_MULTS FLOPCTR_INVS FLOPCTR_FETCHES FLOPCTR_STORES FLOPCTR_CPLXPRIMS;
	sleft = size(left)
	sright = size(right)
	if (or(sleft <> sright)) then
		error("Cannot order-compare matrices of discrepant sizes.")
	end  // if
	r = (FlOpCtr_GetValue(left) < FlOpCtr_GetValue(right))
	nbElems = prod(sleft)
	FLOPCTR_OPCOUNTS(FLOPCTR_ADDS) = FLOPCTR_OPCOUNTS(FLOPCTR_ADDS) + nbElems
	FLOPCTR_OPCOUNTS(FLOPCTR_FETCHES) = FLOPCTR_OPCOUNTS(FLOPCTR_FETCHES) + 2 * nbElems
	FLOPCTR_OPCOUNTS(FLOPCTR_STORES) = FLOPCTR_OPCOUNTS(FLOPCTR_STORES) + nbElems
endfunction


function [r]=%FLOPCTR_1_s(left, right)
	if (right == []) then
		r = []
	else
		r = (left < flopctr(right))
	end  // if
endfunction


function [r]=%s_1_FLOPCTR(left, right)
	if (left == []) then
		r = []
	else
		r = (flopctr(left) < right)
	end  // if
endfunction


function [r]=%FLOPCTR_2_FLOPCTR(left, right)
	// Strict inverse order comparison.
	//
	global FLOPCTR_OPCOUNTS FLOPCTR_ADDS FLOPCTR_MULTS FLOPCTR_INVS FLOPCTR_FETCHES FLOPCTR_STORES FLOPCTR_CPLXPRIMS;
	sleft = size(left)
	sright = size(right)
	if (or(sleft <> sright)) then
		error("Cannot order-compare matrices of discrepant sizes.")
	end  // if
	r = (FlOpCtr_GetValue(left) > FlOpCtr_GetValue(right))
	nbElems = prod(sleft)
	FLOPCTR_OPCOUNTS(FLOPCTR_ADDS) = FLOPCTR_OPCOUNTS(FLOPCTR_ADDS) + nbElems
	FLOPCTR_OPCOUNTS(FLOPCTR_FETCHES) = FLOPCTR_OPCOUNTS(FLOPCTR_FETCHES) + 2 * nbElems
	FLOPCTR_OPCOUNTS(FLOPCTR_STORES) = FLOPCTR_OPCOUNTS(FLOPCTR_STORES) + nbElems
endfunction


function [r]=%FLOPCTR_2_s(left, right)
	if (right == []) then
		r = []
	else
		r = (left > flopctr(right))
	end  // if
endfunction


function [r]=%s_2_FLOPCTR(left, right)
	if (left == []) then
		r = []
	else
		r = (flopctr(left) > right)
	end  // if
endfunction


function [r]=%FLOPCTR_3_FLOPCTR(left, right)
	// Order comparison.
	//
	global FLOPCTR_OPCOUNTS FLOPCTR_ADDS FLOPCTR_MULTS FLOPCTR_INVS FLOPCTR_FETCHES FLOPCTR_STORES FLOPCTR_CPLXPRIMS;
	sleft = size(left)
	sright = size(right)
	if (or(sleft <> sright)) then
		error("Cannot order-compare matrices of discrepant size.")
	end  // if
	r = (FlOpCtr_GetValue(left) <= FlOpCtr_GetValue(right))
	nbElems = prod(sleft)
	FLOPCTR_OPCOUNTS(FLOPCTR_ADDS) = FLOPCTR_OPCOUNTS(FLOPCTR_ADDS) + nbElems
	FLOPCTR_OPCOUNTS(FLOPCTR_FETCHES) = FLOPCTR_OPCOUNTS(FLOPCTR_FETCHES) + 2 * nbElems
	FLOPCTR_OPCOUNTS(FLOPCTR_STORES) = FLOPCTR_OPCOUNTS(FLOPCTR_STORES) + nbElems
endfunction


function [r]=%FLOPCTR_3_s(left, right)
	if (right == []) then
		r = []
	else
		r = (left <= flopctr(right))
	end  // if
endfunction


function [r]=%s_3_FLOPCTR(left, right)
	if (left == []) then
		r = []
	else
		r = (flopctr(left) <= right)
	end  // if
endfunction


function [r]=%FLOPCTR_4_FLOPCTR(left, right)
	// Inverse order comparison.
	//
	global FLOPCTR_OPCOUNTS FLOPCTR_ADDS FLOPCTR_MULTS FLOPCTR_INVS FLOPCTR_FETCHES FLOPCTR_STORES FLOPCTR_CPLXPRIMS;
	sleft = size(left)
	sright = size(right)
	if (or(sleft <> sright)) then
		error("Cannot order-compare matrices of discrepant size.")
	end  // if
	r = (FlOpCtr_GetValue(left) >= FlOpCtr_GetValue(right))
	nbElems = prod(sleft)
	FLOPCTR_OPCOUNTS(FLOPCTR_ADDS) = FLOPCTR_OPCOUNTS(FLOPCTR_ADDS) + nbElems
	FLOPCTR_OPCOUNTS(FLOPCTR_FETCHES) = FLOPCTR_OPCOUNTS(FLOPCTR_FETCHES) + 2 * nbElems
	FLOPCTR_OPCOUNTS(FLOPCTR_STORES) = FLOPCTR_OPCOUNTS(FLOPCTR_STORES) + nbElems
endfunction


function [r]=%FLOPCTR_4_s(left, right)
	if (right == []) then
		r = []
	else
		r = (left >= flopctr(right))
	end  // if
endfunction


function [r]=%s_4_FLOPCTR(left, right)
	if (left == []) then
		r = []
	else
		r = (flopctr(left) >= right)
	end  // if
endfunction


function [r]=%FLOPCTR_log(fc)
	// Natural (element-wise) logarithm.
	//
	global FLOPCTR_OPCOUNTS FLOPCTR_ADDS FLOPCTR_MULTS FLOPCTR_INVS FLOPCTR_FETCHES FLOPCTR_STORES FLOPCTR_CPLXPRIMS;
	r = flopctr(log(FlOpCtr_GetValue(fc)))
	nbElems = size(fc, "*")
	FLOPCTR_OPCOUNTS(FLOPCTR_CPLXPRIMS) = FLOPCTR_OPCOUNTS(FLOPCTR_CPLXPRIMS) + nbElems
	FLOPCTR_OPCOUNTS(FLOPCTR_FETCHES) = FLOPCTR_OPCOUNTS(FLOPCTR_FETCHES) + nbElems
	FLOPCTR_OPCOUNTS(FLOPCTR_STORES) = FLOPCTR_OPCOUNTS(FLOPCTR_STORES) + nbElems
endfunction


function [r]=%FLOPCTR_sin(fc)
	// Circular element-wise sine.
	//
	global FLOPCTR_OPCOUNTS FLOPCTR_ADDS FLOPCTR_MULTS FLOPCTR_INVS FLOPCTR_FETCHES FLOPCTR_STORES FLOPCTR_CPLXPRIMS;
	r = flopctr(sin(FlOpCtr_GetValue(fc)))
	nbElems = size(fc, "*")
	FLOPCTR_OPCOUNTS(FLOPCTR_CPLXPRIMS) = FLOPCTR_OPCOUNTS(FLOPCTR_CPLXPRIMS) + nbElems
	FLOPCTR_OPCOUNTS(FLOPCTR_FETCHES) = FLOPCTR_OPCOUNTS(FLOPCTR_FETCHES) + nbElems
	FLOPCTR_OPCOUNTS(FLOPCTR_STORES) = FLOPCTR_OPCOUNTS(FLOPCTR_STORES) + nbElems
endfunction


function [r]=%FLOPCTR_cos(fc)
	// Circular element-wise cosine.
	//
	global FLOPCTR_OPCOUNTS FLOPCTR_ADDS FLOPCTR_MULTS FLOPCTR_INVS FLOPCTR_FETCHES FLOPCTR_STORES FLOPCTR_CPLXPRIMS;
	r = flopctr(cos(FlOpCtr_GetValue(fc)))
	nbElems = size(fc, "*")
	FLOPCTR_OPCOUNTS(FLOPCTR_CPLXPRIMS) = FLOPCTR_OPCOUNTS(FLOPCTR_CPLXPRIMS) + nbElems
	FLOPCTR_OPCOUNTS(FLOPCTR_FETCHES) = FLOPCTR_OPCOUNTS(FLOPCTR_FETCHES) + nbElems
	FLOPCTR_OPCOUNTS(FLOPCTR_STORES) = FLOPCTR_OPCOUNTS(FLOPCTR_STORES) + nbElems
endfunction


function [r]=%FLOPCTR_abs(fc)
	// Absolute value.
	//
	global FLOPCTR_OPCOUNTS FLOPCTR_ADDS FLOPCTR_MULTS FLOPCTR_INVS FLOPCTR_FETCHES FLOPCTR_STORES FLOPCTR_CPLXPRIMS;
	r = flopctr(abs(FlOpCtr_GetValue(fc)))
	nbElems = size(fc, "*")
	FLOPCTR_OPCOUNTS(FLOPCTR_FETCHES) = FLOPCTR_OPCOUNTS(FLOPCTR_FETCHES) + nbElems
	FLOPCTR_OPCOUNTS(FLOPCTR_STORES) = FLOPCTR_OPCOUNTS(FLOPCTR_STORES) + nbElems
endfunction


function [r]=%FLOPCTR_t(fc)
	// Transposition.
	//
	global FLOPCTR_OPCOUNTS FLOPCTR_ADDS FLOPCTR_MULTS FLOPCTR_INVS FLOPCTR_FETCHES FLOPCTR_STORES FLOPCTR_CPLXPRIMS;
	r = flopctr(FlOpCtr_GetValue(fc)')
	nbElems = size(fc, "*")
	FLOPCTR_OPCOUNTS(FLOPCTR_FETCHES) = FLOPCTR_OPCOUNTS(FLOPCTR_FETCHES) + nbElems
	FLOPCTR_OPCOUNTS(FLOPCTR_STORES) = FLOPCTR_OPCOUNTS(FLOPCTR_STORES) + nbElems
endfunction


function [r]=%FLOPCTR_norm(varargin)
	// Norm.  Only L2 variant is yet implemented.
	// TODO - Unit tests.
	//
	global FLOPCTR_OPCOUNTS FLOPCTR_ADDS FLOPCTR_MULTS FLOPCTR_INVS FLOPCTR_FETCHES FLOPCTR_STORES FLOPCTR_CPLXPRIMS;
	fc = varargin(1)
	if (size(fc, "r") <> 1 & size(fc, "c") <> 1) then
		error("Only L2 vector norm is yet implemented.")
	end  // if
	if (size(fc, "r") == 1) then
		r = sqrt(fc * fc')
	else
		r = sqrt(fc' * fc)
	end  // if
	//nbElems = size(fc, '*')
	//r = flopctr(norm(FlOpCtr_GetValue(fc)))
	//FLOPCTR_OPCOUNTS(FLOPCTR_FETCHES) = FLOPCTR_OPCOUNTS(FLOPCTR_FETCHES) + nbElems
	//FLOPCTR_OPCOUNTS(FLOPCTR_MULTS) = FLOPCTR_OPCOUNTS(FLOPCTR_MULTS) + nbElems
	//FLOPCTR_OPCOUNTS(FLOPCTR_ADDS) = FLOPCTR_OPCOUNTS(FLOPCTR_ADDS) + nbElems - 1
	//FLOPCTR_OPCOUNTS(FLOPCTR_STORES) = FLOPCTR_OPCOUNTS(FLOPCTR_STORES) + 1
endfunction


function [r]=%FLOPCTR_int(fc)
	// Gives the integral part of the argument.
	// TODO - Unit tests.
	//

	global FLOPCTR_OPCOUNTS FLOPCTR_ADDS FLOPCTR_MULTS FLOPCTR_INVS FLOPCTR_FETCHES FLOPCTR_STORES FLOPCTR_CPLXPRIMS;
	r = flopctr(int(FlOpCtr_GetValue(fc)))
	nbElems = size(fc, "*")
	FLOPCTR_OPCOUNTS(FLOPCTR_FETCHES) = FLOPCTR_OPCOUNTS(FLOPCTR_FETCHES) + nbElems
	FLOPCTR_OPCOUNTS(FLOPCTR_MULTS) = FLOPCTR_OPCOUNTS(FLOPCTR_MULTS) + nbElems
	FLOPCTR_OPCOUNTS(FLOPCTR_STORES) = FLOPCTR_OPCOUNTS(FLOPCTR_STORES) + nbElems
endfunction


function [r]=%FLOPCTR_diag(fc)
	// For a vector fc (of R^n), gives a n x n diagonal matrix with the
	// entries of fc on the diagonal.  For a matrix fc, gives the elements of
	// its main diagonal as a column vector.
	// TODO -- Add support for parameter k.
	// TODO -- Unit tests.
	//

	global FLOPCTR_OPCOUNTS FLOPCTR_ADDS FLOPCTR_MULTS FLOPCTR_INVS FLOPCTR_FETCHES FLOPCTR_STORES FLOPCTR_CPLXPRIMS;
	r = flopctr(diag(FlOpCtr_GetValue(fc)))
	sfc = size(fc)
	nbElems = prod(sfc)
	if and(sfc <> [1,1]) then
		// Matrix case.
		nbElems = sfc(1)
	end  // if
	FLOPCTR_OPCOUNTS(FLOPCTR_FETCHES) = FLOPCTR_OPCOUNTS(FLOPCTR_FETCHES) + nbElems
	FLOPCTR_OPCOUNTS(FLOPCTR_STORES) = FLOPCTR_OPCOUNTS(FLOPCTR_STORES) + nbElems
endfunction


function [r]=%FLOPCTR_maxi(fc)
	// Returns maximal element of vector/matrix.
	//
	// TODO -- Support more complex uses of max().
	// TODO -- Unit tests.
	//

	global FLOPCTR_OPCOUNTS FLOPCTR_ADDS FLOPCTR_MULTS FLOPCTR_INVS FLOPCTR_FETCHES FLOPCTR_STORES FLOPCTR_CPLXPRIMS;
	r = flopctr(max(FlOpCtr_GetValue(fc)))
	nbElems = size(fc, '*')
	FLOPCTR_OPCOUNTS(FLOPCTR_FETCHES) = FLOPCTR_OPCOUNTS(FLOPCTR_FETCHES) + nbElems
	FLOPCTR_OPCOUNTS(FLOPCTR_STORES) = FLOPCTR_OPCOUNTS(FLOPCTR_STORES) + 1
	FLOPCTR_OPCOUNTS(FLOPCTR_ADDS) = FLOPCTR_OPCOUNTS(FLOPCTR_ADDS) + nbElems
endfunction


function [r]=%FLOPCTR_mini(fc)
	// Returns minimal element of vector/matrix.
	//
	// TODO -- Support more complex uses of min().
	// TODO -- Unit tests.
	//

	global FLOPCTR_OPCOUNTS FLOPCTR_ADDS FLOPCTR_MULTS FLOPCTR_INVS FLOPCTR_FETCHES FLOPCTR_STORES FLOPCTR_CPLXPRIMS;
	r = flopctr(min(FlOpCtr_GetValue(fc)))
	nbElems = size(fc, '*')
	FLOPCTR_OPCOUNTS(FLOPCTR_FETCHES) = FLOPCTR_OPCOUNTS(FLOPCTR_FETCHES) + nbElems
	FLOPCTR_OPCOUNTS(FLOPCTR_STORES) = FLOPCTR_OPCOUNTS(FLOPCTR_STORES) + 1
	FLOPCTR_OPCOUNTS(FLOPCTR_ADDS) = FLOPCTR_OPCOUNTS(FLOPCTR_ADDS) + nbElems
endfunction


// End of module.
