//
// Functions : definition of FlOpCtr, used through operator overloading.
//
// Author : Benoit Hamelin
// 2003 Universit de Sherbrooke
//


function [fc]=flopctr(value)
	// Makes a new FlOpCtr object associated to given value.
	//
	fc = mlist(FLOPCTR_FIELDS, value);
endfunction


function []=FLOPCTR_Reset()
	// Resets the floating-point operations counter.
	//
	global FLOPCTR_OPCOUNTS FLOPCTR_ADDS FLOPCTR_MULTS FLOPCTR_INVS FLOPCTR_FETCHES FLOPCTR_STORES FLOPCTR_CPLXPRIMS;
	FLOPCTR_OPCOUNTS = zeros(1, 6);
endfunction


function [v]=FlOpCtr_GetValue(this)
	// Gives value bound to FlOpCtr.
	v = this.value
endfunction


function varargout=%FLOPCTR_size(varargin)
	[lhs,rhs] = argn()
	if (rhs < 1) then
		error("Not enough input parameters.")
	end  // if
	v = FlOpCtr_GetValue(varargin(1))

	if (lhs == 1) then
		if (rhs == 1)
			varargout(1) = size(v)
		elseif (rhs == 2)
			varargout(1) = size(v, varargin(2))
		else
			error("Too many input parameters.")
		end  // if
	elseif (lhs == 2) then
		if (rhs > 1) then
			error("Too many input parameters for this call sequence.")
		end  // if
		[r,c] = size(v)
		varargout(1) = r
		varargout(2) = c
	else
		error("Too many output parameters.")
	end  // if
endfunction


function [r]=%FLOPCTR_zeros(fc)
	r = zeros(FlOpCtr_GetValue(fc))
endfunction


function [r]=%FLOPCTR_rand(fc)
	r = rand(FlOpCtr_GetValue(fc))
endfunction


function [r]=%FLOPCTR_ones(fc)
	r = ones(FlOpCtr_GetValue(fc))
endfunction


function [r]=%FLOPCTR_eye(fc)
	r = eye(FlOpCtr_GetValue(fc))
endfunction


function [r]=%FLOPCTR_e(varargin)
	// Extraction operator.
	//
	global FLOPCTR_OPCOUNTS FLOPCTR_ADDS FLOPCTR_MULTS FLOPCTR_INVS FLOPCTR_FETCHES FLOPCTR_STORES FLOPCTR_CPLXPRIMS;
	[lhs,rhs] = argn()
	if (rhs < 2) then
		error("Not enough argument for extraction from a FlOpCounter.")
	elseif (rhs == 2) then
		if (size(definedfields(varargin), "*") <> 2) then
			r = varargin($)
		else
			v = FlOpCtr_GetValue(varargin($))
			r = flopctr(v(varargin(1)))
		end  // if
	elseif (rhs == 3) then
		if (size(definedfields(varargin), "*") <> 3) then
			r = varargin($)
		else
			v = FlOpCtr_GetValue(varargin($))
			r = flopctr(v(varargin(1), varargin(2)))
		end  // if
	else
		error("Too many arguments for extraction from a FlOpCounter.")
	end  // if

	nbElemsExtracted = size(r, "*")
	FLOPCTR_OPCOUNTS(FLOPCTR_FETCHES) = FLOPCTR_OPCOUNTS(FLOPCTR_FETCHES) + nbElemsExtracted
	FLOPCTR_OPCOUNTS(FLOPCTR_STORES) = FLOPCTR_OPCOUNTS(FLOPCTR_STORES) + nbElemsExtracted
endfunction


function [r]=%FLOPCTR_c_FLOPCTR(left, right)
	// Horizontal concatenation.
	//
	global FLOPCTR_OPCOUNTS FLOPCTR_ADDS FLOPCTR_MULTS FLOPCTR_INVS FLOPCTR_FETCHES FLOPCTR_STORES FLOPCTR_CPLXPRIMS;
//	sleft = size(left)
//	sright = size(right)
//	if (prod(sleft) == 0) then
//		r = right
//	elseif (prod(sright) == 0) then
//		r = left
//	elseif (sleft(1) <> sright(1)) then
//		error("Cannot horizontally concatenate matrices with discrepant number or rows.")
//	else
//		mtcLeft = [eye(sleft(2), sleft(2)), zeros(sleft(2), sright(2))]
//		mtcRight = [zeros(sright(2), sleft(2)), eye(sright(2), sright(2))]
//		r = left * mtcLeft + right * mtcRight
//	end  // if
	r = flopctr([FlOpCtr_GetValue(left), FlOpCtr_GetValue(right)])
	nbElems = size(r, "*")
	FLOPCTR_OPCOUNTS(FLOPCTR_FETCHES) = FLOPCTR_OPCOUNTS(FLOPCTR_FETCHES) + ...
			nbElems
	FLOPCTR_OPCOUNTS(FLOPCTR_STORES) = FLOPCTR_OPCOUNTS(FLOPCTR_STORES) + ...
			nbElems
endfunction


function [r]=%FLOPCTR_c_s(left, right)
	r = [left, flopctr(right)]
endfunction


function [r]=%s_c_FLOPCTR(left, right)
	r = [flopctr(left), right]
endfunction


function [r]=%FLOPCTR_f_FLOPCTR(up, down)
	// Vertical concatenation.
	//
	global FLOPCTR_OPCOUNTS FLOPCTR_ADDS FLOPCTR_MULTS FLOPCTR_INVS FLOPCTR_FETCHES FLOPCTR_STORES FLOPCTR_CPLXPRIMS;
//	sup = size(up)
//	sdown = size(down)
//	if (prod(sup) == 0) then
//		r = down
//	elseif (prod(sdown) == 0) then
//		r = up
//	elseif (sup(2) <> sdown(2)) then
//		error("Cannot vertically concatenate matrices with discrepant number of columns.")
//	else
//		mtcUp = [eye(sup(1), sup(1)) ; zeros(sdown(1), sup(1))]
//		mtcDown = [zeros(sup(1), sdown(1)) ; eye(sdown(1), sdown(1))]
//		r = mtcUp * up + mtcDown * down
//	end  // if
	r = flopctr([FlOpCtr_GetValue(up) ; FlOpCtr_GetValue(down)])
	nbElems = size(r, "*")
	FLOPCTR_OPCOUNTS(FLOPCTR_FETCHES) = FLOPCTR_OPCOUNTS(FLOPCTR_FETCHES) + ...
			nbElems
	FLOPCTR_OPCOUNTS(FLOPCTR_STORES) = FLOPCTR_OPCOUNTS(FLOPCTR_STORES) + ...
			nbElems
endfunction


function [r]=%FLOPCTR_f_s(up, down)
	r = [up ; flopctr(down)]
endfunction


function [r]=%s_f_FLOPCTR(up, down)
	r = [flopctr(up) ; down]
endfunction


function [this]=%FLOPCTR_i_FLOPCTR(varargin)
	// Insertion operator.
	//
	global FLOPCTR_OPCOUNTS FLOPCTR_ADDS FLOPCTR_MULTS FLOPCTR_INVS FLOPCTR_FETCHES FLOPCTR_STORES FLOPCTR_CPLXPRIMS;
	[lhs,rhs] = argn()
	if (rhs < 3) then
		error("No index given.")
	end  // if

	this = varargin($)
	thisValue = FlOpCtr_GetValue(this)
	ins = varargin($-1)
	insValue = FlOpCtr_GetValue(ins)

	if (size(definedfields(varargin), "*") == size(varargin)) then
		indexCount = size(varargin) - 2
		if (indexCount > 2) then
			error("No more than two index vectors are supported.")
		elseif (indexCount == 1) then
			//this = Insert1Index(this, ins, varargin(1))
			thisValue(varargin(1)) = insValue
		elseif (indexCount == 2) then
			//this = Insert2Index(this, ins, varargin(1), varargin(2))
			thisValue(varargin(1), varargin(2)) = insValue
		end  // if
		this = flopctr(thisValue)
		nbElemsIns = size(insValue, "*")
		FLOPCTR_OPCOUNTS(FLOPCTR_FETCHES) = ...
				FLOPCTR_OPCOUNTS(FLOPCTR_FETCHES) + nbElemsIns
		FLOPCTR_OPCOUNTS(FLOPCTR_STORES) = ...
				FLOPCTR_OPCOUNTS(FLOPCTR_STORES) + nbElemsIns
	end  // if
endfunction


function [this]=%s_i_FLOPCTR(varargin)
	global FLOPCTR_OPCOUNTS FLOPCTR_ADDS FLOPCTR_MULTS FLOPCTR_INVS FLOPCTR_FETCHES FLOPCTR_STORES FLOPCTR_CPLXPRIMS;
	[lhs,rhs] = argn()
	if (rhs < 3) then
		error("No index given.")
	end  // if

	this = varargin($)
	thisValue = FlOpCtr_GetValue(this)
	insValue = varargin($-1)

	if (size(definedfields(varargin), "*") == size(varargin)) then
		indexCount = size(varargin) - 2
		if (indexCount > 2) then
			error("No more than two index vectors are supported.")
		elseif (indexCount == 1) then
			//this = Insert1Index(this, ins, varargin(1))
			thisValue(varargin(1)) = insValue
		else
			//this = Insert2Index(this, ins, varargin(1), varargin(2))
			thisValue(varargin(1), varargin(2)) = insValue
		end  // if
		this = flopctr(thisValue)
		nbElemsIns = size(insValue, "*")
		FLOPCTR_OPCOUNTS(FLOPCTR_FETCHES) = ...
				FLOPCTR_OPCOUNTS(FLOPCTR_FETCHES) + nbElemsIns
		FLOPCTR_OPCOUNTS(FLOPCTR_STORES) = ...
				FLOPCTR_OPCOUNTS(FLOPCTR_STORES) + nbElemsIns
	end  // if
endfunction


function [this]=%FLOPCTR_i_s(varargin)
	global FLOPCTR_OPCOUNTS FLOPCTR_ADDS FLOPCTR_MULTS FLOPCTR_INVS FLOPCTR_FETCHES FLOPCTR_STORES FLOPCTR_CPLXPRIMS;
	[lhs,rhs] = argn()
	if (rhs < 3) then
		error("No index given.")
	end  // if

//	this = varargin($)
//	if (size(this, "*") == 0) then
//		this = flopctr(0)
//	else
//		this = flopctr(this)
//	end  // if
	thisValue = varargin($)
	ins = varargin($-1)
	insValue = FlOpCtr_GetValue(ins)

	if (size(definedfields(varargin), "*") == size(varargin)) then
		indexCount = size(varargin) - 2
		if (indexCount > 2) then
			error("No more than two index vectors are supported.")
		elseif (indexCount == 1) then
			//this = Insert1Index(this, ins, varargin(1))
			thisValue(varargin(1)) = insValue
		else
			//this = Insert2Index(this, ins, varargin(1), varargin(2))
			thisValue(varargin(1), varargin(2)) = insValue
		end  // if
		this = flopctr(thisValue)
		nbElemsIns = size(insValue, "*")
		FLOPCTR_OPCOUNTS(FLOPCTR_FETCHES) = ...
				FLOPCTR_OPCOUNTS(FLOPCTR_FETCHES) + nbElemsIns
		FLOPCTR_OPCOUNTS(FLOPCTR_STORES) = ...
				FLOPCTR_OPCOUNTS(FLOPCTR_STORES) + nbElemsIns
	end  // if
endfunction


// End of module.
