//
// Mathematically sound implementation of insertion of elements in a vector/matrix.
//
// Author : Benoit Hamelin
// 2003 Universit de Sherbrooke.
//


function this=Insert1Index(this, ins, index)
	// Insertion of elements of a vector in a matrix at arbitrary indices.  Original matrix is seen as a column vector resulting of 
	// the vertical concatenation of its own columns.

	sthis = size(this)
	sins = size(ins)
	sindex = size(index)
	if (prod(sins) <> prod(sindex)) then
		error("Discrepancy between size of replaced and size of inserted sub-matrices.")
	end  // if
	index = concretizeIndexVector(index, prod(sthis))
	if (or(index < 1)) then
		error("Invalid index.")
	end  // if

	// We consider simplifications for insertion of a contiguous vector.
	if (and(index == (index(1):index($)))) then

		// Vector of consecutive indices.
		if (or(sthis == 1)) then

			// Vector to vector insertion.
			if (sthis(1) == 1 & sthis(2) <> 1) then
				// Original is a row vector.
				if (sins(1) <> 1) then
					ins = ins'
				end  // if
				if (index($) > sthis(2)) then
					// Insertion lengthens the vector.
					this = [this, zeros(1, index($) - sthis(2))]
				end  // if
				if (index(1) > sthis(2)) then
					// Inserted sub-vector does not overwrite original.
					this = this + [zeros(1, index(1) - 1), ins]
				else
					// Inserted sub-vector overwrites part of original.
					before = 1:(index(1)-1)
					after = (index($)+1):sthis(2)
					mask = [ones(before), zeros(ins), ones(after)]
					ins = [zeros(before), ins, zeros(after)]
					this = this .* mask + ins
				end  // if
			else
				// Original is a column vector.
				if (sins(2) <> 1) then
					ins = ins'
				end  // if
				if (index($) > sthis(1)) then
					// Insertion lengthens the vector.
					this = [this; zeros(index($) - sthis(1), 1)]
				end  // if
				if (index(1) > sthis(1)) then
					// Inserted sub-vector does not overwrite original.
					this = this + [zeros(index(1) - 1, 1) ; ins]
				else
					// Inserted sub-vector overwrites part of original.
					before = (1:(index(1)-1))'
					after = ((index($)+1):sthis(1))'
					mask = [ones(before) ; zeros(ins) ; ones(after)]
					ins = [zeros(before) ; ins ; zeros(after)]
					this = this .* mask + ins
				end  // if
			end  // if

		else
			// Vector to matrix insertion.  In this case, we don't allow out of bound indices.
			if (or(index > prod(sthis))) then
				error("Out of bound index forbidden in this case.")
			end  // if
			if (sins(2) <> 1) then
				ins = ins'
				sins = size(ins)
			end  // if

			indexFirst = indexUni2Multi(index(1), sthis)
			indexLast = indexUni2Multi(index($), sthis)
			if (indexFirst(2) == indexLast(2)) then
				// Vector overwrites only part of a column of the matrix.
				before = (1:(indexFirst(1)-1))'
				after = ((indexLast(1)+1):sthis(1))'
				ins = [zeros(sthis(1), indexFirst(2) - 1), ...
					[zeros(before) ; ins ; zeros(after)], ...
					zeros(sthis(1), sthis(2) - indexFirst(2))]
			else
				// Vector overwrites part of at least two columns of the matrix.
				mins = zeros(sthis(1), indexFirst(2) - 1)
				lengthBefore = indexFirst(1) - 1
				lengthAfter = indexLast(1)
				mins = [mins , [zeros(indexFirst(1) - 1, 1) ; ins(1:(sthis(1)-lengthBefore))]]
				for i = (sthis(1)-lengthBefore+1):sthis(1):(prod(sins)-sthis(1)) do
					mins = [mins, ins(i:(i + sthis(1) - 1))]
				end  // for
				mins = [mins , [ins(($ - lengthAfter + 1):$) ; zeros(sthis(1) - indexLast(1), 1)], ...
					zeros(sthis(1), sthis(2) - indexLast(2))]
				ins = mins
			end  // if

			mask = ones(this)
			mask(index) = zeros(index)
			this = this .* mask + ins
		end  // if

	else

		// Arbitrary index vector.  Section the index vector into vectors of consecutive indices.
		listIndexVec = list()
		listIns = list()
		current = index(1)
		j = 1
		for i = index(2:$) do
			if (current == []) then
				current = i
			elseif (i == (current($) + 1)) then
				current = [current, i]
			else
				listIndexVec($+1) = current
				sc = size(current, "*")
				listIns($+1) = ins(j:(j+sc-1))
				current = i
				j = j + sc
			end  // if
		end  // for
		listIndexVec($+1) = current
		listIns($+1) = ins(j:$)

		// Process insertion by multiple contiguous-index-vector insertions.
		for i = 1:size(listIndexVec) do
			this(listIndexVec(i)) = listIns(i)
		end  // for
	end
endfunction


function this=Insert2Index(this, ins, indexRow, indexCol)
	// Insertion operator using two-dim indices.

	sthis = size(this)
	sins = size(ins)
	indexRow = concretizeIndexVector(indexRow, sthis(1))
	indexCol = concretizeIndexVector(indexCol, sthis(2))
	if (or(indexRow < 1) | (size(indexRow,"*") <> sins(1))) then
		error("Invalid row index.")
	end  // if
	if (or(indexCol < 1) | (size(indexCol,"*") <> sins(2))) then
		error("Invalid column index.")
	end  // if

	// We consider special case of insertion of a contiguous sub-matrix.
	if (and(indexRow == (indexRow(1):indexRow($))) & and(indexCol == (indexCol(1):indexCol($)))) then

		// Contiguous block sub-matrix.
		if (indexRow($) > sthis(1)) then
			// Original matrix must grow vertically to accomodate sub-matrix.
			this = [this ; zeros(indexRow($) - sthis(1), sthis(2))]
			sthis = size(this)
		end  // if
		if (indexCol($) > sthis(2)) then
			// Original matrix must grow horizontally to accomodate sub-matrix.
			this = [this , zeros(sthis(1), indexCol($) - sthis(2))]
			sthis = size(this)
		end  // if

		// Prepare filter for original matrix.
		mask = ones(this)
		mask(indexRow, indexCol) = zeros(size(indexRow,"*"), size(indexCol,"*"))

		// Resize inserted sub-matrix so it fits to be added to original.
		ins = [zeros(indexRow(1) - 1, sins(1)) ; eye(sins(1), sins(1)) ; zeros(sthis(1) - indexRow($), sins(1))] * ...
			ins * ...
			[zeros(sins(2), indexCol(1) - 1), eye(sins(2), sins(2)), zeros(sins(2), sthis(2) - indexCol($))]

		this = this .* mask + ins
	else

		// Arbitrary sub-matrix.  We define the continuous blocks that compose this sub-matrix and perform the insertion
		// block by block.
		listIndexVecRow = list()
		current = indexRow(1)
		for i = indexRow(2:$) do
			if (i == (current($) + 1)) then
				current = [current, i]
			else
				listIndexVecRow($+1) = current
				current = i
			end  // if
		end  // for
		listIndexVecRow($+1) = current

		listIndexVecCol = list()
		current = indexCol(1)
		for i = indexCol(2:$) do
			if (i == (current($) + 1)) then
				current = [current, i]
			else
				listIndexVecCol($+1) = current
				current = i
			end  // if
		end  // for
		listIndexVecCol($+1) = current

		indexExtrRowBase = 1
		for ir = listIndexVecRow do
			sr = size(ir, "*")
			indexExtrRow = indexExtrRowBase:(indexExtrRowBase + sr - 1)

			indexExtrColBase = 1
			for ic = listIndexVecCol do
				sc = size(ic, "*")
				indexExtrCol = indexExtrColBase:(indexExtrColBase + sc - 1)
				this(ir, ic) = ins(indexExtrRow, indexExtrCol)
				indexExtrColBase = indexExtrColBase + sc
			end  // for

			indexExtrRowBase = indexExtrRowBase + sr
		end  // for

	end  // if
endfunction



// End of script.
