//
// Numerical tools.
//
// Author : Benoit Hamelin
// 2003 Universit de Sherbrooke
//


function [r]=expm1m(x, accuracy)
	// (Regular) Exponential calculus of (exp(x) - 1), given at requested accuracy.
	r = eye(x)
	t = eye(x)
	i = 1
	while (or(abs(t) > accuracy)) do
		i = i + 1
		t = t * (x / i)
		r = r + t
	end  // while
	r = r * x
endfunction


function [r]=expm1(x, accuracy)
	// (Element-wise) Exponential calculus of (exp(x) - 1), given at requested accuracy.
	r = ones(x)
	t = ones(x)
	i = 1
	while (or(abs(t) > accuracy)) do
		i = i + 1
		t = t .* (x ./ i)
		r = r + t
	end  // while
	r = r .* x
endfunction


function [r,nbTerms]=cosm1(x, numzero)
	// MacLaurin series development of cosm1(x) = cos(x) - 1
	// We clamp the argument in the [-2pi, 2pi] interval to ensure accuracy.

	q = x / (2 * %pi)
	x = (q - fix(q)) * 2 * %pi

	i = 0
	d = 1
	n = ones(x)
	x2 = x .^ 2
	terms = []
	while (or(abs(n ./ d) > numzero) | (size(terms,"*") < 2)) do
		i = i + 2
		d = d * (i - 1) * i
		n = -n * x2
		terms = [terms ; n ./ d]
	end  // while
	r = sum(terms)
	nbTerms = size(terms, "*")
endfunction


function [power,listPowers2,listIndex,listProds,rootExpo,rootFactor]=elevatePower(base, exponent)
	// Calculates the power elevation of the base to the given exponent.  For the integral part of the exponent, a
	// divide-to-conquer algorithm is used.  The root part is calculated using Scilab primitive operations.
	// TODO - Write unit tests.

	if (exponent > 1) then
		i = 1
		k = fix(exponent)
		rootExpo = exponent - k
		int_k = 0
		frac_k = 0
		B = base
		listPowers2 = list()
		listIndex = list()
		listProds = list()
		while (%t) do
			k = k / 2
			int_k = fix(k)
			frac_k = k - int_k

			listPowers2(i) = B
			if (abs(frac_k) > %eps) then
				listIndex($+1) = i
			end  // if
			if (abs(int_k) < %eps) then
				break
			end  // if

			i = i + 1
			B = B * B
		end  // while

		power = listPowers2(listIndex(1))
		listProds(1) = power
		for i = 2:size(listIndex) do
			power = power * listPowers2(listIndex(i))
			listProds($+1) = power
		end  // for
		if (abs(rootExpo) > %eps) then
			rootFactor = base ^ rootExpo
			power = power * rootFactor
		else
			rootFactor = eye(base)
		end  // if
	else
		listPowers2 = list()
		listIndex = list()
		listProds = list()
		rootExpo = exponent
		rootFactor = base ^ exponent
		power = rootFactor
	end  // if
endfunction


function [power,listPowers2,listMasks,listProds,rootExpo,rootFactor]=elevatePowerElem(base, exponent)
	// Calculates the element-wise power elevation of the base to the given exponent.  For the integral part of the exponent, a
	// divide-to-conquer algorithm is used.  The root part is calculated using Scilab primitive operations.
	// TODO - Write unit tests.

	k = zeros(exponent)
	indexes_k = find(exponent > 1)
	if (indexes_k <> []) then
		k(indexes_k) = exponent(indexes_k)
	end  // if
	rootExpo = zeros(exponent)
	indexes_root = find(exponent <= 1)
	if (indexes_root <> []) then
		rootExpo(indexes_root) = exponent(indexes_root)
	end  // if
	power = []

	i = 1
	k = fix(k)
	d = exponent - k
	if (indexes_k <> []) then
		rootExpo(indexes_k) = d(indexes_k)
	end  // if
	int_k = 0
	frac_k = 0
	sb = size(base)
	B = base
	if (indexes_root <> []) then
		B(indexes_root) = 1
	end  // if
	indexesB = 1:size(B, "*")
	listPowers2 = list()
	listMasks = list()
	listProds = list()
	power = []
	if (or(k > 0)) then
		while (%t) do
			k = k ./ 2
			int_k = fix(k)
			frac_k = k - int_k
			k = int_k

			listPowers2(i) = B
			mask = ones(B)
			indexesNoNeed = find(abs(frac_k) < %eps)
			if (indexesNoNeed <> []) then
				if (sb(1) == 1) then
					mask(indexesNoNeed) = zeros(indexesNoNeed)
				else
					mask(indexesNoNeed) = zeros(indexesNoNeed)'
				end  // if
			end  // if
			listMasks(i) = mask

			indexDone = find(abs(int_k) < %eps)
			if (and(indexDone == indexesB)) then
				break
			end  // if
			if (indexDone <> []) then
				if (sb(1) == 1) then
					k(indexDone) = zeros(indexDone)
				else
					k(indexDone) = zeros(indexDone)'
				end  // if
			end  // if
			B = B .* B
			i = i + 1
		end  // while

		power = listPowers2(1) .^ listMasks(1)
		listProds(1) = power
		for i = 2:size(listPowers2) do
			index = find(listMasks(i) == 1)
			lp = listPowers2(i)
			if (index <> []) then
				power(index) = power(index) .* lp(index)
			end  // if
			listProds($+1) = power
		end  // for
	end  // if

	if (or(abs(rootExpo) > %eps)) then
		rootFactor = base .^ rootExpo
		if (power == []) then
			power = rootFactor
		else
			power = power .* rootFactor
		end  // if
	else
		rootFactor = ones(base)
	end  // if
endfunction


// End of module.
