//
// Functions : tools for processing multi-dimensional indices.
//
// Author : Benoit Hamelin
// 2002 Universite de Sherbrooke
//


function [ilist]=combineIndexDim2Multi(s)
	// From the given vector of dimensions, combines a matrix of all valid multi-dimensional indices, which stand on the rows of
	// the matrix.
	if (or(s <= 0)) then
		error('One of the given dimensions is negative or null.');
	end  // if
	if (size(s) == [1,1]) then
		ilist = (1:s)'
	else
		subList = combineIndexDim2Multi(s(1:($-1)))
		ssubList = size(subList)
		rowStruct = ones(ssubList(1), 1)
		dim = s($)
		ilist = []
		for i = 1:dim do
			ilist=[ilist ; subList, i .* rowStruct]
		end  // for
	end  // if
endfunction


function [imulti]=combineIndexVec2Multi(vecList)
	// From the elements of the given vectors (passed into a list), combines a matrix of all valid multi-dimensional indices, which
	// stand on the rows of the matrix.  The given vectors are expected to be row vectors.
	if (size(vecList) == 0)
		error('No vectors in given list.')
	end  // if

	lastVec = vecList($)
	if (or(lastVec <= 0))
		error('At least one component of a vector in the given list is null or negative.')
	end  // if
	if (lastVec == [])
		error('Empty vector present in given list.')
	end  // if

	if (size(vecList) == 1)
		imulti = lastVec'
	else
		vecList($) = null()
		subMulti = combineIndexVec2Multi(vecList)
		ssubMulti = size(subMulti)
		rowStruct = ones(ssubMulti(1), 1)
		imulti = []
		for i = lastVec do
			imulti = [imulti; subMulti, i .* rowStruct]
		end  // for
	end  // if
endfunction


function [multi]=indexUni2Multi(uni, s)
	// Transforms an uni-dimensional hypermatrix index into the corresponding multi-dimensional index (row vector).
	if (uni <= 0) then
		error('The uni-dim index is negative or null.');
	end  // if
//	if (uni > prod(s))
//		error('The uni-dim index is beyond range allowed by given dimensions.');
//	end  // if

	if (size(s) == [1,1])
		multi = uni
	else
		rest = s(1:($-1))
		depl = prod(rest)
		lastComp = ((uni - 1) / depl) + 1
		multi = indexUni2Multi(modulo(uni - 1, depl) + 1, rest)
		multi = [multi, lastComp]
	end  // if
	multi = floor(multi)
endfunction


function [uni]=indexMulti2Uni(multi, s)
	// Transforms a matrix of row-defined multi-dimensional indices to a column vector of corresponding uni-dimensional indices.
	// A boundary row vector that has an entry for each component of the multi-dimensional indices is given.
	uni = []
	if (multi ~= []) then
		ss = size(s)
		smulti = size(multi)
		if (smulti(2) ~= ss(2)) then
			error("Multi-indices and boundary vector aren''t the same size.")
		end

		incr = [1,s]
		incr($) = []
		incr = cumprod(incr)
		for col = multi' do
			r = col'
			if (or(r > s) | or(r < 1)) then
				error("A multi-index in the given matrix is too small or out of given boundaries.")
			end
			i = sum((r - 1) .* incr) + 1
			uni = [uni;i]
		end  // for
	end  // if
endfunction


function [v,out?]=concretizeIndexVector(vdesc, maxSize)
	// Concretizes an index vector description (which can contain the '$' polynomial and the eye shortcut) into an integer vector
	// of indices, by switching the $ with the given maximal size the vector can be.  Eye (identity) is substituted with a row
	// vector from 1 to given maximal size.
	//
	// The concretized vector of indices is returned as well as a flag stating whether maximal size is exceeded by index
	// specification.
	if (prod(size(maxSize)) ~= 1) then
		error("Maximal size is expected to be a scalar.")
	end  // if
	if (maxSize < 1) then
		error("Maximal size must be 1 or greater.")
	end  // if

	v = []
	if (type(vdesc) == 1) then  // vdesc is a simple constant, vector or eye.
		svdesc = size(vdesc)
		if (and(svdesc == [0,0])) then
			error("Empty index vector description.")
		elseif (and(svdesc == [-1,-1])) then
			v = 1:maxSize
		else
			v = vdesc
		end  // if
	else
		v = horner(vdesc, maxSize)
	end  // if

	//if (v == [] | or(v < 1) | or(v > maxSize)) then
	if (v == [] | or(v < 1)) then
		error("Eye and $ substitutions lead to empty vector of indices.")
	end
	out? = or(v > maxSize)
endfunction


// End of module.

