//
// Functions for the Derivator.
//
// Only implements derivation for functions built from "2D components" (no
// tensors involved).  Final derivative is given as a list of one vector
// (gradient), one matrix (jacobian or hessian) or (number of given
// directions) matrices (derivative of a function of R^(m*n), m > 1 and
// n > 1).
//
// FOR THE MOMENT, we correctly derive only expressions of a single R^n
// variable.
//
// HUGE KLUDGE : for efficiency matters, many interface wrappers of the Expr
// ADT (class) are ignored and data members of Exprs and read into and
// manipulated explicitly -- of course, respecting conventions that ensure
// integrity of Exprs.  Hence, the day the implementation of Expr is altered,
// this whole module of code has to be revised.
//
// Author : Benoit Hamelin
// 2003 Universit de Sherbrooke
//


function [derivator]=makeDerivator()
	derivator = tlist(SCIAD_DERIVATOR_FIELDS, ...
			list(), ...
			Derivator_VisitIVar, ...
			Derivator_VisitConst, ...
			Derivator_VisitAdd, ...
			Derivator_VisitAddInv, ...
			Derivator_VisitT, ...
			Derivator_VisitMult, ...
			Derivator_VisitElemMult, ...
			Derivator_VisitInv, ...
			Derivator_VisitElemInv, ...
			Derivator_VisitExp, ...
			Derivator_VisitElemExp, ...
			Derivator_VisitLog, ...
			Derivator_VisitMonome, ...
			Derivator_VisitElemMonome, ...
			Derivator_VisitSin, ...
			Derivator_VisitCos, ...
			Derivator_VisitAtan, ...
			Derivator_VisitAbs, ...
			Derivator_VisitStub, ...
			0, ...
			list(), ...
			[], ...
			%f ...
			)
endfunction


function [d]=Derivator_Process(this, expr, n, initialAdjoint, numericalOnly)
	// Processes the derivative calculation job on the given expression.
	// Derivative is calculated in each direction specified as column of
	// matrix D.
	nbNodes = Expr_GetNodeCount(expr)
	for i = 1:nbNodes do
		this.adjointList(i) = list()
	end  // for
	this.n = n
	this.ivarNodes = []
	this.numericalOnly = numericalOnly
	exprRootIndex = Expr_GetRoot(expr)
	this.adjointList(exprRootIndex) = initialAdjoint

	[expr,this] = Expr_VisitBackward(expr, this)

	// Compute derivative from adjoints gathered at independant variable
	// nodes.
	if numericalOnly then
		d = Derivator_MakeDerivNum(this, expr)
	else
		d = Derivator_MakeDerivGraph(this, expr)
	end  // if
endfunction


function [adjOperand]=Derivator_CalcElemAdj(operand, adjResult)
	// For element-wise primitives, computes the adjoint of some operand.
	//
	// operand    - Value combined to the adjoint at result node to obtain the
	//              operand's adjoint.
	// adjResult  - Adjoint derivative frame list for result node.
	//

	global SCIAD_EXPR SCIAD_VISITOR
	soperand = size(operand)
	if (and(soperand == [1,1])) then
		// Scalar operand.
		adjResult = adjResult(1)
		adjOperand = list(operand .* adjResult)
	elseif (soperand(1) == 1) then
		// Row vector operand.
		adjResult = adjResult(1)
		nbFrames = size(adjResult, 2)
		operand = operand' * ones(1, nbFrames)
		adjOperand = list(operand .* adjResult)
	elseif (soperand(2) == 1) then
		// Column vector operand.
		adjResult = adjResult(1)
		nbFrames = size(adjResult, 1)
		operand = ones(nbFrames, 1) * operand'
		adjOperand = list(operand .* adjResult)
	else
		// Matrix operand.
		nbFrames = size(adjResult)
		adjOperand = list()
		for i = 1:nbFrames do
			adjOperand(i) = operand .* adjResult(i)
		end  // for
	end  // if
endfunction


function [operand]=Derivator_GetOperand(nodeIndex, node)
	global SCIAD_EXPR SCIAD_VISITOR
	if SCIAD_VISITOR($).numericalOnly | ...
			(node.isConstant & type(node.value) == 1) then
		operand = node.value
	else
		if node.Visit == NodeConst_Visit | ...
				node.Visit == NodeIVar_Visit then
			operand = makeExpr(node)
		else
			operand = []
			if node.Visit == NodeT_Visit then
				indexPredec = SCIAD_VISITOR($).listPredec(nodeIndex)
				nodePredec = SCIAD_EXPR($).nodes(indexPredec)
				if nodePredec.Visit == NodeIVar_Visit then
					operand = makeExpr(nodePredec)'
				end  // if
			end  // if
			if typeof(operand) <> "ADEXPR" then
				operand = ad_stub(node.value, %t, nodeIndex)
			end  // if
		end  // if
	end  // if
endfunction


function [adjoint]=Derivator_GetAdjoint(nodeIndex)
	// Gives the adjoint frame list already computed for indexed node.
	//

	global SCIAD_EXPR SCIAD_VISITOR
	if SCIAD_VISITOR($).numericalOnly then
		adjoint = SCIAD_VISITOR($).adjointList(nodeIndex)
	else
		adjoint = list()
		for i = 1:size(SCIAD_VISITOR($).adjointList(nodeIndex)) do
			if Derivator_MustStubAdjoint(nodeIndex, i) then
				adjoint(i) = ad_stub(...
						SCIAD_VISITOR($).adjointList(nodeIndex)(i).nodes($).value, ...
						%f, ...
						[nodeIndex, i])
			else
				adjoint(i) = SCIAD_VISITOR($).adjointList(nodeIndex)(i)
			end  // if
		end  // for
	end  // if
endfunction


function [r]=Derivator_MustStubAdjoint(indexNode, indexFrame)
	// Tells whether indexed frame of adjoint for indexed node must be stubbed
	// or given as is.
	r = %t
	if typeof(SCIAD_VISITOR($).adjointList(indexNode)(indexFrame)) <> "ADEXPR" then
		r = %f
	elseif Expr_IsIVar(SCIAD_VISITOR($).adjointList(indexNode)(indexFrame)) then
		r = %f
	elseif Expr_IsConstant(SCIAD_VISITOR($).adjointList(indexNode)(indexFrame)) then
		r = %f
	end  // if
endfunction


function []=Derivator_AddAdjoint(nodeIndex, adjointTerm)
	// Adds a term to the adjoint sum for the indexed node.
	global SCIAD_EXPR SCIAD_VISITOR

	if size(SCIAD_VISITOR($).adjointList(nodeIndex)) == 0 then
		SCIAD_VISITOR($).adjointList(nodeIndex) = adjointTerm
	else
		for i = 1:size(SCIAD_VISITOR($).adjointList(nodeIndex)) do
			SCIAD_VISITOR($).adjointList(nodeIndex)(i) = ...
					SCIAD_VISITOR($).adjointList(nodeIndex)(i) + ...
					adjointTerm(i)
		end  // for
	end  // if
endfunction


function [index]=Derivator_VisitIVar(node, index)
	// Visit of an IVar node.
	//
	// We consider some arbitrary sub-vector of the full variable vector as a
	// projection of fvv into another euclidean space using an A matrix.  In
	// that case, we reconstitute partial gradient using the matrix*vector
	// adjoint calculation rule.

	global SCIAD_EXPR SCIAD_VISITOR
//	indices = node.index
//	sindices = size(indices, "*")
//	A = zeros(sindices, SCIAD_VISITOR($).n)
//	for i = 1:sindices do
//		A(i, indices(i)) = 1
//	end  // for
//
//	adjHere = SCIAD_VISITOR($).adjointList(index)
//	adjHere = adjHere(1)
//	SCIAD_VISITOR($).adjointList(index) = list(adjHere * A)

	SCIAD_VISITOR($).ivarNodes = [SCIAD_VISITOR($).ivarNodes, index]
endfunction


function [index]=Derivator_VisitConst(node, index)
	// Visit of a constant node.  Nothing to do here, should avoid to define
	// an adjoint for a constant in general.
	global SCIAD_EXPR SCIAD_VISITOR
endfunction


function [index]=Derivator_VisitAdd(node, index)
	// Visit of an addition node.
	//
	// z = x + y
	//     x, y, z : m x n
	// x* = z* ; y* = z*

	global SCIAD_EXPR SCIAD_VISITOR
	predec = SCIAD_VISITOR($).listPredec(index)
	if (size(predec, "*") <> 2) then
		error("Unexpected number of predecessors for addition node.")
	end  // if

	nodeLeft = SCIAD_EXPR($).nodes(predec(1))
	adjointList = Derivator_GetAdjoint(index)
	if (~nodeLeft.isConstant) then
		Derivator_AddAdjoint(predec(1), adjointList)
	end  // if

	nodeRight = SCIAD_EXPR($).nodes(predec(2))
	if (~nodeRight.isConstant) then
		Derivator_AddAdjoint(predec(2), adjointList)
	end  // if
endfunction


function [index]=Derivator_VisitAddInv(node, index)
	// Visit of an additive inverse node.
	//
	// z = -x
	//     x, z : m x n
	// x* = -z*

	global SCIAD_EXPR SCIAD_VISITOR
	predec = SCIAD_VISITOR($).listPredec(index)
	if (size(predec, "*") <> 1) then
		error("Unexpected number of predecessors for additive inverse node.")
	end  // if

	adjHere = Derivator_GetAdjoint(index)
	adjPredec = list()
	for i = 1:size(adjHere) do
		adjPredec(i) = -adjHere(i)
	end  // for
	Derivator_AddAdjoint(predec, adjPredec)
endfunction


function [index]=Derivator_VisitT(node, index)
	// Visit of a transposition node.
	//
	// z = x^T
	//     x : m x n
	//     z : n x m
	// x* = (z*)^T

	global SCIAD_EXPR SCIAD_VISITOR
	predec = SCIAD_VISITOR($).listPredec(index)
	if (size(predec, "*") <> 1) then
		error("Unexpected number of predecessors for transposition node.")
	end  // if

	adjHere = Derivator_GetAdjoint(index)
	adjPredec = list()
	for i = 1:size(adjHere) do
		adjPredec(i) = adjHere(i)'
	end  // for
	Derivator_AddAdjoint(predec, adjPredec)
endfunction


function [index]=Derivator_VisitMult(node, index)
	// Visit of a regular product node.
	//
	// z = x * y
	// 8 cases :
	//
	// a) Scalar * Scalar
	//    x : 1 x 1 ==> x* : p x 1
	//    y : 1 x 1 ==> y* : p x 1
	//    z : 1 x 1 ==> y* : p x 1
	//
	//    x* = y * z*
	//    y* = x * z*
	//
	// b) Vector * Scalar
	//    x : n x 1 ==> x* : p x n
	//    y : 1 x 1 ==> y* : p x 1
	//    z : n x 1 ==> z* : p x n
	//
	//    x* = y * z*
	//    y* = z* * x
	//
	// c) Scalar * Vector
	//    x : 1 x 1 ==> x* : p x 1
	//    y : 1 x m ==> y* : m x p
	//    z : 1 x m ==> z* : m x p
	//
	//    x* = z*^T * y^T
	//    y* = x * z*
	//
	// d) Inner product
	//    x : 1 x n ==> x* : n x p
	//    y : n x 1 ==> y* : p x n
	//    z : 1 x 1 ==> z* : p x 1
	//
	//    x* = y * z*^T
	//    y* = z* * x
	//
	// e) Outer product
	//    x : m x 1 ==> x* : p x m
	//    y : 1 x n ==> y* : n x p
	//    z : m x n ==> z* : m x n x p
	//
	//    x*(i,:) = y * z*(:,:,i)^T
	//    y*(:,i) = z*(:,:,i)^T * x
	//
	// f) Vector * Matrix
	//    x : 1 x m ==> x* : m x p
	//    y : m x n ==> y* : m x n x p
	//    z : 1 x n ==> z* : n x p
	//
	//    x* = y * z*
	//    y*(:,:,i) = x^T * z*(:,i)^T
	//
	// g) Matrix * Vector
	//    x : m x n ==> x* : m x n x p
	//    y : n x 1 ==> y* : p x n
	//    z : m x 1 ==> z* : p x m
	//
	//    x*(:,:,i) = z*(i,:)^T * y^T
	//    y* = z* * x
	//
	// h) Matrix * Matrix
	//    x : m x n ==> x* : m x n x p
	//    y : n x q ==> y* : n x q x p
	//    z : m x q ==> z* : m x q x p
	//
	//    x*(:,:,i) = z*(:,:,i) * y^T
	//    y*(:,:,i) = x^T * z*(:,:,i)

	global SCIAD_EXPR SCIAD_VISITOR
	predec = SCIAD_VISITOR($).listPredec(index)
	if (size(predec, "*") <> 2) then
		error("Unexpected number of predecessors for regular product node.")
	end  // if

	adjHere = Derivator_GetAdjoint(index)
	nodeLeft = SCIAD_EXPR($).nodes(predec(1))
	nodeRight = SCIAD_EXPR($).nodes(predec(2))
	opLeft = []
	if (~nodeRight.isConstant) then
		opLeft = Derivator_GetOperand(predec(1), nodeLeft)
	end  // if
	opRight = []
	if (~nodeLeft.isConstant) then
		opRight = Derivator_GetOperand(predec(2), nodeRight)
	end  // if

	sleft = size(nodeLeft)
	sright = size(nodeRight)
	if (and(sleft == [1,1])) then
		// Left operand is a scalar.  Right may either be another scalar or a
		// row vector.
		if (and(sright == [1,1])) then
			// Right is a scalar.
			if (~nodeLeft.isConstant) then
				Derivator_AddAdjoint(predec(1), list(adjHere(1) * opRight))
			end  // if
			if (~nodeRight.isConstant) then
				Derivator_AddAdjoint(predec(2), list(adjHere(1) * opLeft))
			end  // if
		else
			// Right operand is a row vector.
			if (~nodeLeft.isConstant) then
				Derivator_AddAdjoint(predec(1), list(adjHere(1)' * opRight'))
			end  // if
			if (~nodeRight.isConstant) then
				Derivator_AddAdjoint(predec(2), list(opLeft * adjHere(1)))
			end  // if
		end  // if
	elseif (sleft(1) == 1) then
		// Left operand is a row vector.  Right may either be a column vector
		// or a matrix.
		if (sright(2) == 1) then
			// Right is a column vector; we have an inner product.
			if (~nodeLeft.isConstant) then
				Derivator_AddAdjoint(predec(1), list(opRight * adjHere(1)'))
			end  // if
			if (~nodeRight.isConstant) then
				Derivator_AddAdjoint(predec(2), list(adjHere(1) * opLeft))
			end  // if
		else
			// Right operand is a matrix.
			adjHere = adjHere(1)
			sadj = size(adjHere)
			if (~nodeLeft.isConstant) then
				Derivator_AddAdjoint(predec(1), list(opRight * adjHere))
			end  // if
			if (~nodeRight.isConstant) then
				adjRight = list()
				opLeft = opLeft'
				for i = 1:sadj(2) do
					adjRight(i) = opLeft * adjHere(:,i)'
				end  // for
				Derivator_AddAdjoint(predec(2), adjRight)
			end  // if
		end  // if
	elseif (sleft(2) == 1) then
		// Left operand is a column vector.  Right may either be a scalar or a
		// row vector.
		if (and(sright == [1,1])) then
			// Right operand is a scalar.
			if (~nodeLeft.isConstant) then
				Derivator_AddAdjoint(predec(1), list(opRight * adjHere(1)))
			end  // if
			if (~nodeRight.isConstant) then
				Derivator_AddAdjoint(predec(2), list(adjHere(1) * opLeft))
			end  // if
		else
			// Right operand is a row vector; we have an outer product.
			nbfr = size(adjHere)
			if (~nodeLeft.isConstant) then
				adjLeft = []
				for i = 1:nbfr do
					adjLeft = [adjLeft ; opRight * adjHere(i)']
				end  // for
				Derivator_AddAdjoint(predec(1), list(adjLeft))
			end  // if
			if (~nodeRight.isConstant) then
				adjRight = []
				for i = 1:nbfr do
					adjRight = [adjRight , adjHere(i)' * opLeft]
				end  // for
				Derivator_AddAdjoint(predec(2), list(adjRight))
			end  // if
		end  // if
	else
		// Left operand is a matrix.  Right may either be a column vector or
		// another matrix.
		if (sright(2) == 1) then
			// Right operand is a column vector.
			adjHere = adjHere(1)
			sadj = size(adjHere)
			if (~nodeLeft.isConstant) then
				adjLeft = list()
				opRight = opRight'
				for i = 1:sadj(1) do
					adjLeft(i) = adjHere(i,:)' * opRight
				end  // for
				Derivator_AddAdjoint(predec(1), adjLeft)
			end  // if
			if (~nodeRight.isConstant) then
				Derivator_AddAdjoint(predec(2), list(adjHere * opLeft))
			end  // if
		else
			// Right operand is another matrix.
			nbfr = size(adjHere)
			if (~nodeLeft.isConstant) then
				adjLeft = list()
				for i = 1:nbfr do
					adjLeft(i) = adjHere(i) * opRight'
				end  // for
				Derivator_AddAdjoint(predec(1), adjLeft)
			end  // if
			if (~nodeRight.isConstant) then
				adjRight = list()
				for i = 1:nbfr do
					adjRight(i) = opLeft' * adjHere(i)
				end  // for
				Derivator_AddAdjoint(predec(2), adjRight)
			end  // if
		end  // if
	end  // if
endfunction


function [index]=Derivator_VisitElemMult(node, index)
	// Visit of an element-wise product node.
	//
	// z = x .* y
	// x* = y .* z*
	// y* = x .* z*
	//

	global SCIAD_EXPR SCIAD_VISITOR
	predec = SCIAD_VISITOR($).listPredec(index)
	if (size(predec, "*") <> 2) then
		error("Unexpected number of predecessors for element-wise product node.")
	end  // if

	nodeLeft = SCIAD_EXPR($).nodes(predec(1))
	nodeRight = SCIAD_EXPR($).nodes(predec(2))
	adjHere = Derivator_GetAdjoint(index)

	if (~nodeLeft.isConstant) then
		opRight = Derivator_GetOperand(predec(2), nodeRight)
		Derivator_AddAdjoint(predec(1), ...
				Derivator_CalcElemAdj(opRight, adjHere))
	end  // if
	if (~nodeRight.isConstant) then
		opLeft = Derivator_GetOperand(predec(1), nodeLeft)
		Derivator_AddAdjoint(predec(2), ...
				Derivator_CalcElemAdj(opLeft, adjHere))
	end  // if
endfunction


function [index]=Derivator_VisitInv(node, index)
	// Visit of an inversion node.
	// NOTE : for the moment, only derivation of a scalar inversion is
	// possible.
	//
	// z = 1 / x = x^(-1)
	// x* = -z* * x^(-2)
	//

	global SCIAD_EXPR SCIAD_VISITOR
	predec = SCIAD_VISITOR($).listPredec(index)
	if (size(predec, "*") <> 1) then
		error("Unexpected number of predecessors for inversion node.");
	end  // if

	invOpPredec = Derivator_GetOperand(index, node)
	adjHere = Derivator_GetAdjoint(index)
	adjHere = adjHere(1)
	Derivator_AddAdjoint(predec, list(-adjHere * (invOpPredec ^ 2)))
endfunction


function [index]=Derivator_VisitElemInv(node, index)
	// Visit of an element-wise inversion node.
	//
	// z = 1 ./ x = x .^ (-1)
	// x* = -z* .* (x .^ (-2))
	//

	global SCIAD_EXPR SCIAD_VISITOR
	predec = SCIAD_VISITOR($).listPredec(index)
	if (size(predec, "*") <> 1) then
		error("Unexpected number of predecessors for element-wise inversion node.")
	end  // if

	invOpPredec = Derivator_GetOperand(index, node)
	adjHere = Derivator_GetAdjoint(index)
	Derivator_AddAdjoint(predec, Derivator_CalcElemAdj(-invOpPredec .^ 2, adjHere))
endfunction


function [index]=Derivator_VisitExp(node, index)
	// Visit of an exponential node.
	// NOTE : for the moment, only derivative of scalar-base exponentials is
	// possible.
	//
	// z = a^x
	// x* = z* * a^x * log(a)

	global SCIAD_EXPR SCIAD_VISITOR
	predec = SCIAD_VISITOR($).listPredec(index)
	if (size(predec, "*") <> 1) then
		error("Unexpected number of predecessors for exponential node.")
	end  // if

	adjHere = Derivator_GetAdjoint(index)
	adjHere = adjHere(1)
	base = node.base
	nodePredec = SCIAD_EXPR($).nodes(predec)
	resultHere = Derivator_GetOperand(index, node)
	Derivator_AddAdjoint(predec, list(adjHere * resultHere * log(base)))
endfunction


function [index]=Derivator_VisitElemExp(node, index)
	// Visit of an element-wise exponential node.
	//
	// z = a .^ x
	// x* = z* .* (a .^ x) .* log(a)
	//

	global SCIAD_EXPR SCIAD_VISITOR
	predec = SCIAD_VISITOR($).listPredec(index)
	if (size(predec, "*") <> 1) then
		error("Unexpected number of predecessors for element-wise exponential node.")
	end  // if

	adjHere = Derivator_GetAdjoint(index)
	base = node.base
	resultHere = Derivator_GetOperand(index, node)
	igz = find(base > 0)
	logBase = zeros(base)
	logBase(igz) = log(base(igz))
	Derivator_AddAdjoint(predec, ...
			Derivator_CalcElemAdj(resultHere .* logBase, adjHere))
endfunction


function [index]=Derivator_VisitLog(node, index)
	// Visit of an element-wise logarithm node.
	//
	// z = log(x)
	// x* = z* ./ x
	//

	global SCIAD_EXPR SCIAD_VISITOR
	predec = SCIAD_VISITOR($).listPredec(index)
	if (size(predec, "*") <> 1) then
		error("Unexpected number of predecessors for element-wise logarithm node.")
	end  // if

	adjHere = Derivator_GetAdjoint(index)
	nodePredec = SCIAD_EXPR($).nodes(predec)
	opPredec = Derivator_GetOperand(predec, nodePredec)
	Derivator_AddAdjoint(predec, ...
			Derivator_CalcElemAdj(1 ./ opPredec, adjHere))
endfunction


function [index]=Derivator_VisitMonome(node, index)
	// Visit of a monome node.
	// NOTE : for the moment, only derivation of scalar monomes is possible.
	//
	// z = x^k
	// x* = k * z* * x^(k-1)
	//

	global SCIAD_EXPR SCIAD_VISITOR
	predec = SCIAD_VISITOR($).listPredec(index)
	if (size(predec, "*") <> 1) then
		error("Unexpected number of predecessors for monome node.")
	end  // if

	nodePredec = SCIAD_EXPR($).nodes(predec)
	opPredec = Derivator_GetOperand(predec, nodePredec)
	expo = node.exponent
	adjHere = Derivator_GetAdjoint(index)
	adjHere = adjHere(1)
	Derivator_AddAdjoint(predec, list(expo * adjHere * (opPredec ^ (expo-1))))
endfunction


function [index]=Derivator_VisitElemMonome(node, index)
	// Visit of an element-wise monome node.
	//
	// z = x .^ k
	// x* = k .* z* .* (x .^ (k-1))
	//

	global SCIAD_EXPR SCIAD_VISITOR
	predec = SCIAD_VISITOR($).listPredec(index)
	if (size(predec, "*") <> 1) then
		error("Unexpected number of predecessors for element-wise monome node.")
	end  // if

	nodePredec = SCIAD_EXPR($).nodes(predec)
	opPredec = Derivator_GetOperand(predec, nodePredec)
	expo = node.exponent
	adjHere = Derivator_GetAdjoint(index)
	Derivator_AddAdjoint(predec, ...
			Derivator_CalcElemAdj(expo .* (opPredec .^ (expo - 1)), adjHere))
endfunction


function [index]=Derivator_VisitSin(node, index)
	// Visit of an element-wise circular sine node.
	//
	// z = sin(x)
	// x* = z* .* cos(x)
	//

	global SCIAD_EXPR SCIAD_VISITOR
	predec = SCIAD_VISITOR($).listPredec(index)
	if (size(predec, "*") <> 1) then
		error("Unexpected number of predecessors for a sine node.")
	end  // if

	nodePredec = SCIAD_EXPR($).nodes(predec)
	opPredec = Derivator_GetOperand(predec, nodePredec)
	adjHere = Derivator_GetAdjoint(index)
	Derivator_AddAdjoint(predec, ...
			Derivator_CalcElemAdj(cos(opPredec), adjHere))
endfunction


function [index]=Derivator_VisitCos(node, index)
	// Visit of an element-wise circular cosine node.
	//
	// z = cos(x)
	// x* = -z* .* sin(x)
	//

	global SCIAD_EXPR SCIAD_VISITOR
	predec = SCIAD_VISITOR($).listPredec(index)
	if (size(predec, "*") <> 1) then
		error("Unexpected number of predecessors for a cosine node.")
	end  // if

	nodePredec = SCIAD_EXPR($).nodes(predec)
	opPredec = Derivator_GetOperand(predec, nodePredec)
	adjHere = Derivator_GetAdjoint(index)
	Derivator_AddAdjoint(predec, ...
			Derivator_CalcElemAdj(-sin(opPredec), adjHere))
endfunction


function [index]=Derivator_VisitAtan(node, index)
	// Visit of an element-wise circular arc-tangent node.
	//
	// z = arctan(x)
	// x* = z* ./ (1 + x .^ 2)
	//

	global SCIAD_EXPR SCIAD_VISITOR
	predec = SCIAD_VISITOR($).listPredec(index)
	if (size(predec, "*") <> 1) then
		error("Unexpected number of predecessors for an arctangent node.")
	end  // if

	nodePredec = SCIAD_EXPR($).nodes(predec)
	opPredec = Derivator_GetOperand(predec, nodePredec)
	adjHere = Derivator_GetAdjoint(index)
	Derivator_AddAdjoint(predec, ...
			Derivator_CalcElemAdj(1 ./ (1 + opPredec .^ 2), adjHere))
endfunction


function [index]=Derivator_VisitAbs(node, index)
	// Visit of an absolute value node.
	//
	// x(i) < 0 ==> x*(i) = -z*(i)
	// x(i) > 0 ==> x*(i) = z*(i)
	// x(i) = 0 ==> x*(i) = 0
	//

	global SCIAD_EXPR SCIAD_VISITOR
	predec = SCIAD_VISITOR($).listPredec(index)
	if (size(predec, "*") <> 1) then
		error("Unexpected number of predecessors for an absolute value node.")
	end  // if

	nodePredec = SCIAD_EXPR($).nodes(predec)
	valuePredec = nodePredec.value
	factor = zeros(valuePredec)
	isz = find(valuePredec < 0)
	factor(isz) = -ones(factor(isz))
	igz = find(valuePredec > 0)
	factor(igz) = ones(factor(igz))
	adjHere = Derivator_GetAdjoint(index)
	Derivator_AddAdjoint(predec, Derivator_CalcElemAdj(factor, adjHere))
endfunction


function [index]=Derivator_VisitStub(node, index)
	error("Derivator : unable to work with sub-expression stubs.")
endfunction


function [deriv]=Derivator_MakeDerivNum(this, expr)
	// After the backward traversal of the expression graph, this method
	// constructs the numerical expression of the derivative of the function.

	exprRootIndex = expr.root
	initialAdjoint = this.adjointList(exprRootIndex)
	indic = "r"
	if (size(expr, 'r') == 1) then
		indic = "c"
	end  // if
	deriv = zeros(size(initialAdjoint(1), indic), this.n)
	if (this.ivarNodes <> []) then
		for n = this.ivarNodes($:-1:1) do
			ivarNode = expr.nodes(n)
			derivPart = this.adjointList(n)(1)
			for i = 1:length(ivarNode.index) do
				deriv(:, ivarNode.index(i)) = deriv(:, ivarNode.index(i)) + ...
						derivPart(:, i)
			end  // for
		end  // for
	end  // if
endfunction


function [deriv]=Derivator_MakeDerivGraph(this, expr)
	// After the backward traversal of the expression graph, this method
	// constructs the graph of the derivative of the function.
	global SCIAD_RGD_ADJ_LIST SCIAD_RGD_ORIG_EXPR SCIAD_RGD_LIST_PREDEC;
	global SCIAD_RGD_DICT_NODE SCIAD_RGD_DICT_ADJ;
	global SCIAD_RGD_DERIV;

	indic = "r"
	if (size(expr, 'r') == 1) then
		indic = "c"
	end  // if
	exprRootIndex = Expr_GetRoot(expr)
	initialAdjoint = this.adjointList(exprRootIndex)(1)
	if (this.ivarNodes == []) then
		deriv = ad_const(zeros(size(initialAdjoint, indic), this.n))
	else

		SCIAD_RGD_ADJ_LIST = this.adjointList
		SCIAD_RGD_ORIG_EXPR = expr
		SCIAD_RGD_LIST_PREDEC = Expr_ListPredecessors(expr)
		nbNodes = Expr_GetNodeCount(expr)
		SCIAD_RGD_DICT_NODE = zeros(nbNodes, 1)
		SCIAD_RGD_DICT_ADJ = zeros(nbNodes, size(initialAdjoint, indic))

		// Pre-compile constant adjoints together.
		ivarAdjoints = this.ivarNodes($:-1:1)
		nonConstAdjoints = []
		initialDeriv = zeros(size(initialAdjoint, indic), this.n)
		for i = ivarAdjoints do
			if typeof(SCIAD_RGD_ADJ_LIST(i)(1)) <> "ADEXPR" then
				nodeIVar = expr.nodes(i)
				adjValue = SCIAD_RGD_ADJ_LIST(i)(1)
				for j = 1:length(nodeIVar.index) do
					initialDeriv(:, nodeIVar.index(j)) = ...
							initialDeriv(:, nodeIVar.index(j)) + adjValue(:,j)
				end  // for
			elseif Expr_IsConstant(SCIAD_RGD_ADJ_LIST(i)(1)) then
				nodeIVar = expr.nodes(i)
				adjValue = Expr_GetEvaluation(SCIAD_RGD_ADJ_LIST(i)(1))
				for j = 1:length(nodeIVar.index) do
					initialDeriv(:, nodeIVar.index(j)) = ...
							initialDeriv(:, nodeIVar.index(j)) + adjValue(:,j)
				end  // for
			else
				nonConstAdjoints($+1) = i
			end  // for
		end  // for
		nonConstAdjoints = nonConstAdjoints(:)'
		SCIAD_RGD_DERIV = ad_const(initialDeriv)

		// Incorporate other (variable) adjoints.
		indexLeft = 1
		nodeLeft = SCIAD_RGD_DERIV.nodes(1)
		for i = nonConstAdjoints do
			nodeIVar = expr.nodes(i)

			// Include adjoint in derivative.
			Derivator_RGD_InclAdjoint(i, 1)
			indexLastAdjoint = SCIAD_RGD_DICT_ADJ(i, 1)
			nodeLastAdjoint = SCIAD_RGD_DERIV.nodes(indexLastAdjoint)
			indexRight = indexLastAdjoint

			// Compose a product to the adjoint so it is effectively projected
			// into the Jacobian matrix space.
			[projMatrix,projAdjoint] = Derivator_RGD_IVarProjection(this.n, ...
					nodeIVar, nodeLastAdjoint.value)
			sproj = size(projMatrix)
			if (sproj(1) <> sproj(2)) | or(projMatrix <> eye(projMatrix)) then
				SCIAD_RGD_DERIV.nodes($+1) = makeNodeConst(projMatrix)
				indexNodeProj = size(SCIAD_RGD_DERIV.nodes)
				SCIAD_RGD_DERIV.nodes($+1) = makeNodeMult(projAdjoint)
				indexRight = size(SCIAD_RGD_DERIV.nodes)
				SCIAD_RGD_DERIV.edges(:, [$+1,$+2]) = ...
					[...
						indexLastAdjoint, indexNodeProj ; ...
						indexRight      , indexRight   ...
					]
			end  // if

			// Cumulate the last adjoint.
			valueAdd = nodeLeft.value
			for j = 1:length(nodeIVar.index) do
				valueAdd(:, nodeIVar.index(j)) = ...
						valueAdd(:, nodeIVar.index(j)) + ...
						nodeLastAdjoint.value(:, j)
			end  // for
			nodeAdd = makeNodeAdd(valueAdd)
			SCIAD_RGD_DERIV.nodes($+1) = nodeAdd
			indexNodeAdd = size(SCIAD_RGD_DERIV.nodes)
			SCIAD_RGD_DERIV.edges(:,$+1) = [indexLeft; indexNodeAdd]
			SCIAD_RGD_DERIV.edges(:,$+1) = [indexRight; indexNodeAdd]
			indexLeft = indexNodeAdd
			nodeLeft = nodeAdd
		end  // for
		SCIAD_RGD_DERIV.root = size(SCIAD_RGD_DERIV.nodes)

		deriv = SCIAD_RGD_DERIV
	end  // if
endfunction


function [projMatrix,projAdjoint]=Derivator_RGD_IVarProjection(n, nodeIVar, ...
		adjointValue)
	// Builds a projection matrix that expresses how some independant variable
	// sub-vector has been obtained from the original.  Also gives the result
	// of the projection of the adjoint value using this new matrix.
	//
	indices = nodeIVar.index
	sindices = size(indices, "*")
	projMatrix = zeros(sindices, n)
	for i = 1:sindices do
		projMatrix(i, indices(i)) = 1
	end  // for

	projAdjoint = zeros(size(adjointValue, 'r'), n)
	for i = 1:length(nodeIVar.index) do
		projAdjoint(:, nodeIVar.index(i)) = ...
				projAdjoint(:, nodeIVar.index(i)) + adjointValue(:, i)
	end  // for
endfunction


function []=Derivator_RGD_InclAdjoint(indexAdj, indexFrame)
	// Recursively includes the given adjoint derivative to the derivative
	// expression being reconstructed.
	//
	// NOTE : this does NOT set the root of the derivative expression.
	//

	global SCIAD_RGD_ADJ_LIST SCIAD_RGD_ORIG_EXPR SCIAD_RGD_LIST_PREDEC;
	global SCIAD_RGD_DICT_NODE SCIAD_RGD_DICT_ADJ;
	global SCIAD_RGD_DERIV;

	if SCIAD_RGD_DICT_ADJ(indexAdj, indexFrame) <> 0 then
		error("RGD : we don''t include adjoints that are already there!")
	end  // if

	stackAdjoint = list(list(indexAdj, indexFrame, 1, []))
	while size(stackAdjoint) > 0 do
		// Acquire state of adjoint to include.
		indexAdj = stackAdjoint($)(1)
		indexFrame = stackAdjoint($)(2)
		adjoint = SCIAD_RGD_ADJ_LIST(indexAdj)(indexFrame)
		if (typeof(adjoint) <> "ADEXPR") then
			adjoint = ad_const(adjoint)
		end  // if
		nbNodes = size(adjoint.nodes)
		indexNodeAt = stackAdjoint($)(3)
		dict = stackAdjoint($)(4)
		if (dict == []) then
			dict = zeros(nbNodes, 1)
		end  // if

		// Add nodes of the adjoint.
		for indexNode = indexNodeAt:nbNodes do
			node = adjoint.nodes(indexNode)
			if node.Visit == NodeStub_Visit then
				// Stub node.  Do we need to substitute it?
				if node.subexpr then
					// Sub-expression stub.  Is it in?
					if SCIAD_RGD_DICT_NODE(node.index) == 0 then
						Derivator_RGD_InclSubExpr(node.index)
					end  // if
					dict(indexNode) = SCIAD_RGD_DICT_NODE(node.index)
				else
					// Adjoint stub.  Is it in?
					if SCIAD_RGD_DICT_ADJ(node.index(1), ...
							node.index(2)) == 0 then
						// No!  We push the inclusion of this adjoint.
						stackAdjoint($)(3) = indexNode
						stackAdjoint($)(4) = dict
						stackAdjoint($+1) = ...
								list(node.index(1), node.index(2), 1, [])
						break  // Return to outer loop.
					end  // if
					dict(indexNode) = SCIAD_RGD_DICT_ADJ(node.index(1), ...
							node.index(2))
				end  // if
			else
				// Any other node.  We put it in.
				SCIAD_RGD_DERIV.nodes($+1) = node
				dict(indexNode) = size(SCIAD_RGD_DERIV.nodes)
			end  // if
		end  // for

		// Pursue only if all nodes have been added!
		if dict($) <> 0 then
			// Add edges.
			for edge = Expr_GetEdges(adjoint) do
				SCIAD_RGD_DERIV.edges(:, $+1) = ...
						[dict(edge(1)) ; dict(edge(2))]
			end  // for

			// Add entry in adjoint dictionary.
			SCIAD_RGD_DICT_ADJ(indexAdj, indexFrame) = dict($)

			// Pop adjoint from stack.
			stackAdjoint($) = null()
		end  // if
	end  // while
endfunction


function []=Derivator_RGD_InclSubExpr(indexRoot)
	// Includes the sub-expression rooted at the indexed node into the
	// reconstructed derivative.
	//
	// NOTE : the does NOT set the root of the reconstructed derivative.

	global SCIAD_RGD_ADJ_LIST SCIAD_RGD_ORIG_EXPR SCIAD_RGD_LIST_PREDEC;
	global SCIAD_RGD_DICT_NODE SCIAD_RGD_DICT_ADJ;
	global SCIAD_RGD_DERIV;

	if SCIAD_RGD_DICT_NODE(indexRoot) <> 0 then
		error("RGD : we don''t include sub-expressions that are already there!")
	end  // if

	stackNode = list(-indexRoot)
	while size(stackNode) > 0 do
		indexNode = stackNode($)
		stackNode($) = null()
		if indexNode < 0 then
			// Top node is being visited for the first time...
			indexNode = -indexNode
			if (SCIAD_RGD_DICT_NODE(indexNode) == 0) then
				// This node isn't already in the reconstructed derivative.
				// Put its predecessors first, then visit it over.
				stackNode($+1) = indexNode
				predec = SCIAD_RGD_LIST_PREDEC(indexNode)
				for p = predec($:-1:1) do
					stackNode($+1) = -p
				end  // for
			end  // if
		else
			// Top node is visited for the second time around.  This means
			// its predecessors have been inserted in the reconstructed
			// derivative.  Let us insert the node, then edges from its
			// predecessors.
			SCIAD_RGD_DERIV.nodes($+1) = SCIAD_RGD_ORIG_EXPR.nodes(indexNode)
			SCIAD_RGD_DICT_NODE(indexNode) = size(SCIAD_RGD_DERIV.nodes)
			predec = SCIAD_RGD_LIST_PREDEC(indexNode)
			for p = predec do
				SCIAD_RGD_DERIV.edges(:, $+1) = ...
					[SCIAD_RGD_DICT_NODE(p); SCIAD_RGD_DICT_NODE(indexNode)]
			end  // for
		end  // if
	end  // while
endfunction


// End of module.
