//
// Functions to implement analytical expressions.
//
// Author : Benoit Hamelin
// 2002 Universit de Sherbrooke
//


function [expr]=makeExpr(node)
	// Instantiates a 1-node expression.
	s = size(node)
	expr = mlist(SCIAD_EXPR_FIELDS, s(:)', list(node), [0;0], 1);
			// Edges have dummy first entry to initialize shape.
endfunction


function [expr]=extendExpr(expr, node)
	// Appends a node at the end of expr.  Dimension considerations are ignored.
	root = Expr_GetRoot(expr)
	expr.nodes($+1) = node
	expr.edges(:, $+1) = [root ; root + 1]
	expr.dims = size(node)
	expr.root = Expr_GetNodeCount(expr)
	expr.nodes(expr.root).isConstant = Expr_IsConstant(expr)
endfunction


function [expr]=composeExpr(expr, joined, node)
	// Joins together two expressions by connecting them both to the same
	// ending node.  Dimension considerations are ignored.
	nbNodesExpr = Expr_GetNodeCount(expr)
	rootExpr = Expr_GetRoot(expr)
	joinedEdgesCount = size(joined.edges, "c")
	expr.edges = [expr.edges , ...
		joined.edges(:, 2:$) + nbNodesExpr .* ones(2, joinedEdgesCount - 1)]
	expr.nodes = lstcat(expr.nodes, joined.nodes)
	rootJoined = Expr_GetRoot(joined) + nbNodesExpr
	expr.nodes($+1) = node
	newRoot = Expr_GetNodeCount(expr)
	expr.edges(:, ($+1):($+2)) = [rootExpr , rootJoined ; newRoot , newRoot]
	expr.dims = size(node)
	expr.root = newRoot
	expr.nodes(newRoot).isConstant = ...
			(Expr_IsConstant(expr) & Expr_IsConstant(joined))
endfunction


// Special expression matrices.

function [ivar]=ad_ivar(value)
	// Creates an independant variable R^n vector.
	ivar = makeExpr(makeNodeIVar((1:size(value, "*"))', value(:)))
endfunction


function [const]=ad_const(value)
	// Creates an expression around a constant hypermatrix.
	const = makeExpr(makeNodeConst(value))
endfunction


function [stub]=ad_stub(value, subexpr, index)
	// Creates the stub of some sub-expression rooted at node indexed "root", and that has to be sub-indexed using given indexList.
	stub = makeExpr(makeNodeStub(value, subexpr, index))
endfunction


// Access to an expression.

function varargout=%ADEXPR_size(expr, desc)
	// Size operator.
	[lhs,rhs] = argn()

	s = expr.dims
	if (rhs == 2) then
		if (type(desc) == 10) then  // Descriptor is a string.
			if (desc == "*") then
				varargout(1) = prod(s)
			elseif (desc == "r") then
				varargout(1) = s(1)
			elseif (desc == "c") then
				varargout(1) = s(2)
			else
				error("Bad size information description.")
			end  // if
		else  // Descriptor is a number.
			varargout(1) = s(desc)
		end  // if
	else
		if (lhs == 1) then
			varargout(1) = s
		else
			if (lhs <> size(s, "*")) then
				error("Not as many output parameters as dimensions of expression matrix.")
			end  // if
			for i=(1:lhs) do
				varargout(i) = s(i)
			end  // for
		end  // if
	end  // if
endfunction


function [v]=Expr_GetEvaluation(this)
	// Gives result of expression evaluation.
	lastNode = this.nodes($)
	v = lastNode.value
endfunction


function [const?]=Expr_IsConstant(this)
	// Tells whether given expression is a simple constant.
	rootNode = Expr_GetNode(this, Expr_GetRoot(this))
	const? = rootNode.isConstant
endfunction


function [isIVar,isVertical]=Expr_IsIVar(this, varargin)
	// Tells whether given expression is an independent variable; in that
	// case, also tells whether it's a column or row vector (transposed
	// column).
	//
	// If other arguments are given, they are taken to be the indexes of nodes
	// that compose the sub-expression for which we want to know whether it's
	// an independent variable or not.

	if (size(varargin) == 0) then
		nbNodes = Expr_GetNodeCount(this)
		indexNodes = 1:nbNodes
	else
		nbNodes = size(varargin)
		if (nbNodes == 1) then
			indexNodes = varargin(1)
		else
			indexNodes = [varargin(1),varargin(2)]
		end  // if
	end  // if

	node1 = Expr_GetNode(this, indexNodes(1))
	if (nbNodes == 1) then
		// Only one node.  Is it a column independant variable?
		isVertical = %t
		isIVar = (node1.Visit == NodeIVar_Visit)
	elseif (nbNodes == 2) then
		// Two nodes.  Is it a row independant variable?
		node2 = Expr_GetNode(this, indexNodes(2))
		isIVar = (node1.Visit == NodeIVar_Visit & node2.Visit == NodeT_Visit)
		isVertical = ~isIVar
	else
		isIVar = %f
		isVertical = %f  // Indefinite.
	end  // if
endfunction


function [this]=Expr_Reevaluate(this, newIVarValue)
	// Re-evaluates the given expression graph by taking given new valuation
	// for independant variables.
	this = Expr_VisitForward(this, makeEvaluator(newIVarValue))
endfunction


// Access to the graph.

function [nc]=Expr_GetNodeCount(this)
	// Gives number of nodes the graph of this expression is made of.
	nc = size(this.nodes)
endfunction


function [node]=Expr_GetNode(this, i)
	// Gives indexed node of the graph of this expression.
	node = this.nodes(i)
endfunction


function [root]=Expr_GetRoot(this)
	// Gives root node of the graph of this expression.
	root = this.root
endfunction


function [succ]=Expr_GetSuccessors(this, i)
	// Gives indices of successors of the indexed node within the expression's graph.
	succ = this.edges(2, find(this.edges(1,:) == i))
endfunction


function [predec]=Expr_GetPredecessors(this, i)
	// Gives indices of predecessors of the indexed node within the expression's graph.
	predec = this.edges(1, find(this.edges(2,:) == i))
endfunction


function [predecList]=Expr_ListPredecessors(this)
	// Computes a list for which each entry contains the row vector of
	// predecessors of the node that indexes the entry.
	predecList = list()
	for i=1:size(this.edges, 'c') do
		predecList(i) = []
	end  // for
	for edge = Expr_GetEdges(this) do
		predecList(edge(2)) = [predecList(edge(2)), edge(1)]
	end  // for
endfunction


function [edges]=Expr_GetEdges(this)
	// Gives edge array (clean from representation idiosynchrasies).
	edges = this.edges(:, 2:$)
endfunction


function [subexpr]=Expr_ExtractSubExpr(this, i)
	// Extracts sub-expression rooted at node i.  Partially extracted built sub-expression is given, as its current number of nodes.
	//
	if (i <= 0 | i > Expr_GetNodeCount(this)) then
		error("Invalid node index for sub-expression extraction.")
	end  // if

	subexpr = mlist(SCIAD_EXPR_FIELDS, [], list(), [0;0], 0);
	nbNodesSubExpr = 0
	listTodo = list(-i)
	while size(listTodo) > 0 do
		nodesAtLevel = listTodo($)
		[workNode,iwork] = Expr_FindWorkNode(nodesAtLevel)

		if workNode == 0 then
			// All nodes at this level have been processed.  Link them to
			// their root, which is the work node at lower level.  Their
			// successor will be this lower-level work node, which must be
			// added.
			if (size(listTodo) > 1) then
				[lowerNode,ilower] = Expr_FindWorkNode(listTodo($-1))
				if (lowerNode == 0) then
					error("Successor-to-be should be a workable node in the lower level.")
				end  // if
				nbNodesSubExpr = nbNodesSubExpr + 1
				subexpr.nodes($+1) = Expr_GetNode(this, lowerNode)
				successor = nbNodesSubExpr
				for predecessor = nodesAtLevel do
					subexpr.edges = [subexpr.edges, [predecessor ; successor]]
				end  // for

				// Node at lower level has been fully processed.
				nodesAtLowerLevel = listTodo($-1)
				nodesAtLowerLevel(ilower) = successor
				listTodo($-1) = nodesAtLowerLevel
				//pause
			end  // if
			
			// Level all processed : pop the level stack.
			listTodo($) = null()
		else
			// Must process next node at this level.  Check if it has
			// predecessors.
			predec = Expr_GetPredecessors(this, workNode)
			if (predec == []) then
				// No predecessor.  We add this node to the sub-expression and
				// translate it at this level.
				nbNodesSubExpr = nbNodesSubExpr + 1
				newNode = nbNodesSubExpr
				subexpr.nodes($+1) = Expr_GetNode(this, workNode)
				nodesAtLevel(iwork) = newNode
				listTodo($) = nodesAtLevel
				//pause
			else
				// Got predecessors.  We push these on the stack and process
				// them first.  Current node remains unaltered on the stack so
				// it can be further processed after its children.
				listTodo($+1) = -predec
				//pause
			end  // if
		end  // if
	end  // while
	lastNodeSubExpr = Expr_GetNode(subexpr, nbNodesSubExpr)
	subexpr.dims = size(lastNodeSubExpr)
	subexpr.root = Expr_GetNodeCount(subexpr)
endfunction


function [workNode,i]=Expr_FindWorkNode(nodes)
	// Private to ExtractSubExpr : finds the work node for the given level.
	workNode = 0
	for i = 1:size(nodes,"*") do
		if (nodes(i) < 0) then
			workNode = -nodes(i)
			break
		end  // if
	end  // for
endfunction


// NOTE : this function does not work on 20040722.
//
//function [this]=Expr_ReplaceLeaf(this, index, subexpr)
//	// Replaces the indexed leaf node with a new sub-expression.
//	// WARNING : the fact that "index" points to a leaf node is not explicitly checked.
//
//	if (or(size(this.nodes(index)) <> size(subexpr))) then
//		//error("Trying to replace a leaf node with a sub-expression that has not the same size.")
//	end  // if
//
//	// Gather things from the sub-expression that will be inserted herein.
//	subexprNodes = subexpr.nodes
//	nbNodesSub = Expr_GetNodeCount(subexpr)
//	subexprEdges = Expr_GetEdges(subexpr)
//	if (subexprEdges <> []) then
//		subexprEdges = subexprEdges + index - 1
//	end  // if
//
//	// Divide the list of nodes in two : those before the replaced leaf and those after -- which will remain in this expression's
//	// node list.
//	nodesBefore = list()
//	i = 1
//	while (i < index) do
//		nodesBefore = lstcat(nodesBefore, list(this.nodes(1)))
//		this.nodes(1) = null()
//		i = i + 1
//	end  // while
//	this.nodes(1) = null()  // Eliminate the replaced leaf.
//	this.nodes = lstcat(nodesBefore, subexprNodes, this.nodes)
//
//	// Modify the edge matrix : all nodes indexed greater than or equal to the replaced leaf index must be offset.
//	indexGreaterLeaf = find(this.edges >= index)
//	this.edges(indexGreaterLeaf) = this.edges(indexGreaterLeaf) + nbNodesSub - 1
//
//	// Insert sub-expression edges just before the first edge that involves the replaced leaf.
//	iedge1 = min(find(this.edges(1,:) == index + nbNodesSub - 1))
//	if (iedge1 == []) then
//		this.edges = [this.edges, subexprEdges]
//	else
//		this.edges = [this.edges(:,1:(iedge1-1)), subexprEdges, this.edges(:,iedge1:$)]
//	end  // if
//endfunction


// Visiting.

function [this,visitor]=Expr_Visit(this, visitor, nodeIndices)
	//
	// Gets a visitor to traverse this expression's graph in order given through the node list.
	//
	// Note that the visit algorithm doesn't follow the expression expr paths directly.  Hence, all node visit methods must be
	// implemented out of node traversal order context.  The correct traversal context for some node should be gathered from this
	// expression passed as argument.
	//
	// For the moment, alteration of the expression graph during traversal is forbidden.

	global SCIAD_VISITOR SCIAD_EXPR
	SCIAD_VISITOR($+1) = visitor
	SCIAD_EXPR($+1) = this

	SCIAD_VISITOR($).listPredec = Expr_ListPredecessors(this)
	for node = nodeIndices do
		[dummy] = ExprNode_Visit(this.nodes(node), node)
	end;  // for

	visitor = SCIAD_VISITOR($)
	SCIAD_VISITOR($) = null()
	this = SCIAD_EXPR($)
	SCIAD_EXPR($) = null()
endfunction


function [this,visitor]=Expr_VisitForward(this, visitor)
	//
	// Gets a visitor to traverse this expression from top to bottom
	// (following edge sense).
	//
	// Note that the visit algorithm doesn't follow the expression expr paths
	// directly.  Hence, all node visit methods must be implemented out of
	// node traversal order context.  The correct traversal context for some
	// node should be gathered from this expression passed as argument.
	//
	// For the moment, alteration of the structure of the expression graph
	// during traversal is forbidden; one can modify the contents of the
	// nodes, however.
	//
	[this,visitor] = Expr_Visit(this, visitor, 1:Expr_GetNodeCount(this))
endfunction


function [this,visitor]=Expr_VisitBackward(this, visitor)
	// Gets a visitor to traverse this expression from bottom to top (inverse
	// of edge sense).
	//
	// Note that the visit algorithm doesn't follow the expression expr paths
	// directly.  Hence, all node visit methods must be implemented out of
	// node traversal order context.  The correct traversal context for some
	// node should be gathered from this expression passed as argument.
	//
	// For the moment, alteration of the expression graph during traversal is
	// forbidden.
	//
	[this,visitor] = Expr_Visit(this, visitor, Expr_GetNodeCount(this):-1:1)
endfunction


function [transpose]=%ADEXPR_t(expr)
	// Calculates and returns the transposition of the given expression.

	sexpr = size(expr)
	if (Expr_IsConstant(expr)) then
		transpose = ad_const(Expr_GetEvaluation(expr)')
	elseif and(sexpr == [1,1]) then
		transpose = expr
	else
		rootIndex = Expr_GetRoot(expr)
		rootNode = Expr_GetNode(expr, rootIndex)
		if (rootNode.Visit == NodeT_Visit) then
			// Root node is a transposition : we cancel this operation.
			transpose = expr
			predec = Expr_GetPredecessors(transpose, rootIndex)
			trIndex = rootIndex
			transpose.root = predec
			transpose.nodes(trIndex) = null()
			transpose.edges(:, find(transpose.edges(2,:) == trIndex)) = []
			i = find(transpose.edges(:) > trIndex)
			if i <> [] then
				transpose.edges(i) = transpose.edges(i) - 1
			end  // if
			transpose.dims = size(Expr_GetNode(transpose, predec))
		else
			transpose = extendExpr(expr, makeNodeT(Expr_GetEvaluation(expr)'))
		end  // if
	end  // if
endfunction


// Extraction.

function [extr]=%ADEXPR_e(varargin)
	// Extraction operator.  Two types of extractions are allowed :
	// uni-dimensional index and n-dimensional index, where n corresponds to
	// the number of dimensions of this hypermatrix.  Out-of-bounds indices
	// pop an error.

	[lhs,rhs] = argn()
	if (rhs < 2) then
		error("No index given.")
	end  // if

	this = varargin($)
	varargin($) = null()

	sVecList = size(varargin)
	sthis = size(this)
	if Expr_IsConstant(this) then

		// We build a new constant expression with the extracted value.  We
		// use the "raw" index vectors.
		value = Expr_GetEvaluation(this)
		if (sVecList == 1) then
			if (and(size(this) <> 1)) then
				error("Cannot use a single index on a matricial SCIAD expression.")
			end  // if
			value = value(varargin(1))
		elseif (sVecList == 2) then
			value = value(varargin(1), varargin(2))
		else
			error("Illegal are expressions that express on matrices of more than 2 dimensions.")
		end  // if
		extr = ad_const(value)

	else

		// Concretize given index vectors.
		index = []
		if (sVecList == 1) then
			if (and(size(this) <> 1)) then
				error("Cannot use a single index on a matricial SCIAD expression.")
			end  // if
			[index,out?] = concretizeIndexVector(varargin(1), prod(sthis))
			if (out?) then
				error("Given index vector refers out of bounds of expression vector.")
			end  // if
			indexVecList = list(index(:)')
		elseif (sVecList == size(sthis, "*")) then
			indexVecList = list()
			for i = 1:sVecList do
				[indexVec,out?] = concretizeIndexVector(varargin(i), sthis(i))
				if (out?) then
					error("Index vector #" + i + " refers out of bounds of expression matrix.")
				end  // if
				indexVecList(i) = indexVec(:)'
			end  // for
		else
			error("Too many or too few index arguments given.")
		end  // if

		// Two cases we manage here :
		// a) Expr is an independant variable (original or transposed) : we make
		//    a new IVar expression with the appropriate indexing, and transpose
		//    it if appropriate.
		// b) Expr is anything else : we just use a matricial product to select
		//    indexed rows and columns.
		[thisIsIVar,isVertical] = Expr_IsIVar(this)
		node1 = Expr_GetNode(this, 1)
		selectFullVector = %f
		if sVecList == 1 & type(varargin(1)) == 1 then
			selectFullVector = (size(varargin(1)) == [-1,-1])
		end  // if

		if thisIsIVar then
			// We compute a new IVar expression with the selected indexes.
			if (sVecList == 1) then
				extr = makeExpr(...
						makeNodeIVar(...
							node1.index(indexVecList(1)), ...
							node1.value(indexVecList(1)) ...
							) ...
						)
				if (~isVertical & ~selectFullVector) then
					extr = extr'
				end  // if
			else
				iIndexVec = 1
				if (~isVertical) then
					iIndexVec = 2
				end  // if
				extr = makeExpr(...
						makeNodeIVar(...
							node1.index(indexVecList(iIndexVec)), ...
							node1.value(indexVecList(iIndexVec)) ...
							) ...
						)
				if (~isVertical) then
					extr = extr'
				end  // if
			end  // if
		else
			// We compute row and column selection matrices that we'll multiply to
			// this expression.
			thisIsRow = (sthis(1) == 1)
			if sVecList == 1 then
				// Convert it into a two-index list.
				if thisIsRow then
					indexVecList(2) = indexVecList(1)
					indexVecList(1) = 1
				else
					indexVecList(2) = 1
				end  // if
			end  // if

			// Row selection matrix.
			indexVec = indexVecList(1)
			nbIndexes = size(indexVec, "*")
			rowSel = zeros(nbIndexes, sthis(1))
			for r = 1:nbIndexes do
				rowSel(r, indexVec(r)) = 1
			end  // for

			// Column selection matrix.
			indexVec = indexVecList(2)
			nbIndexes = size(indexVec, "*")
			colSel = zeros(sthis(2), nbIndexes)
			for c = 1:nbIndexes do
				colSel(indexVec(c), c) = 1
			end  // for

			// Extract!  Transpose the result for selection of a whole row
			// vector.
			extr = rowSel * this * colSel
			if (thisIsRow & selectFullVector) then
				extr = extr'
			end  // if
		end  // if

	end  // if
endfunction


// Concatenation.

function [conc]=%ADEXPR_c_ADEXPR(left, right)
	// Matrix horizontal concatenation operator.  We concatenate using product and addition.
	sleft = size(left)
	sright = size(right)
	if ((size(sleft,"*") > 2) | (size(sright,"*") > 2)) then
		error("Can only concatenate 2-dimensional objects.")
	elseif (and(sleft == zeros(sleft))) then
		conc = right
	elseif (and(sright == zeros(sright))) then
		conc = left
	elseif (sleft(1) <> sright(1)) then
		error("Can only concatenate horizontally two objects that share their number of rows.")
	else
		left_col = sleft(2)
		right_col = sright(2)
		conc = ...
			left * [eye(left_col,left_col) , zeros(left_col,right_col)] + ...
			right * [zeros(right_col,left_col) , eye(right_col,right_col)]
	end  // if
endfunction


function [conc]=%ADEXPR_c_s(left, right)
	conc = [left,ad_const(right)]
endfunction


function [conc]=%s_c_ADEXPR(left, right)
	conc = [ad_const(left),right]
endfunction


function [conc]=%ADEXPR_f_ADEXPR(top, bottom)
	// Matrix vertical concatenation operator.  We concatenate using product and addition.
	stop = size(top)
	sbottom = size(bottom)
	if ((size(stop,"*") > 2) | (size(sbottom,"*") > 2)) then
		error("Can only concatenate 2-dimensional objects.")
	elseif (and(stop == zeros(stop))) then
		conc = bottom
	elseif (and(sbottom == zeros(sbottom))) then
		conc = top
	elseif (stop(2) <> sbottom(2)) then
		error("Can only concatenate vertically two objects that share their number of columns.")
	else
		top_row = stop(1)
		bottom_row = sbottom(1)
		conc = ...
			[eye(top_row,top_row) ; zeros(bottom_row,top_row)] * top + ...
			[zeros(top_row,bottom_row) ; eye(bottom_row,bottom_row)] * bottom
	end  // if
endfunction


function [conc]=%ADEXPR_f_s(top, bottom)
	conc = [top ; ad_const(bottom)]
endfunction


function [conc]=%s_f_ADEXPR(top, bottom)
	conc = [ad_const(top) ; bottom]
endfunction


// Insertion.

function this=%ADEXPR_i_ADEXPR(varargin)
	// Insertion operator.  Two types of insertions are allowed : unidimensional index and bidimensional index, where n
	// corresponds to the number of dimensions of this matrix.  For the moment, we can only insert in a bidimensional matrix.
	// The inserted object must have same dimensions as the indexed sub-matrix that is replaced.

	[lhs,rhs] = argn()
	if (rhs < 3) then
		error("No index given.")
	end  // if

	this = varargin($)
	ins = varargin($-1)

	indexCount = size(varargin) - 2
	if (indexCount > 2) then
		error("No more than two index vectors are supported.")
	elseif (indexCount == 1) then
		if (and(size(this) <> 1)) then
			error("Single-index insertion not supported in SCIAD expression matrices -- only in vectors.")
		end  // if
		this = Insert1Index(this, ins, varargin(1))
	else
		this = Insert2Index(this, ins, varargin(1), varargin(2))
	end  // if
endfunction


function this=%s_i_ADEXPR(varargin)
	// Insertion operator (constant in expression).  Two types of insertions are allowed : unidimensional index and bidimensional
	// index, where n corresponds to the number of dimensions of this matrix.  For the moment, we can only insert in a
	// bidimensional matrix.  The inserted object must have same dimensions as the indexed sub-matrix that is replaced.

	[lhs,rhs] = argn()
	if (rhs < 3) then
		error("No index given.")
	end  // if

	this = varargin($)
	ins = ad_const(varargin($-1))

	indexCount = size(varargin) - 2
	if (indexCount > 2) then
		error("No more than two index vectors are supported.")
	elseif (indexCount == 1) then
		if (and(size(this) <> 1)) then
			error("Single-index insertion not supported in SCIAD expression matrices -- only in vectors.")
		end  // if
		this = Insert1Index(this, ins, varargin(1))
	else
		this = Insert2Index(this, ins, varargin(1), varargin(2))
	end  // if
endfunction


function this=%ADEXPR_i_s(varargin)
	// Insertion operator (expression in constant).  Two types of insertions are allowed : unidimensional index and bidimensional
	// index, where n corresponds to the number of dimensions of this matrix.  For the moment, we can only insert in a
	// bidimensional matrix.  The inserted object must have same dimensions as the indexed sub-matrix that is replaced.

	[lhs,rhs] = argn()
	if (rhs < 3) then
		error("No index given.")
	end  // if

	dest = varargin($)
	if (size(dest, "*") == 0) then
		this = ad_const(0)
	else
		this = ad_const(varargin($))
	end  // if
	ins = varargin($-1)

	indexCount = size(varargin) - 2
	if (indexCount > 2) then
		error("No more than two index vectors are supported.")
	elseif (indexCount == 1) then
		if (and(size(this) <> 1)) then
			error("Single-index insertion not supported in SCIAD expression matrices -- only in vectors.")
		end  // if
		this = Insert1Index(this, ins, varargin(1))
	else
		this = Insert2Index(this, ins, varargin(1), varargin(2))
	end  // if
endfunction


// Finite difference calculus.

function [Delta,n,t]=Expr_CalcFiniteDiff(this, increment, numzero)
	// Calculates right finite difference on expression for given increment
	// and required numerical precision.  Also gives the number of nodes for
	// which AFD involved computing a convergent series (n), as well as the
	// total number of convergent series terms that were computed (t).
	increment = increment(:)
	fdcalc = makeFiniteDiffCalc(numzero)
	[Delta,n,t] = FDCalc_Process(fdcalc, this, increment)
endfunction


// Derivative calculus.

function [d]=Expr_Gradient(this, n, numericalOnly)
	// Calculates the gradient of a R^n->R function using the backward mode of automatic differentiation.
	// FOR THE MOMENT, allows only derivation of expressions of a single R^n variable.
	//
	// n - Dimension of the independant variable vector.
	// numericalOnly - Boolean flag.  If true, only a numerical evaluation of the gradient of the given expression will be
	//                 computed; otherwise, the expression graph of the gradient will be built, allowing for hessian computation
	//                 using the the Expr_Jacobian function.

	d = Derivator_Process(makeDerivator(), this, n, list(1), numericalOnly)
endfunction


function [d]=Expr_Jacobian(this, n, initialAdjoint, numericalOnly)
	// Calculates the jacobian of a R^n->R^m function using the backward mode of automatic differentiation.
	// FOR THE MOMENT, allows only derivation of expressions of a single R^n variable.
	//
	// n - Dimension of the independant variable vector.
	// initialAdjoint - Matrix of initial adjoint directions; must be a p x m matrix when image is a column vector and a m x p
	//                  matrix when image is a row vector (a gradient, for example).  For instance, giving [1 0 ... 0] will yield
	//                  the first line of the jacobian; giving I_m will yield the full jacobian.
	// numericalOnly - Boolean flag.  If true, only a numerical evaluation of the jacobian of the given expression will be
	//                 computed; otherwise, the expression graph of the jacobian will be built, allowing for computation of
	//                 higher-order directional derivatives of the function.

	d = Derivator_Process(makeDerivator(), this, n, list(initialAdjoint), numericalOnly)
endfunction


function [taylorCoeff]=Expr_TaylorCoeff(this, order, anchor)
	// Using this expression as f(x), calculates "order+1" coefficients of the
	// Taylor development of f(x-anchor).  Only works with x in R^1.

	[this,taylorVisitor] = Expr_VisitForward(this, makeTaylorDev(order, anchor))
	taylorCoeff = taylorVisitor.coeffList($)
endfunction


// End of module.

