//
// Functions regarding nodes that compose an expression graph.
//
// Author : Benoit Hamelin
// 2002 Universite de Sherbrooke
//


// Common to all nodes.

function [n]=ExprNode_GetDisplay(node)
	// Returns the display string representation of the given expression node.
	n = node.display;
endfunction


function [index]=ExprNode_Visit(this, index)
	// Visits the given node in context of a graph traversal.  Index of this visited node in context of the expression it belongs
	// to is given, as well as the expression itself (for gathering of visit context) and the indexes of predecessors and
	// successors.
	visit = this("Visit")
	[index] = visit(this, index)
endfunction


function []=%ADNODE_p(node)
	mprintf("%s\n", node.display)
endfunction


function varargout=%ADNODE_size(varargin)
	// Size operator.
	[lhs,rhs] = argn()
	node = varargin(1)
	if (lhs > 1) then
		[l,c] = size(node.value)
		varargout(1) = l
		varargout(2) = c
	elseif (rhs > 1) then
		varargout(1) = size(node.value, varargin(2))
	else
		s = size(node.value)
		varargout(1) = s(:)'
	end  // if
endfunction


// Independant variable nodes.

function [node]=makeNodeIVar(index, value)
	// Makes a new independant variable node, with given index reference
	// vector and evaluation value.
	index = index(:)
	value = value(:)
	n = ""
	if (size(index, "*") == 1) then
		n = "Indep. variable (" + string(index) + ")"
	else
		n = "Indep. variable [" + strcat(string(index), ",") + "]"
	end  // if
	node = mlist(SCIAD_NODEIVAR_FIELDS, n, value, %f, NodeIVar_Visit, index)
endfunction


function [index]=NodeIVar_Visit(this, index)
	visitIVar = visitor.VisitIVar
	[index] = visitIVar(this, index)
endfunction


// Constant node.

function [node]=makeNodeConst(value)
	// Makes a new constant node.
	s = size(value)
	node = mlist(SCIAD_NODECONST_FIELDS, "Constant", value, %t, NodeConst_Visit);
endfunction


function [index]=NodeConst_Visit(this, index)
	visitConst = visitor.VisitConst
	[index] = visitConst(this, index)
endfunction


// Addition node.

function [node]=makeNodeAdd(value)
	// Makes a new addition node.
	node = mlist(SCIAD_NODEADD_FIELDS, "+", value, %f, NodeAdd_Visit)
endfunction


function [index]=NodeAdd_Visit(this, index)
	visitAdd = visitor.VisitAdd
	[index] = visitAdd(this, index)
endfunction


// Additive inverse node.

function [node]=makeNodeAddInv(value)
	node = mlist(SCIAD_NODEADDINV_FIELDS, "- (unary)", value, %f, NodeAddInv_Visit)
endfunction


function [index]=NodeAddInv_Visit(this, index)
	visitAddInv = visitor.VisitAddInv
	[index] = visitAddInv(this, index)
endfunction


// Transposition node.

function [node]=makeNodeT(value)
	// Makes a new transposition node.
	node = mlist(SCIAD_NODET_FIELDS, "T", value, %f, NodeT_Visit)
endfunction


function [index]=NodeT_Visit(this, index)
	visitT = visitor.VisitT
	[index] = visitT(this, index)
endfunction


// Regular product node.

function [node]=makeNodeMult(value)
	// Makes a new regular product node.
	node = mlist(SCIAD_NODEMULT_FIELDS, "*", value, %f, NodeMult_Visit)
endfunction


function [index]=NodeMult_Visit(this, index)
	visitMult = visitor.VisitMult
	[index] = visitMult(this, index)
endfunction


// Element-wise product node.

function [node]=makeNodeElemMult(value)
	// Makes a new element-wise product node.
	node = mlist(SCIAD_NODEELEMMULT_FIELDS, ".*", value, %f, NodeElemMult_Visit)
endfunction


function [index]=NodeElemMult_Visit(this, index)
	visitElemMult = visitor.VisitElemMult
	[index] = visitElemMult(this, index)
endfunction


// Regular inverse (multiplicative) node.

function [node]=makeNodeInv(value)
	// Makes a new regular inverse node.
	node = mlist(SCIAD_NODEINV_FIELDS, "inv", value, %f, NodeInv_Visit)
endfunction


function [index]=NodeInv_Visit(this, index)
	visitInv = visitor.VisitInv
	[index] = visitInv(this, index)
endfunction


// Element-wise inverse (multiplicative) node.

function [node]=makeNodeElemInv(value)
	// Makes a new element-wise inverse node.
	node = mlist(SCIAD_NODEELEMINV_FIELDS, "e-inv", value, %f, NodeElemInv_Visit)
endfunction


function [index]=NodeElemInv_Visit(this, index)
	visitElemInv = visitor.VisitElemInv
	[index] = visitElemInv(this, index)
endfunction


// Regular exponential node.

function [node]=makeNodeExp(base, value)
	// Makes a new regular exponential node.
	node = mlist(SCIAD_NODEEXP_FIELDS, "exp", value, %f, NodeExp_Visit, base)
endfunction

function [index]=NodeExp_Visit(this, index)
	visitExp = visitor.VisitExp
	[index] = visitExp(this, index)
endfunction


// Element-wise exponential node.

function [node]=makeNodeElemExp(base, value)
	// Makes a new element-wise exponential node.
	node = mlist(SCIAD_NODEELEMEXP_FIELDS, ".exp", value, %f, NodeElemExp_Visit, base)
endfunction

function [index]=NodeElemExp_Visit(this, index)
	visitElemExp = visitor.VisitElemExp
	[index] = visitElemExp(this, index)
endfunction


// Logarithm node.

function [node]=makeNodeLog(value)
	// Makes a new element-wise logarithm node.
	node = mlist(SCIAD_NODELOG_FIELDS, "log", value, %f, NodeLog_Visit);
endfunction

function [index]=NodeLog_Visit(this, index)
	visitLog = visitor.VisitLog
	[index] = visitLog(this, index)
endfunction


// Regular monome node.

function [node]=makeNodeMonome(base, exponent)
	// Makes a new regular monome node.  The trace of the calculus is kept, so to be used when evaluating finite differences.
	[value,listPowers,listIndex,listProds,rootExpo,rootFactor] = ...
			elevatePower(base, exponent)
	node = mlist(SCIAD_NODEMONOME_FIELDS, ...
			"monome", ...
			value, ...
			%f, ...
			NodeMonome_Visit, ...
			exponent, ...
			listPowers, ...
			listIndex, ...
			listProds, ...
			rootExpo, ...
			rootFactor)
endfunction


function [index]=NodeMonome_Visit(this, index)
	visitMonome = visitor.VisitMonome
	[index] = visitMonome(this, index)
endfunction


// Element-wise monome node.

function [node]=makeNodeElemMonome(base, exponent)
	[value,listPowers,listMasks,listProds,rootExpo,rootFactor] = ...
			elevatePowerElem(base, exponent)
	node = mlist(SCIAD_NODEELEMMONODE_FIELDS, ...
			".monome", ...
			value, ...
			%f, ...
			NodeElemMonome_Visit, ...
			exponent, ...
			listPowers, ...
			listMasks, ...
			listProds, ...
			rootExpo, ...
			rootFactor)
endfunction

function [index]=NodeElemMonome_Visit(this, index)
	visitElemMonome = visitor.VisitElemMonome
	[index] = visitElemMonome(this, index)
endfunction


// Element-wise circular sine node.

function [node]=makeNodeSin(value)
	// Makes a new sine node.
	node = mlist(SCIAD_NODESIN_FIELDS, "sine", value, %f, NodeSin_Visit)
endfunction

function [index]=NodeSin_Visit(this, index)
	visitSin = visitor.VisitSin
	[index] = visitSin(this, index)
endfunction


// Element-wise circular cosine node.

function [node]=makeNodeCos(value)
	// Makes a new cosine node.
	node = mlist(SCIAD_NODECOS_FIELDS, "cosine", value, %f, NodeCos_Visit)
endfunction

function [index]=NodeCos_Visit(this, index)
	visitCos = visitor.VisitCos
	[index] = visitCos(this, index)
endfunction


// Element-wise circular arc-tangent node.

function [node]=makeNodeAtan(value)
	// Makes a new arc-tangent node.
	node = mlist(SCIAD_NODEATAN_FIELDS, "arc-tangent", value, %f, NodeAtan_Visit)
endfunction

function [index]=NodeAtan_Visit(this, index)
	visitAtan = visitor.VisitAtan
	[index] = visitAtan(this, index)
endfunction


// Absolute value node.

function [node]=makeNodeAbs(value)
	// Makes a new absolute value node.
	node = mlist(SCIAD_NODEABS_FIELDS, "absolute value", value, %f, NodeAbs_Visit)
endfunction

function [index]=NodeAbs_Visit(this, index)
	visitAbs = visitor.VisitAbs
	[index] = visitAbs(this, index)
endfunction


// Sub-expression stub node.

function [node]=makeNodeStub(value, subexpr, index)
	// Makes a new sub-expression stub node.
	//
	if subexpr then
		display = "Sub-expr rooted at " + string(index)
	else
		display = "Adjoint of node " + string(index(1)) + ...
				", frame " + string(index(2))
	end  // if
	node = mlist(SCIAD_NODESTUB_FIELDS, display, value, %f, ...
			NodeStub_Visit, subexpr, index)
endfunction

function [index]=NodeStub_Visit(this, index)
	visitStub = visitor.VisitStub
	[index] = visitStub(this, index)
endfunction


// End of function module.

