//
// Operators and overloaded functions to build Sciad expressions in regular Scilab syntax.
//
// Author : Benoit Hamelin
// 2002 Universit de Sherbrooke
//


function [r]=%ADEXPR_a(expr)
	// Unary plus - no alteration to given expression.
	r = expr
endfunction


function [r]=%ADEXPR_a_ADEXPR(left, right)
	// Addition operator.

	sleft = size(left)
	sright = size(right)
	if (or(sleft <> sright)) then

		if (prod(sleft) == 1) then
			r = (ones(sright(1), 1) * left * ones(1, sright(2))) + right
		elseif (prod(sright) == 1) then
			r = left + (ones(sleft(1), 1) * right * ones(1, sleft(2)))
		else
			error("Can''t add two hypermatrices with discrepant sizes.")
		end  // if

	else

		leftIsConst = Expr_IsConstant(left)
		rightIsConst = Expr_IsConstant(right)
		leftValue = Expr_GetEvaluation(left)
		leftValueVec = leftValue(:)
		rightValue = Expr_GetEvaluation(right)
		rightValueVec = rightValue(:)
		z = zeros(leftValueVec)
		r = []

		if (leftIsConst & rightIsConst) then
			r = ad_const(leftValue + rightValue)
		elseif (leftIsConst) then
			if (and(leftValueVec == z)) then
				r = right
			end  // if
		elseif (rightIsConst) then
			if (and(rightValueVec == z)) then
				r = left
			end  // if
		end  // if

		if (typeof(r) <> "ADEXPR") then
			r = composeExpr(left, right, makeNodeAdd(leftValue + rightValue))
		end  // if
	
	end  // if
endfunction


function [r]=%ADEXPR_a_s(left, right)
	r = left + ad_const(right)
endfunction


function [r]=%ADEXPR_a_hm(left, right)
	r = left + ad_const(right)
endfunction


function [r]=%s_a_ADEXPR(left, right)
	r = ad_const(left) + right
endfunction


function [r]=%hm_a_ADEXPR(left, right)
	r = ad_const(left) + right
endfunction


function [r]=%ADEXPR_sum(varargin)
	// Sum operator.

	[lhs,rhs] = argn()
	if (rhs < 1) then
		error("Takes at least one argument.");
	end  // if
	expr = varargin(1)
	op = 0
	if (rhs > 1) then
		arg2 = varargin(2)
		if (arg2 == 1 | arg2 == "r") then
			op = 1
		elseif (arg2 == 2 | arg2 == "c") then
			op = 2
		else
			error("Invalid second argument; must be 1, 2, ''r'' or ''c''.");
		end  // if
	end  // if

	[nbRows,nbCols] = size(expr)
	if (op == 0) then
		// Must add together all elements of the vectorized expression.
		if (nbRows == 1) then
			r = expr * ones(nbCols, 1)
		elseif (nbCols == 1) then
			r = ones(1, nbRows) * expr
		else
			r = ones(1, nbRows) * expr * ones(nbCols, 1)
		end  // if
	elseif (op == 1) then
		// Must add together all rows of the expression.
		r = ones(1, nbRows) * expr
	elseif (op == 2) then
		// Must add together all columns of the expression.
		r = expr * ones(nbCols, 1)
	else
		error("Bad programming of primitive -- email benoit@benoithamelin.com with insult.")
	end  // if
endfunction


function [r]=%ADEXPR_x_ADEXPR(left, right)
	// Element-wise product operator.

	sleft = size(left)
	sright = size(right)
	if (or(sleft <> sright)) then

		if (prod(sleft) == 1) then
			r = (ones(sright(1), 1) * left * ones(1, sright(2))) .* right
		elseif (prod(sright) == 1) then
			r = left .* (ones(sleft(1), 1) * right * ones(1, sleft(2)))
		else
			error("Can''t dot-multiply two hypermatrices with discrepant sizes.")
		end  // if

	else

		leftIsConst = Expr_IsConstant(left)
		leftValue = Expr_GetEvaluation(left)
		leftValueVec = leftValue(:)
		rightIsConst = Expr_IsConstant(right)
		rightValue = Expr_GetEvaluation(right)
		rightValueVec = rightValue(:)
		prodValues = leftValue .* rightValue
		z = zeros(leftValueVec)
		o = ones(leftValueVec)
		r = []

		if (leftIsConst & rightIsConst) then
			r = ad_const(prodValues)
		elseif (leftIsConst) then
			if (and(leftValueVec == z)) then
				r = ad_const(zeros(leftValue))
			elseif (and(leftValueVec == o)) then
				r = right
			end  // if
		elseif (rightIsConst) then
			if (and(rightValueVec == z)) then
				r = ad_const(zeros(rightValue))
			elseif (and(rightValueVec == o)) then
				r = left
			end  // if
		end  // if

		if (typeof(r) <> "ADEXPR") then
			r = composeExpr(left, right, makeNodeElemMult(prodValues))
		end  // if

	end  // if
endfunction


function [pr]=%ADEXPR_x_s(left, right)
	pr = left .* ad_const(right)
endfunction


function [pr]=%ADEXPR_x_hm(left, right)
	pr = left .* ad_const(right)
endfunction


function [pr]=%s_x_ADEXPR(left, right)
	pr = ad_const(left) .* right
endfunction


function [pr]=%hm_x_ADEXPR(left, right)
	pr = ad_const(left) .* right
endfunction


function [r]=%ADEXPR_m_ADEXPR(left, right)
	// Regular (matricial inner) product operator.

	sleft = size(left)
	sright = size(right)
	if ((size(sleft,"*") > 2) | (size(sright,"*") > 2)) then
		// TODO - Incorporate rules for product of n-dimensional objects.
		error("Inner product can only be computed on 2-dimensional objects.")
	end  // if

	if ((size(sleft,"*") == 2) & (size(sright,"*") == 2) & (sleft(2) == sright(1))) then

		leftIsConst = Expr_IsConstant(left)
		leftValue = Expr_GetEvaluation(left)
		rightIsConst = Expr_IsConstant(right)
		rightValue = Expr_GetEvaluation(right)
		prodValues = leftValue * rightValue
		r = []

		if (leftIsConst & rightIsConst) then
			r = ad_const(prodValues)
		elseif (leftIsConst) then
			if (and(leftValue == zeros(leftValue))) then
				r = ad_const(zeros(prodValues))
			elseif ((sleft(1) == sleft(2)) & and(leftValue == eye(leftValue))) then
				r = right
			end  // if
		elseif (rightIsConst) then
			if (and(rightValue == zeros(rightValue))) then
				r = ad_const(zeros(prodValues))
			elseif ((sright(1) == sright(2)) & and(rightValue == eye(rightValue))) then
				r = left
			end  // if
		end  // if

		if (typeof(r) <> "ADEXPR") then
			r = composeExpr(left, right, makeNodeMult(prodValues))
		end  // if

	elseif ((prod(sleft) == 0) | (prod(sright) == 0)) then
		// Product with an empty matrix!
		r = []
	elseif (prod(sleft) == 1) then
		// Product of a scalar object to a vector/matrix.  We convert to an element-wise product.
		r = (ones(sright(1), 1) * left * ones(1, sright(2))) .* right
	elseif (prod(sright) == 1) then
		// Product of a vector/matrix to a scalar object.  We convert to an element-wise product.
		r = left .* (ones(sleft(1), 1) * right * ones(1, sleft(2)))
	else
		// Incoherent product.
		error("Number of columns of left matrix doesn''t match number of lines of right matrix.")
	end  // if
endfunction


function [r]=%ADEXPR_m_s(left, right)
	r = left * ad_const(right)
endfunction


function [r]=%s_m_ADEXPR(left, right)
	r = ad_const(left) * right
endfunction


function [r]=%ADEXPR_s(expr)
	// Additive inversion of expression.
	r = extendExpr(expr, makeNodeAddInv(-Expr_GetEvaluation(expr)))
endfunction


function [r]=%ADEXPR_s_ADEXPR(left, right)
	// Subtraction of an expression hypermatrix to another.
	r = left + (-right)
endfunction


function [r]=%ADEXPR_s_s(left, right)
	r = left + (-right)
endfunction


function [r]=%s_s_ADEXPR(left, right)
	r = left + (-right)
endfunction


function [r]=%ADEXPR_s_hm(left, right)
	r = left + (-right)
endfunction


function [r]=%hm_s_ADEXPR(left, right)
	r = left + (-right)
endfunction


function [r]=%ADEXPR_inv(expr)
	// Regular (matricial) inversion of an expression hypermatrix.
	sexpr = size(expr)
	if ((size(sexpr,"*") <> 2) | (sexpr(1) <> sexpr(2))) then
		error("Can only invert square 2-dim matrices.");
	end  // if

	if (Expr_IsConstant(expr)) then
		r = ad_const(inv(Expr_GetEvaluation(expr)))
	else
		r = extendExpr(expr, makeNodeInv(inv(Expr_GetEvaluation(expr))))
	end  // if
endfunction


function [r]=%ADEXPR_r_ADEXPR(left, right)
	// Regular (matricial) right-division of two expression matrices.
	// TODO - Support pseudo-inversion of non square matrices.
	r = left * inv(right)
endfunction


function [r]=%ADEXPR_r_s(left, right)
	r = left * inv(right)
endfunction


function [r]=%s_r_ADEXPR(left, right)
	r = left * inv(right)
endfunction


function [r]=%ADEXPR_l_ADEXPR(left, right)
	// Regular (matricial) left-division of two expression matrices.
	// TODO - Support pseudo-inversion of non square matrices.
	r = inv(left) * right
endfunction


function [r]=%ADEXPR_l_s(left, right)
	r = inv(left) * right
endfunction


function [r]=%s_l_ADEXPR(left, right)
	r = inv(left) * right
endfunction


function [r]=eleminv(expr)
	// Element-wise multiplicative inverse of the given expression hypermatrix.
	if (Expr_IsConstant(expr)) then
		r = ad_const(1 ./ Expr_GetEvaluation(expr))
	else
		r = extendExpr(expr, makeNodeElemInv(1 ./ Expr_GetEvaluation(expr)))
	end  // if
endfunction


function [r]=%ADEXPR_d_ADEXPR(left, right)
	// Element-wise right-division of two expression hypermatrices.
	r = left .* eleminv(right)
endfunction


function [r]=%ADEXPR_d_s(left, right)
	r = left .* (1 ./ right)
endfunction


function [r]=%ADEXPR_d_hm(left, right)
	r = left .* (ones(right) ./ right)
endfunction


function [r]=%s_d_ADEXPR(left, right)
	r = left .* eleminv(right)
endfunction


function [r]=%hm_d_ADEXPR(left, right)
	r = left .* eleminv(right)
endfunction


function [r]=%ADEXPR_q_ADEXPR(left, right)
	// Element-wise left-division of two expression hypermatrices.
	r = eleminv(left) .* right
endfunction


function [r]=%ADEXPR_q_s(left, right)
	r = eleminv(left) .* right
endfunction


function [r]=%ADEXPR_q_hm(left, right)
	r = eleminv(left) .* right
endfunction


function [r]=%s_q_ADEXPR(left, right)
	r = (1 ./ left) .* right
endfunction


function [r]=%hm_q_ADEXPR(left, right)
	r = (ones(left) ./ left) .* right
endfunction


function [r]=%ADEXPR_p_ADEXPR(base, exponent)
	// Regular exponentiation.

	sbase = size(base)
	ssbase = size(sbase)
	sexponent = size(exponent)
	if (or(ssbase <> [1,2]) | (sbase(1) <> sbase(2)) | or(sexponent <> [1,1])) then
		// This is not a typical case of regular exponentiation, we defer it to element-wise exponentiation.
		r = base .^ exponent
	else

		// Base is a square matrix and exponent is a scalar expression.
		baseIsConst = Expr_IsConstant(base)
		baseValue = Expr_GetEvaluation(base)
		exponentIsConst = Expr_IsConstant(exponent)
		exponentValue = Expr_GetEvaluation(exponent)
		r = []

		if (baseIsConst & exponentIsConst) then
			r = ad_const(baseValue ^ exponentValue)
		elseif (baseIsConst) then
			if (and(baseValue == zeros(baseValue))) then
				r = ad_const(zeros(baseValue))
			elseif (and(baseValue == eye(baseValue))) then
				r = ad_const(eye(baseValue))
			else
				r = extendExpr(exponent, makeNodeExp(baseValue, baseValue ^ exponentValue))
			end  // if
		elseif (exponentIsConst) then
			if (exponentValue == 0) then
				r = ad_const(eye(baseValue))
			elseif (exponentValue == 1) then
				r = base
			elseif (exponentValue < 0) then
				r = inv(base) ^ (-exponentValue)
			else
				r = extendExpr(base, makeNodeMonome(baseValue, exponentValue))
			end  // if
		else
			error("Exponentiation : can only handle constant-base exponential or constant-exponent polynome.")
		end  // if
	end  // if
endfunction


function [r]=%ADEXPR_p_s(base, exponent)
	r = base ^ ad_const(exponent)
endfunction


function [r]=%s_p_ADEXPR(base, exponent)
	r = ad_const(base) ^ exponent
endfunction


function [r]=%ADEXPR_j_ADEXPR(base, exponent)
	// Element-wise exponentiation.

	sbase = size(base)
	sexponent = size(exponent)
	if (or(sbase <> sexponent)) then

		if (prod(sbase) == 1) then
			r = (ones(sexponent(1), 1) * base * ones(1, sexponent(2))) .^ exponent
		elseif (prod(sexponent) == 1) then
			r = base .^ (ones(sbase(1), 1) * exponent * ones(1, sbase(2)))
		else
			error("Can''t dot-power two hypermatrices with discrepant sizes.")
		end  // if

	else

		baseIsConst = Expr_IsConstant(base)
		baseValue = Expr_GetEvaluation(base)
		baseValueVec = baseValue(:)
		exponentIsConst = Expr_IsConstant(exponent)
		exponentValue = Expr_GetEvaluation(exponent)
		exponentValueVec = exponentValue(:)
		r = []
		z = zeros(baseValueVec)
		o = ones(baseValueVec)

		if (baseIsConst & exponentIsConst) then
			r = ad_const(baseValue .^ exponentValue)
		elseif (baseIsConst) then
			if (and(baseValueVec == z)) then
				r = ad_const(zeros(baseValue))
			elseif (and(baseValueVec == o)) then
				r = ad_const(ones(baseValue))
			else
				r = extendExpr(exponent, makeNodeElemExp(baseValue, baseValue .^ exponentValue))
			end  // if
		elseif (exponentIsConst) then
			if (and(exponentValueVec == z)) then
				r = ad_const(ones(baseValue))
			elseif (and(exponentValueVec == o)) then
				r = base
			elseif (and(exponentValueVec == -o)) then
				r = eleminv(base)
			else
				r = extendExpr(base, makeNodeElemMonome(baseValue, exponentValue))
			end  // if
		else
			error("Exponentiation : can only handle constant-base exponential or constant-exponent polynome.")
		end  // if
	end  // if
endfunction


function [r]=%ADEXPR_j_s(base, exponent)
	r = base .^ ad_const(exponent)
endfunction


function [r]=%ADEXPR_j_hm(base, exponent)
	r = base .^ ad_const(exponent)
endfunction


function [r]=%s_j_ADEXPR(base, exponent)
	r = ad_const(base) .^ exponent
endfunction


function [r]=%hm_j_ADEXPR(base, exponent)
	r = ad_const(Base) .^ exponent
endfunction


function [r]=%ADEXPR_exp(expr)
	r = %e .^ expr
endfunction


function [r]=%ADEXPR_log(expr)
	// Element-wise logarithm of an expression hypermatrix.
	r = extendExpr(expr, makeNodeLog(log(Expr_GetEvaluation(expr))))
endfunction


function [r]=%ADEXPR_sqrt(expr)
	// Element-wise square root of an expression hypermatrix.
	r = expr .^ (0.5)
endfunction


function [r]=%ADEXPR_o_s(left, right)
	// Value equality comparison (with constant matrix at right).
	r = (Expr_GetEvaluation(left) == right)
endfunction


function [r]=%s_o_ADEXPR(left, right)
	// Value equality comparison (with constant matrix at left).
	r = (right == left)
endfunction


function [r]=%ADEXPR_o_hm(left, right)
	// Value equality comparison (with hypermatrix at right).
	r = (Expr_GetEvaluation(left) == right)
endfunction


function [r]=%hm_o_ADEXPR(left, right)
	// Value equality comparison (with hypermatrix at left).
	r = (right == left)
endfunction


function [r]=%ADEXPR_n_s(left, right)
	// Value difference comparison (with constant matrix at right).
	r = (Expr_GetEvaluation(left) <> right)
endfunction


function [r]=%s_n_ADEXPR(left, right)
	// Value difference comparison (with constant matrix at left).
	r = (right <> left)
endfunction


function [r]=%ADEXPR_n_hm(left, right)
	// Value difference comparison (with hypermatrix at right).
	r = (Expr_GetEvaluation(left) <> right)
endfunction


function [r]=%hm_n_ADEXPR(left, right)
	// Value difference comparison (with hypermatrix at left).
	r = (right <> left)
endfunction


function [r]=%ADEXPR_1_s(left, right)
	// Value total order comparison (constant matrix at right).
	r = (Expr_GetEvaluation(left) < right)
endfunction


function [r]=%s_1_ADEXPR(left, right)
	// Value total order comparison (constant matrix at left).
	r = (right > left)
endfunction


function [r]=%ADEXPR_1_hm(left, right)
	// Value total order comparison (hypermatrix at right).
	r = (Expr_GetEvaluation(left) < right)
endfunction


function [r]=%hm_1_ADEXPR(left, right)
	// Value total order comparison (hypermatrix at left).
	r = (right > left)
endfunction


function [r]=%ADEXPR_2_s(left, right)
	// Value inverse total order comparison (constant matrix at right).
	r = (Expr_GetEvaluation(left) > right)
endfunction


function [r]=%s_2_ADEXPR(left, right)
	// Value inverse total order comparison (constant matrix at left).
	r = (right < left)
endfunction


function [r]=%ADEXPR_2_hm(left, right)
	// Value inverse total order comparison (hypermatrix at right).
	r = (Expr_GetEvaluation(left) > right)
endfunction


function [r]=%hm_2_ADEXPR(left, right)
	// Value inverse total order comparison (hypermatrix at left).
	r = (right < left)
endfunction


function [r]=%ADEXPR_3_s(left, right)
	// Value partial order comparison (constant matrix at right).
	r = (Expr_GetEvaluation(left) <= right)
endfunction


function [r]=%s_3_ADEXPR(left, right)
	// Value partial order comparison (constant matrix at left).
	r = (right >= left)
endfunction


function [r]=%ADEXPR_3_hm(left, right)
	// Value partial order comparison (hypermatrix at right).
	r = (Expr_GetEvaluation(left) <= right)
endfunction


function [r]=%hm_3_ADEXPR(left, right)
	// Value partial order comparison (hypermatrix at left).
	r = (right >= left)
endfunction


function [r]=%ADEXPR_4_s(left, right)
	// Value inverse partial order comparison (constant matrix at right).
	r = (Expr_GetEvaluation(left) >= right)
endfunction


function [r]=%s_4_ADEXPR(left, right)
	// Value inverse partial order comparison (constant matrix at left).
	r = (right <= left)
endfunction


function [r]=%ADEXPR_4_hm(left, right)
	// Value inverse partial order comparison (hypermatrix at right).
	r = (Expr_GetEvaluation(left) >= right)
endfunction


function [r]=%hm_4_ADEXPR(left, right)
	// Value inverse partial order comparison (hypermatrix at left).
	r = (right <= left)
endfunction


function [r]=%ADEXPR_abs(expr)
	// Absolute value of an expression hypermatrix.
	r = extendExpr(expr, makeNodeAbs(abs(Expr_GetEvaluation(expr))))
endfunction


// End of module.

