//
// Functions for forward visitor that computes right finite difference for some function built as an expression, using finite
// difference chain rule.
//
// This works on expressions that depend on a single R^n variable.
//
// Author : Benoit Hamelin
// 2003 Universit de Sherbrooke
//


function [fdcalc]=makeFiniteDiffCalc()
	fdcalc = tlist(SCIAD_FDCALC_FIELDS, ...
			list(), ...
			FDCalc_VisitIVar, ...
			FDCalc_VisitConst, ...
			FDCalc_VisitAdd, ...
			FDCalc_VisitAddInv, ...
			FDCalc_VisitT, ...
			FDCalc_VisitMult, ...
			FDCalc_VisitElemMult, ...
			FDCalc_VisitInv, ...
			FDCalc_VisitElemInv, ...
			FDCalc_VisitExp, ...
			FDCalc_VisitElemExp, ...
			FDCalc_VisitLog, ...
			FDCalc_VisitMonome, ...
			FDCalc_VisitElemMonome, ...
			FDCalc_VisitSin, ...
			FDCalc_VisitCos, ...
			FDCalc_VisitAtan, ...
			FDCalc_VisitAbs, ...
			FDCalc_VisitStub, ...
			0, ...
			numzero, ...
			list(), ...
			0, ...
			0 ...
			)
endfunction


function [r,nbNodesSeries,nbTermsSeries]=FDCalc_Process(this, expr, increment)
	// Processes the right finite difference calculation job on given expression, for given increment.
	this.increment = increment
	[expr,this] = Expr_VisitForward(expr, this)
	r = this.fdList(Expr_GetRoot(expr))
	nbNodesSeries = this.nbNodesSeries
	nbTermsSeries = this.nbTermsSeries
endfunction


function [index]=FDCalc_VisitIVar(node, index)
	// Not a composed function : f(x) = x ==> f(x+h) - f(x) = x + h - x = h
	global SCIAD_EXPR SCIAD_VISITOR
	node = Expr_GetNode(SCIAD_EXPR($), index)
	i = node.index
	SCIAD_VISITOR($).fdList(index) = SCIAD_VISITOR($).increment(i)
endfunction


function [index]=FDCalc_VisitConst(node, index)
	// Not a composed function : f(x) = K ==> f(x+h) - f(x) = K - K = 0
	global SCIAD_EXPR SCIAD_VISITOR
	node = Expr_GetNode(SCIAD_EXPR($), index)
	SCIAD_VISITOR($).fdList(index) = zeros(node.value)
endfunction


function [index]=FDCalc_VisitAdd(node, index)
	// f(x+h) + g(x+h) - f(x) - g(x) = D_h f(x) + D_h g(x)
	global SCIAD_EXPR SCIAD_VISITOR
	predec = SCIAD_VISITOR($).listPredec(index)
	SCIAD_VISITOR($).fdList(index) = SCIAD_VISITOR($).fdList(predec(1)) + ...
			SCIAD_VISITOR($).fdList(predec(2))
endfunction


function [index]=FDCalc_VisitAddInv(node, index)
	// f(x) = -g(x) ==> f(x+h) - f(x) = -g(x+h) - (-g(x)) = -D_h g(x)
	global SCIAD_EXPR SCIAD_VISITOR
	predec = SCIAD_VISITOR($).listPredec(index)
	SCIAD_VISITOR($).fdList(index) = -SCIAD_VISITOR($).fdList(predec)
endfunction


function [index]=FDCalc_VisitT(node, index)
	// f(x) = g(x)' ==> f(x+h) - f(x) = g(x+h)' - g(x)' = [g(x+h) - g(x)]' = [D_h g(x)]'
	global SCIAD_EXPR SCIAD_VISITOR
	predec = SCIAD_VISITOR($).listPredec(index)
	SCIAD_VISITOR($).fdList(index) = SCIAD_VISITOR($).fdList(predec)'
endfunction


function [index]=FDCalc_VisitMult(node, index)
	// f(x) = g(x)h(x) ==> f(x+t) - f(x) = g(x+t)h(x+t) - g(x)h(x)
	//                                   = [D_t g(x) + g(x)][D_t h(x) + h(x)] - g(x)h(x)
	//                                   = D_t g(x) D_t h(x) + h(x) D_t g(x) + g(x) D_t h(x) + g(x)h(x) - g(x)h(x)
	//                                   = D_t g(x) D_t h(x) + h(x) D_t g(x) + g(x) D_t h(x)
	global SCIAD_EXPR SCIAD_VISITOR
	predec = SCIAD_VISITOR($).listPredec(index)

	dfLeft = SCIAD_VISITOR($).fdList(predec(1))
	nodeLeft = Expr_GetNode(SCIAD_EXPR($), predec(1))
	fLeft = nodeLeft.value

	dfRight = SCIAD_VISITOR($).fdList(predec(2))
	nodeRight = Expr_GetNode(SCIAD_EXPR($), predec(2))
	fRight = nodeRight.value

	SCIAD_VISITOR($).fdList(index) = dfLeft * dfRight + dfLeft * fRight + fLeft * dfRight
endfunction


function [index]=FDCalc_VisitElemMult(node, index)
	// Same formula as for inner product.
	global SCIAD_EXPR SCIAD_VISITOR
	predec = SCIAD_VISITOR($).listPredec(index)

	dfLeft = SCIAD_VISITOR($).fdList(predec(1))
	nodeLeft = Expr_GetNode(SCIAD_EXPR($), predec(1))
	fLeft = nodeLeft.value

	dfRight = SCIAD_VISITOR($).fdList(predec(2))
	nodeRight = Expr_GetNode(SCIAD_EXPR($), predec(2))
	fRight = nodeRight.value

	SCIAD_VISITOR($).fdList(index) = dfLeft .* dfRight + fRight .* dfLeft + fLeft .* dfRight
endfunction


function [index]=FDCalc_VisitInv(node, index)
	// f(x) = inv[g(x)] ==>              f(x+h) - f(x) = inv[g(x+h)] - inv[g(x)]
	//                                                 = inv[D_h g(x) + g(x)] - inv[g(x)]
	//                  ==> [D_h g(x) + g(x)] D_h f(x) = I - [D_h g(x) + g(x)] inv[g(x)]
	//                                                 = I - D_h g(x) inv[g(x)] - I
	//                                                 = -D_h g(x) inv[g(x)]
	//                  ==>                   D_h f(x) = -inv[D_h g(x) + g(x)] * [D_h g(x)] * inv[g(x)]
	global SCIAD_EXPR SCIAD_VISITOR
	predec = SCIAD_VISITOR($).listPredec(index)
	dfPred = SCIAD_VISITOR($).fdList(predec)
	nodePred = Expr_GetNode(SCIAD_EXPR($), predec)
	fPred = nodePred.value
	nodeHere = Expr_GetNode(SCIAD_EXPR($), index)
	fHere = nodeHere.value

	SCIAD_VISITOR($).fdList(index) = -inv(dfPred + fPred) * dfPred * fHere
endfunction


function [index]=FDCalc_VisitElemInv(node, index)
	// For each element of operands, following reasoning holds :
	// f(x) = 1 / g(x) ==> D_h f(x) = 1 / g(x+h) - 1 / g(x)
	//                              = 1 / [D_h g(x) + g(x)] - 1 / g(x)
	//                              = [g(x) - D_h g(x) - g(x)] / [g(x) D_h g(x) + g(x)^2]
	//                              = -D_h g(x) / [g(x) D_h g(x) + g(x)^2]
	global SCIAD_EXPR SCIAD_VISITOR
	predec = SCIAD_VISITOR($).listPredec(index)
	dfPred = SCIAD_VISITOR($).fdList(predec)
	nodePred = Expr_GetNode(SCIAD_EXPR($), predec)
	fPred = nodePred.value

	SCIAD_VISITOR($).fdList(index) = -dfPred ./ (fPred .* dfPred + fPred .^ 2)
endfunction


function [index]=FDCalc_VisitExp(node, index)
	// f(x) = a^{g(x)} ==> f(x+h) - f(x) = a^{g(x+h)} - a^{g(x)}
	//                                   = a^{g(x) + D_h g(x)} - a^{g(x)}
	//                                   = a^{g(x)}[a^{D_h g(x)} - 1]
	//                                   = a^{g(x)}[1 + (1/log(a)) D_h g(x) + (1/log(a))^2 (1/2) [D_h g(x)]^2 +
	//                                         (1/log(a))^3 (1/6) [D_h g(x)]^3 + ... - 1]
	//                                   = a^{g(x)} D_h g(x) [1/log(a) + (1/log(a))^2 (1/2) [D_h g(x)] + ...]

	global SCIAD_EXPR SCIAD_VISITOR
	predec = SCIAD_VISITOR($).listPredec(index)
	exp_gx = node.value
	Dgx = SCIAD_VISITOR($).fdList(predec)

	// Evaluate infinite serie up to required precision.
	log_base = logm(node.base)
	d = 1
	n = log_base
	i = 1
	last = 0
	curr = n / d
	while (or(abs(curr - last) > SCIAD_VISITOR($).numzero) | (i < 2)) do
		i = i + 1
		d = d * i
		n = Dgx * n * log_base  // Dgx is scalar.
		last = curr
		curr = curr + n / d  // d is scalar.
	end  // while

	SCIAD_VISITOR($).nbNodesSeries = SCIAD_VISITOR($).nbNodesSeries + 1
	SCIAD_VISITOR($).nbTermsSeries = SCIAD_VISITOR($).nbTermsSeries + i
	SCIAD_VISITOR($).fdList(index) = Dgx * exp_gx * curr
endfunction


function [index]=FDCalc_VisitElemExp(node, index)
	// Same as regular exponentiation, except that we use element-wise operations.

	global SCIAD_EXPR SCIAD_VISITOR
	predec = SCIAD_VISITOR($).listPredec(index)
	value_here = node.value
	Dgx = SCIAD_VISITOR($).fdList(predec)

	// Evaluate infinite serie up to required precision.
	log_base = zeros(node.base)
	index_non_null = find(node.base > 0)
	log_base(index_non_null) = log(node.base(index_non_null))
	d = 1
	n = log_base
	i = 1
	last = 0
	curr = n ./ d
	while (or(abs(curr - last) > SCIAD_VISITOR($).numzero) | i < 2) do
		i = i + 1
		d = d * i
		n = n .* log_base .* Dgx
		last = curr
		curr = curr + n ./ d
	end  // while

	SCIAD_VISITOR($).nbNodesSeries = SCIAD_VISITOR($).nbNodesSeries + 1
	SCIAD_VISITOR($).nbTermsSeries = SCIAD_VISITOR($).nbTermsSeries + i
	SCIAD_VISITOR($).fdList(index) = value_here .* Dgx .* curr
endfunction


function [index]=FDCalc_VisitLog(node, index)
	// f(x) = log g(x)
	// D_h f(x) = log g(x+h) - log g(x)
	//          = log [g(x+h) / g(x)]
	//          = log [(D_h g(x) + g(x)) / g(x)]
	//          = log [1 + D_h g(x) / g(x)]

	global SCIAD_EXPR SCIAD_VISITOR
	predec = SCIAD_VISITOR($).listPredec(index)
	node_gx = Expr_GetNode(SCIAD_EXPR($), predec)
	gx = node_gx.value
	Dgx = SCIAD_VISITOR($).fdList(predec)

	SCIAD_VISITOR($).fdList(index) = log1p(Dgx ./ gx)
endfunction


function [index]=FDCalc_VisitMonome(node, index)
	// CAREFUL - This implementation, for the moment, does not work for a matricial base.
	//
	// f(x) = [g(x)]^k
	//
	// The integral positive part of the exponent is used as a decomposition into a product of recursive squares.  For each power
	// of two, the square elevation formula is used :
	//
	// D_h (g(x)^2) = g(x) D_h g(x) + [D_h g(x)] g(x) + [D_h g(x)]^2
	//
	// Then product rules are applied on combinations of appropriate squares :
	//
	// D_h (p(x) q(x)) = p(x) D_h q(x) + [D_h p(x)] q(x) + [D_h p(x)][D_h q(x)]
	//
	// For the remaining part of the exponent, we can express f(x) with a Taylor development around a = g(x), which yields
	//     f(x) = g(x)^k + k[g(x)^{k-1}](x - g(x)) + (1/2)k(k-1)[g(x)^{k-2}](x - g(x))^2 + (1/6)k(k-1)(k-2)[g(x)^{k-3}](x - g(x))^3 + ...
	//
	// Obviously, it is assumed that g(x)^{k-i} exists for all i.  In light of this, we deduct :
	// ==> D_h f(x) = g(x+h)^k - g(x)^k
	//              = [g(x) + D_h g(x)]^k - g(x)^k
	//              = SUM_{i=1}^{INF} (1/i!) (PROD_{j=1}^{i} (k-j+1)) g(x)^{k-i} [D_h g(x)]^i
	//
	// Actually, this implementation will work even for non-invertible g(x) : in this case, finite difference calculation will
	// however be an approximation (terms with negative k-i will be truncated).

	global SCIAD_EXPR SCIAD_VISITOR
	predec = SCIAD_VISITOR($).listPredec(index)
	result = []
	node_gx = Expr_GetNode(SCIAD_EXPR($), predec)
	gx = node_gx.value
	gx_k = node.value
	Dgx = SCIAD_VISITOR($).fdList(predec)

	if (and(abs(gx) < SCIAD_VISITOR($).numzero)) then
		result = Dgx ^ node.exponent
	else
		listPowers2 = node.listPowers2
		listIndex = node.listIndex
		listProducts = node.listProds
		s_listPowers2 = size(listPowers2)
		if (s_listPowers2 <> 0) then

			// Compute list of deltas.
			listDeltaPowers2 = list(Dgx)
			for i = 2:s_listPowers2 do
				lastPower = listPowers2(i-1)
				lastDelta = listDeltaPowers2($)
				listDeltaPowers2($+1) = lastPower * lastDelta + lastDelta * lastPower + lastDelta^2
			end  // for

			// Compute deltas of successive products.
			result = listDeltaPowers2(listIndex(1))
			s_listIndex = size(listIndex)
			for i = 2:s_listIndex do
				p = listProducts(listIndex(i - 1))
				q = listPowers2(listIndex(i))
				dp = result
				dq = listDeltaPowers2(listIndex(i))
				result = p * dq + dp * q + dp * dq
			end  // for
		end  // if

		rootExpo = node.rootExpo
		rootFactor = node.rootFactor
		if (rootExpo <> 0) then
			// Other exponent : use Taylor development summation.
			d = 1
			n = rootExpo

			det_gx = det(gx)
			inv_gx = []
			i = 1
			Dgx_i = Dgx
			if (det_gx > SCIAD_VISITOR($).numzero) then
				inv_gx = inv(gx)
				gx_i = rootFactor * inv_gx
				curr = 0
			else
				gx_i = []
				curr = (n / d) * Dgx_i
			end  // if

			if (gx_i <> []) then
				while (%t) do
					last = curr
					curr = curr + (n / d) * gx_i * Dgx_i
					if ((i >= 2) & and(abs(last - curr) < SCIAD_VISITOR($).numzero)) then
						break
					end  // if

					i = i + 1
					d = d * i
					n = n * (rootExpo - i + 1)
					Dgx_i = Dgx_i * Dgx
					if (isinf(Dgx_i)) then
						break
					end  // if
					gx_i = gx_i * inv_gx
					if (isinf(gx_i)) then
						break
					end  // if
					SCIAD_VISITOR($).nbNodesSeries = SCIAD_VISITOR($).nbNodesSeries + 1
					SCIAD_VISITOR($).nbTermsSeries = SCIAD_VISITOR($).nbTermsSeries + i
				end  // while
			end  // if
			if (result == []) then
				result = curr
			else
				result = listProducts($) * curr + result * node.rootFactor + result * curr
			end  // if
		end  // if
	end  // if

	if (result == []) then
		error("Bad implementation of AFD of regular monomials.")
	end  // if
	SCIAD_VISITOR($).fdList(index) = result
endfunction


function [index]=FDCalc_VisitElemMonome(node, index)
	// Same as regular monome, but with element-wise operations.

	global SCIAD_EXPR SCIAD_VISITOR
	predec = SCIAD_VISITOR($).listPredec(index)
	result = []
	node_gx = Expr_GetNode(SCIAD_EXPR($), predec)
	gx = node_gx.value
	gx_k = node.value
	Dgx = SCIAD_VISITOR($).fdList(predec)

	listPowers2 = node.listPowers2
	listMasks = node.listMasks
	listProducts = node.listProds
	s_listPowers2 = size(listPowers2)
	if (s_listPowers2 <> 0) then

		// Compute list of deltas.
		listDeltaPowers2 = list(Dgx)
		for i = 2:s_listPowers2 do
			lastPower = listPowers2(i-1)
			lastDelta = listDeltaPowers2($)
			listDeltaPowers2($+1) = lastPower .* lastDelta + lastDelta .* lastPower + lastDelta .^ 2
		end  // for

		// Compute deltas of successive products.
		result = zeros(gx)
		mask = find(listMasks(1) == 1)
		result(mask) = listDeltaPowers2(1)(mask)
		for i = 2:size(listDeltaPowers2) do
			mask = find(listMasks(i) == 1)
			p = listProducts(i - 1)(mask)
			q = listPowers2(i)(mask)
			dp = result(mask)
			dq = listDeltaPowers2(i)(mask)
			result(mask) = p .* dq + dp .* q + dp .* dq
		end  // for
	end  // if

	rootExpo = node.rootExpo
	rootFactor = node.rootFactor
	if (or(rootExpo <> 0)) then

		// Other exponent : use Taylor development summation.
		d = 1
		n = rootExpo
		orig_sgx = size(gx)
		gx = gx(:)
		i = 1
		index_nonzero = find(abs(gx) > SCIAD_VISITOR($).numzero)
		gx_i = zeros(gx)
		gx_nonzero = gx(index_nonzero)
		gx_i(index_nonzero) = rootFactor(index_nonzero) ./ gx_nonzero
		Dgx_i = Dgx
		curr = zeros(gx)
		if (or(abs(gx_i) > SCIAD_VISITOR($).numzero)) then
			while (%t) do
				last = curr
				curr = curr + (n ./ d) .* gx_i .* Dgx_i
				if ((i >= 2) & and(abs(last - curr) < SCIAD_VISITOR($).numzero)) then
					break
				end  // if

				i = i + 1
				d = d * i
				n = n .* (rootExpo - i + 1)
				Dgx_i = Dgx_i .* Dgx
				if (or(isinf(Dgx_i))) then
					break
				end  // if
				gx_i(index_nonzero) = gx_i(index_nonzero) ./ gx_nonzero
				if (or(isinf(gx_i))) then
					break
				end  // if
			end  // while
			SCIAD_VISITOR($).nbNodesSeries = SCIAD_VISITOR($).nbNodesSeries + 1
			SCIAD_VISITOR($).nbTermsSeries = SCIAD_VISITOR($).nbTermsSeries + i
		end  // if
		curr = matrix(curr, orig_sgx)
		
		if (result == []) then
			result = curr
		else
			result = listProducts($) .* curr + result .* node.rootFactor + result .* curr
		end  // if
	end  // if

	if (result == []) then
		error("Bad implementation of AFD of element-wise monomials.")
	end  // if

	index_zero = find(abs(gx) < SCIAD_VISITOR($).numzero)
	if (index_zero <> []) then
		result(index_zero) = Dgx(index_zero) .^ node.exponent(index_zero)
	end  // if

	SCIAD_VISITOR($).fdList(index) = result
endfunction


function [index]=FDCalc_VisitSin(node, index)
	// f(x) = sin g(x)
	// D_h f(x) = sin g(x+h) - sin g(x)
	//          = sin [D_h g(x) + g(x)] - sin g(x)
	//          = sin g(x) cos [D_h g(x)] + cos g(x) sin [D_h g(x)] - sin g(x)
	//          = sin g(x) [cos D_h g(x) - 1] + cos g(x) sin [D_h g(x)]
	//          = sin g(x) cosm1 [D_h g(x)] + cos g(x) sin [D_h g(x)]

	global SCIAD_EXPR SCIAD_VISITOR
	predec = SCIAD_VISITOR($).listPredec(index)
	node_gx = Expr_GetNode(SCIAD_EXPR($), predec)
	gx = node_gx.value
	sin_gx = node.value
	Dgx = SCIAD_VISITOR($).fdList(predec)

	[cm1,nbTerms] = cosm1(Dgx)
	SCIAD_VISITOR($).fdList(index) = sin_gx .* cm1 + cos(gx) .* sin(Dgx)
	SCIAD_VISITOR($).nbNodesSeries = SCIAD_VISITOR($).nbNodesSeries + 1
	SCIAD_VISITOR($).nbTermsSeries = SCIAD_VISITOR($).nbTermsSeries + nbTerms
endfunction


function [index]=FDCalc_VisitCos(node, index)
	// f(x) = cos g(x)
	// D_h f(x) = cos g(x+h) - cos g(x)
	//          = cos [D_h g(x) + g(x)] - cos g(x)
	//          = cos g(x) cos [D_h g(x)] - sin g(x) sin [D_h g(x)] - cos g(x)
	//          = cos g(x) [cos D_h g(x) - 1] - sin g(x) sin [D_h g(x)]
	//          = cos g(x) cosm1 D_h g(x) - sin g(x) sin [D_h g(x)]

	global SCIAD_EXPR SCIAD_VISITOR
	predec = SCIAD_VISITOR($).listPredec(index)
	node_gx = Expr_GetNode(SCIAD_EXPR($), predec)
	gx = node_gx.value
	cos_gx = node.value
	Dgx = SCIAD_VISITOR($).fdList(predec)

	[cm1,nbTerms] = cosm1(Dgx)
	SCIAD_VISITOR($).fdList(index) = cos_gx .* cm1 - sin(gx) .* sin(Dgx)
	SCIAD_VISITOR($).nbNodesSeries = SCIAD_VISITOR($).nbNodesSeries + 1
	SCIAD_VISITOR($).nbTermsSeries = SCIAD_VISITOR($).nbTermsSeries + nbTerms
endfunction


function [index]=FDCalc_VisitAtan(node, index)
	// f(x) = arctan g(x)
	// D_h f(x) = arctan g(x+h) - arctan g(x)
	//          = arctan [D_h g(x) + g(x)] + arctan (-g(x))
	//          = arctan [ (D_h g(x) + g(x) - g(x)) / ((1 + g(x))(g(x) + D_h g(x))) ]
	//          = arctan [ D_h g(x) / (1 + x(x+h)) ]
	//
	// Other details must be considered to unclamp the difference from the ]-pi/2, pi/2[ interval.

	global SCIAD_EXPR SCIAD_VISITOR
	predec = SCIAD_VISITOR($).listPredec(index)
	node_gx = Expr_GetNode(SCIAD_EXPR($), predec)
	gx = node_gx.value
	Dgx = SCIAD_VISITOR($).fdList(predec)

	// Treat each element according to the argument being positive or negative or null.
	y = gx + Dgx
	r = zeros(gx)
	indexPos = find(gx > SCIAD_VISITOR($).numzero)
	indexNeg = find(gx < -SCIAD_VISITOR($).numzero)
	indexNull = find(abs(gx) < SCIAD_VISITOR($).numzero)

	// Positive argument.
	spos = size(indexPos, "*")
	if (spos > 0) then
		rpos = zeros(indexPos')
		gxPos = gx(indexPos)
		DgxPos = Dgx(indexPos)
		ypos = y(indexPos)
		xyp1 = gxPos .* ypos + 1
		indexInv = find(abs(xyp1) < SCIAD_VISITOR($).numzero)
		indexNotInv = (1:spos)
		indexNotInv(indexInv) = []
		rpos(indexInv) = -%pi / 2
		rpos(indexNotInv) = atan(DgxPos(indexNotInv) ./ xyp1(indexNotInv))
		indexLeftHalf = find(xyp1 < -SCIAD_VISITOR($).numzero)
		rpos(indexLeftHalf) = rpos(indexLeftHalf) - %pi
		r(indexPos) = rpos
	end  // if

	sneg = size(indexNeg, "*")
	if (sneg > 0) then
		rneg = zeros(indexNeg')
		gxNeg = gx(indexNeg)
		DgxNeg = Dgx(indexNeg)
		yneg = y(indexNeg)
		xyp1 = gxNeg .* yneg + 1
		indexInv = find(abs(xyp1) < SCIAD_VISITOR($).numzero)
		indexNotInv = (1:sneg)
		indexNotInv(indexInv) = []
		rneg(indexInv) = %pi / 2
		rneg(indexNotInv) = atan(DgxNeg(indexNotInv) ./ xyp1(indexNotInv))
		indexLeftHalf = find(xyp1 < -SCIAD_VISITOR($).numzero)
		rneg(indexLeftHalf) = rneg(indexLeftHalf) + %pi
		r(indexNeg) = rneg
	end  // if

	r(indexNull) = atan(Dgx(indexNull))
	SCIAD_VISITOR($).fdList(index) = r
endfunction


function [index]=FDCalc_VisitAbs(node, index)
	// f(x) = abs g(x)
	// D_h f(x) = abs g(x+h) - abs g(x)
	//          = abs [D_h g(x) + g(x)] - abs g(x)
	//
	// (D_h g(x) + g(x) > 0) and (g(x) > 0) ==> D_h f(x) = D_h g(x) + g(x) - g(x) = D_h g(x)
	// (D_h g(x) + g(x) > 0) and (g(x) <= 0) ==> D_h f(x) = D_h g(x) + g(x) + g(x) = D_h g(x) + 2g(x)
	// (D_h g(x) + g(x) <= 0) and (g(x) > 0) ==> D_h f(x) = -D_h g(x) - g(x) - g(x) = -D_h g(x) - 2g(x)
	// (D_h g(x) + g(x) <= 0) and (g(x) <= 0) ==> D_h f(x) = -D_h g(x) - g(x) + g(x) = -D_h g(x)

	global SCIAD_EXPR SCIAD_VISITOR
	predec = SCIAD_VISITOR($).listPredec(index)
	node_gx = Expr_GetNode(SCIAD_EXPR($), predec)
	gx = node_gx.value
	Dgx = SCIAD_VISITOR($).fdList(predec)
	DgxPgx = Dgx + gx

	index1 = find((DgxPgx > 0) & (gx > 0))
	index2 = find((DgxPgx > 0) & (gx <= 0))
	index3 = find((DgxPgx <= 0) & (gx > 0))
	index4 = find((DgxPgx <= 0) & (gx <= 0))
	r = zeros(gx)
	r(index1) = Dgx(index1)
	r(index2) = Dgx(index2) + 2 .* gx(index2)
	r(index3) = -Dgx(index3) - 2 .* gx(index3)
	r(index4) = -Dgx(index4)
	SCIAD_VISITOR($).fdList(index) = r
endfunction


function [index]=FDCalc_VisitStub(node, index)
	error("Finite difference calculator : not supposed to find a stub here!")
endfunction


// End of module.

