//
// Functions to interface Sciad system of automatic differentiation.
//
// Author : Benoit Hamelin
// 2002-2004 Universit de Sherbrooke
//


// One builds the computation graph of function f:R^n->R^m by calling f with
// an object built out of the ad_ivar(v) function.  For instance, for applying
// function f at point x \in R^n, one should make :
//
// varx = ad_ivar(x)
// fx = f(varx)
//
// or, shorter :
//
// fx = f(ad_ivar(x))
//
// The call ad_ivar(v), as a matter of fact, gives the computation graph of
// the identity function f(x)=x, applied at point v.
//
// Due to operator overloading, computation graphs can be manipulated as any
// other Scilab scalar/vector/matrix.  Supported operators (yet) :
//
// size
// transposition
// +, - (unary and binary)
// *, .*, /, ./, \, .\
// inv
// ^, .^ (either the base or the exponent must be a constant)
// exp, log, sqrt
// sin, cos, atan
// Horizontal and vertical concatenation
// Extraction :
//     Take v, the computation graph of an operation that gives a vector.
//     Components of v can be extracted as of any vector :
//         v([1,3,$])  --> Gives the comp. graph of operations that yield a
//                         R^3 vector that contains first, third and last
//                         elements of v.
//     Exception to regular usage : one single-index computation graphs of
//     operations that yield a matricial result.  For instance :
//         x = ad_ivar([7;0;-2;3]);
//         M = x * x';
//         M([2,7]) --> This is forbidden.
//     Note that the strange semantics of full vector extraction is respected.
//     For instance, with w the comp. graph of an operation that yields a row
//     vector, w(:) will give w', whereas with v a column vector, v(:) == v.
// Insertion : again, as regular Scilab vector/matrices, the exception
// mentioned above withstanding.


function [v]=ad_value(fx)
	// For fx, the computation graph of function f applied at point x, this
	// returns the numerical value of f(x).
	v = Expr_GetEvaluation(fx)
endfunction


function [newfx]=ad_reeval(fx, newx)
	// Using computation graph fx of function f applied at point x, re-applies
	// the function at point newx without rebuilding the computation graph.
	// For complex computation graphs of fully differentiable functions, this
	// is very time-efficient, as graph-building can get pretty longish.
	//
	// For the moment, only numerical values can be given for newx; one cannot
	// replace independant variables in fx with some ad hoc sub-graph.
	//
	newfx = Expr_Reevaluate(fx, newx)
endfunction


function [g]=ad_gradient_num(fx, n)
	// Gives evaluation of the gradient of f at point x (set when computation
	// graph was built).  One also passes n, the size of the domain vector x.
	// Only the numerical evaluation of the gradient is computed, not the
	// computation graph of the gradient function.  This is much faster than
	// building the gradient graph when one only wishes for the gradient
	// value.
	//
	sfx = size(fx, "*")
	if (sfx > 1) then
		error("Cannot obtain the gradient of a R^n->R^" + string(sfx) + " function.  Use ad_jacobian_num.")
	end  // if
	g = Expr_Gradient(fx, n, %t)
endfunction


function [gx]=ad_gradient_graph(fx, n)
	// Computes the computation graph of the gradient of function f at point x
	// (set when the computation graph fx was built).  One also passes n, the
	// size of the domain vector x.  This is useful when one wants to derive
	// over again to obtain the Hessian of f at x, or even higher-order
	// derivatives.
	//
	sfx = size(fx, "*")
	if (sfx > 1) then
		error("Cannot obtain the gradient of a R^n->R^" + string(sfx) + " function.  Use ad_jacobian_num.")
	end  // if
	gx = Expr_Gradient(fx, n, %f)
endfunction


function [SJ]=ad_jacobian_num(fx, n, S)
	// Evaluates the Jacobian* of function f:R^n->R^m at point x (set when the
	// computation graph fx was built).  One also passes n, the size of the
	// domain vector x, and S, a row selection matrix such that, for J(x)
	// being the Jacobian of f(x), SJ(x) is returned.
	//
	// Only the numerical evaluation of SJ(x) is returned.  This is very
	// time-efficient when one wants only to know the numerical value, since
	// building the computation graph of J(x) can be a lengthy process.
	//
	// * For the computation graph gx of the gradient of f(x), this function
	//   returns the Hessian of f(x).
	//
	sfx = size(fx)
	if (and(sfx > 1)) then
		error("Cannot compute the Jacobian of a matricial function.  Multiply the matrix by a vector first.")
	end  // if
	SJ = Expr_Jacobian(fx, n, S, %t)
endfunction


function [SJ]=ad_jacobian_graph(fx, n, S)
	// Evaluates the Jacobian* of function f:R^n->R^m at point x (set when the
	// computation graph fx was built).  One also passes n, the size of the
	// domain vector x, and S, a row selection matrix such that, for J(x)
	// being the Jacobian of f(x), SJ(x) is returned.
	//
	// The whole computation graph of the Jacobian is built and returned.
	// This is useful when one wants to obtain higher-order derivatives of
	// f(x) (by multiplying the obtained matrix by a direction vector, which
	// yields a new R^n->R^m function, then calling this function over).
	//
	// * For the computation graph gx of the gradient of f(x), this function
	//   returns the Hessian of f(x).
	//
	sfx = size(fx)
	if (and(sfx > 1)) then
		error("Cannot compute the Jacobian of a matricial function.  Multiply the matrix by a vector first.")
	end  // if
	SJ = Expr_Jacobian(fx, n, S, %f)
endfunction


function [D]=ad_finite_diff(fx, d, acc)
	// Calculates right finite difference (using automatic finite differences)
	// D_d f(x) = f(x+d) - f(x), from computation graph fx of the application
	// of f at x.  One must pass an accuracy parameter, which is used as a
	// "numerical zero" for some numerical developments that occur during
	// calculations of automatic finite differences.
	//
	D = Expr_CalcFiniteDiff(fx, d, acc)
endfunction


function [tcoeff]=ad_taylor_coeff(fx, order, x0)
	// Computes a list of the coefficients of the terms of the Taylor
	// development of f(x) around x0; the coefficients computed go from order
	// 0 to given order.  This function only works for R->R functions.
	//
	tcoeff = Expr_TaylorCoeff(fx, order, x0)
endfunction


// End of module.
