//
// Functions for the Taylor developer (for R->R functions).
//
// Taylor constants are gotten on a structure one dim more than the thing the node is for.  For scalars, this structure is a column
// vector.
//
// Author : Benoit Hamelin
// 2004 Universit de Sherbrooke
//


function [taylorDev]=makeTaylorDev(order, anchor)
	taylorDev = tlist(SCIAD_TAYLORDEV_FIELDS, ...
			list(), ...
			TaylorDev_VisitIVar, ...
			TaylorDev_VisitConst, ...
			TaylorDev_VisitAdd, ...
			TaylorDev_VisitAddInv, ...
			TaylorDev_VisitT, ...
			TaylorDev_VisitMult, ...
			TaylorDev_VisitElemMult, ...
			TaylorDev_VisitInv, ...
			TaylorDev_VisitElemInv, ...
			TaylorDev_VisitExp, ...
			TaylorDev_VisitElemExp, ...
			TaylorDev_VisitLog, ...
			TaylorDev_VisitMonome, ...
			TaylorDev_VisitElemMonome, ...
			TaylorDev_VisitSin, ...
			TaylorDev_VisitCos, ...
			TaylorDev_VisitAtan, ...
			TaylorDev_VisitAbs, ...
			TaylorDev_VisitStub, ...
			list(), ...
			order, ...
			anchor ...
			)
endfunction


function [index]=TaylorDev_VisitIVar(node, index)
	// Visit of an independant variable node.
	// The variable is expected to be of R^1.
	global SCIAD_EXPR SCIAD_VISITOR
	if (or(size(node) ~= [1,1])) then
		error("Taylor developer not used on a R->R function.")
	end  // if
	coeffNode = zeros(SCIAD_VISITOR($).order + 1, 1)
	coeffNode(1) = SCIAD_VISITOR($).anchor
	coeffNode(2) = 1
	SCIAD_VISITOR($).coeffList(index) = coeffNode
endfunction


function [index]=TaylorDev_VisitConst(node, index)
	// Visit of a constant node.
	global SCIAD_EXPR SCIAD_VISITOR
	snode = size(node)
	coeffNode = 0
	if (and(snode == [1,1])) then
		// This constant is a scalar.
		coeffNode = zeros(SCIAD_VISITOR($).order+1, 1)
		coeffNode(1) = node.value
	elseif (snode(1) == 1) then
		// This constant is a row vector.
		coeffNode = zeros(SCIAD_VISITOR($).order+1, snode(2))
		coeffNode(1,:) = node.value
	elseif (snode(2) == 1) then
		// This constant is a column vector.
		coeffNode = zeros(snode(1), SCIAD_VISITOR($).order+1)
		coeffNode(:,1) = node.value
	else
		coeffNode = zeros(snode(1), snode(2), SCIAD_VISITOR($).order+1)
		coeffNode(:,:,1) = node.value
	end  // if
	SCIAD_VISITOR($).coeffList(index) = coeffNode
endfunction


function [index]=TaylorDev_VisitAdd(node, index)
	// Addition node : Taylor coefficients are simply added.
	global SCIAD_EXPR SCIAD_VISITOR
	predec = SCIAD_VISITOR($).listPredec(index)
	SCIAD_VISITOR($).coeffList(index) = ...
			SCIAD_VISITOR($).coeffList(predec(1)) + ...
			SCIAD_VISITOR($).coeffList(predec(2))
endfunction


function [index]=TaylorDev_VisitAddInv(node, index)
	// Additive inverse node : Taylor coefficients are opposed.
	global SCIAD_EXPR SCIAD_VISITOR
	predec = SCIAD_VISITOR($).listPredec(index)
	SCIAD_VISITOR($).coeffList(index) = -SCIAD_VISITOR($).coeffList(predec)
endfunction


function [index]=TaylorDev_VisitT(node, index)
	// Transposition node : Taylor coefficient structures are transposed, unless the node is scalar.
	global SCIAD_EXPR SCIAD_VISITOR
	predec = SCIAD_VISITOR($).listPredec(index)
	s = size(node)
	if (and(size(node) == [1,1])) then
		SCIAD_VISITOR($).coeffList(index) = SCIAD_VISITOR($).coeffList(predec)
	elseif (and(s > [1,1])) then
		// Must go through weird ways to transpose an hypermatrix...
		h = SCIAD_VISITOR($).coeffList(predec)
		sh = size(h)
		sframe = prod(s)
		transpElems = []
		for i = 1:sframe:(prod(sh) - sframe + 1) do
			frame = matrix(h(i:(i+sframe-1)), [sh(1),sh(2)])'
			transpElems = [transpElems; frame(:)]
		end  // for
		SCIAD_VISITOR($).coeffList(index) = hypermat([sh(2),sh(1),sh(3:$)], transpElems)
	else
		SCIAD_VISITOR($).coeffList(index) = SCIAD_VISITOR($).coeffList(predec)'
	end  // if
endfunction


function [index]=TaylorDev_VisitMult(node, index)
	// Product node.  Multiple rules for defining the coefficient structures
	// must be considered.

	global SCIAD_EXPR SCIAD_VISITOR
	predec = SCIAD_VISITOR($).listPredec(index)
	indexLeft = predec(1)
	nodeLeft = Expr_GetNode(SCIAD_EXPR($), indexLeft)
	sizeLeft = size(nodeLeft)
	coeffLeft = SCIAD_VISITOR($).coeffList(indexLeft)
	indexRight = predec(2)
	nodeRight = Expr_GetNode(SCIAD_EXPR($), indexRight)
	sizeRight = size(nodeRight)
	coeffRight = SCIAD_VISITOR($).coeffList(indexRight)
	coeffNode = 0

	if (and(sizeLeft == [1,1])) then
		// Left is a scalar.
		if (sizeRight(1) ~= 1) then
			error("TaylorDev : in product node " + index + ", left and right are not coherent operands.")
		end  // if
		if (sizeRight(2) == 1) then
			// Right is a scalar too.
			coeffNode = zeros(SCIAD_VISITOR($).order+1, 1)
			for k = 0:SCIAD_VISITOR($).order do
				for i = 0:k do
					coeffNode(k + 1) = coeffNode(k + 1) + coeffLeft(i + 1) * coeffRight(k - i + 1)
				end  // for
			end  // for
		else
			// Right is a row vector.
			coeffNode = zeros(SCIAD_VISITOR($).order+1, sizeRight(2))
			for k = 0:SCIAD_VISITOR($).order do
				for i = 0:k do
					coeffNode(k+1,:) = coeffNode(k+1,:) + coeffLeft(i+1) * coeffRight(k-i+1,:)
				end  // for
			end  // for
		end  // if
	elseif (sizeLeft(1) == 1) then
		// Left is a row vector.
		if (sizeRight(2) == 1) then
			// Right is a column vector.
			coeffNode = zeros(SCIAD_VISITOR($).order+1, 1)
			for k = 0:SCIAD_VISITOR($).order do
				for i = 0:k do
					coeffNode(k+1) = coeffNode(k+1) + coeffLeft(i+1,:) * coeffRight(:,k-i+1)
				end  // for
			end  // for
		else
			// Right is a matrix.
			coeffNode = zeros(SCIAD_VISITOR($).order+1, sizeRight(2))
			for k = 0:SCIAD_VISITOR($).order do
				for i = 0:k do
					coeffNode(k+1,:) = coeffNode(k+1,:) + coeffLeft(i+1,:) * coeffRight(:,:,k-i+1)
				end  // for
			end  // for
		end  // if
	elseif (sizeLeft(2) == 1) then
		// Left is a column vector.
		if (sizeRight(2) == 1) then
			// Right is a scalar.
			coeffNode = zeros(sizeLeft(1), SCIAD_VISITOR($).order+1)
			for k = 0:SCIAD_VISITOR($).order do
				for i = 0:k do
					coeffNode(:,k+1) = coeffNode(:,k+1) + coeffLeft(:,i+1) * coeffRight(k-i+1)
				end  // for
			end  // for
		else
			// Right is a row vector.
			coeffNode = zeros(sizeLeft(1), sizeRight(2), SCIAD_VISITOR($).order+1)
			for k = 0:SCIAD_VISITOR($).order do
				for i = 0:k do
					coeffNode(:,:,k+1) = coeffNode(:,:,k+1) + coeffLeft(:,i+1) * coeffRight(k-i+1,:)
				end  // for
			end  // for
		end  // if
	else
		// Left is a matrix.
		if (sizeRight(2) == 1) then
			// Right is a column vector.
			coeffNode = zeros(sizeLeft(1), SCIAD_VISITOR($).order+1)
			for k = 0:SCIAD_VISITOR($).order do
				for i = 0:k do
					coeffNode(:,k+1) = coeffNode(:,k+1) + coeffLeft(:,:,i+1) * coeffRight(:,k-i+1)
				end  // for
			end  // for
		else
			// Right is another matrix.
			coeffNode = zeros(sizeLeft(1), sizeRight(2), SCIAD_VISITOR($).order+1)
			for k = 0:SCIAD_VISITOR($).order do
				for i = 0:k do
					coeffNode(:,:,k+1) = coeffNode(:,:,k+1) + coeffLeft(:,:,i+1) * coeffRight(:,:,k-i+1)
				end  // for
			end  // for
		end // if
	end  // if

	SCIAD_VISITOR($).coeffList(index) = coeffNode
endfunction


function [index]=TaylorDev_VisitElemMult(node, index)
	// Element-wise product node.

	global SCIAD_EXPR SCIAD_VISITOR
	predec = SCIAD_VISITOR($).listPredec(index)
	nodeLeft = Expr_GetNode(SCIAD_EXPR($), predec(1))
	sizeLeft = size(nodeLeft)
	coeffLeft = SCIAD_VISITOR($).coeffList(predec(1))
	nodeRight = Expr_GetNode(SCIAD_EXPR($), predec(2))
	sizeRight = size(nodeRight)
	coeffRight = SCIAD_VISITOR($).coeffList(predec(2))
	coeffNode = 0

	if (and(sizeLeft == [1,1])) then
		// Operands are scalars.
		coeffNode = zeros(SCIAD_VISITOR($).order+1, 1)
		for k = 0:SCIAD_VISITOR($).order do
			for i = 0:k do
				coeffNode(k+1) = coeffNode(k+1) + coeffLeft(i+1) .* coeffRight(k-i+1)
			end  // for
		end  // for
	elseif (sizeLeft(1) == 1) then
		// Operands are row vectors.
		coeffNode = zeros(SCIAD_VISITOR($).order+1, sizeLeft(2))
		for k = 0:SCIAD_VISITOR($).order do
			for i = 0:k do
				coeffNode(k+1,:) = coeffNode(k+1,:) + coeffLeft(i+1,:) .* coeffRight(k-i+1,:)
			end  // for
		end  // for
	elseif (sizeLeft(2) == 1) then
		// Operands are column vectors.
		coeffNode = zeros(sizeLeft(1), SCIAD_VISITOR($).order+1)
		for k = 0:SCIAD_VISITOR($).order do
			for i = 0:k do
				coeffNode(:,k+1) = coeffNode(:,k+1) + coeffLeft(:,i+1) .* coeffRight(:,k-i+1)
			end  // for
		end  // for
	else
		// Operands are matrices.
		coeffNode = zeros(sizeLeft(1), sizeLeft(2), SCIAD_VISITOR($).order+1)
		for k = 0:SCIAD_VISITOR($).order do
			for i = 0:k do
				coeffNode(:,:,k+1) = coeffNode(:,:,k+1) + coeffLeft(:,:,i+1) .* coeffRight(:,:,k-i+1)
			end  // for
		end  // for
	end  // if

	SCIAD_VISITOR($).coeffList(index) = coeffNode
endfunction


function [index]=TaylorDev_VisitInv(node, index)
	// Inversion node.
	// We only support scalar inversion for now.

	global SCIAD_EXPR SCIAD_VISITOR
	predec = SCIAD_VISITOR($).listPredec(index)
	sz = size(Expr_GetNode(SCIAD_EXPR($), predec))
	if (or(sz ~= [1,1])) then
		error("TaylorDev : inversion node " + index + " relates to non-scalar operand, which isn''t supported.")
	end  // if
	coeffOp = SCIAD_VISITOR($).coeffList(predec)
	if (coeffOp(1) == 0) then
		error("TaylorDev : at inversion node " + index + ", operand has null 0-order term.")
	end  // if

	coeffNode = zeros(SCIAD_VISITOR($).order+1, 1)
	coeffNode(1) = inv(coeffOp(1))
	for k = 1:SCIAD_VISITOR($).order do
		s = 0
		for i = 1:k do
			s = s + coeffOp(i+1) * coeffNode(k-i+1)
		end  // for
		coeffNode(k+1) = -s / coeffOp(1)
	end  // for

	SCIAD_VISITOR($).coeffList(index) = coeffNode
endfunction


function [index]=TaylorDev_VisitElemInv(node, index)
	// Element-wise inversion node.

	global SCIAD_EXPR SCIAD_VISITOR
	predec = SCIAD_VISITOR($).listPredec(index)
	sizeOp = size(Expr_GetNode(SCIAD_EXPR($), predec))
	coeffOp = SCIAD_VISITOR($).coeffList(predec)

	coeffNode = zeros(coeffOp)
	if (and(sizeOp == [1,1])) then
		// Operand is scalar.
		if (coeffOp(1) == 0) then
			error("TaylorDev : at elem-inversion node " + index + ", scalar operand has null 0-order term.")
		end  // if
		coeffNode(1) = 1 ./ coeffOp(1)
		for k = 1:SCIAD_VISITOR($).order do
			s = 0
			for i = 1:k do
				s = s + coeffOp(i+1) .* coeffNode(k-i+1)
			end  // for
			coeffNode(k+1) = -s ./ coeffOp(1)
		end  // for
	elseif (sizeOp(1) == 1) then
		// Operand is a row vector.
		if (or(coeffOp(1,:) == 0)) then
			error("TaylorDev : at elem-inversion node " + index + ", row-vector operand has null 0-order term.")
		end  // if
		coeffNode(1,:) = ones(1,sizeOp(2)) ./ coeffOp(1,:)
		for k = 1:SCIAD_VISITOR($).order do
			s = zeros(1, sizeOp(2))
			for i = 1:k do
				s = s + coeffOp(i+1,:) .* coeffNode(k-i+1,:)
			end  // for
			coeffNode(k+1,:) = -s ./ coeffOp(1,:)
		end  // for
	elseif (sizeOp(2) == 1) then
		// Operand is a column vector.
		if (or(coeffOp(:,1) == 0)) then
			error("TaylorDev : at elem-inversion node " + index + ", column-vector operand has null 0-order term.")
		end  // if
		coeffNode(:,1) = ones(sizeOp(1),1) ./ coeffOp(:,1)
		for k = 1:SCIAD_VISITOR($).order do
			s = zeros(sizeOp(1), 1)
			for i = 1:k do
				s = s + coeffOp(:,i+1) .* coeffNode(:,k-i+1)
			end  // for
			coeffNode(:,k+1) = -s ./ coeffOp(:,1)
		end  // for
	else
		// Operand is a matrix.
		if (or(coeffOp(:,:,1) == 0)) then
			error("TaylorDev : at elem-inversion node " + index + ", matrix operand has null 0-order term.")
		end  // if
		coeffNode(:,:,1) = ones(sizeOp(1), sizeOp(2)) ./ coeffOp(:,:,1)
		for k = 1:SCIAD_VISITOR($).order do
			s = zeros(sizeOp(1), sizeOp(2))
			for i = 1:k do
				s = s + coeffOp(:,:,i+1) .* coeffNode(:,:,k-i+1)
			end  // for
			coeffNode(:,:,k+1) = -s ./ coeffOp(:,:,1)
		end  // for
	end  // if

	SCIAD_VISITOR($).coeffList(index) = coeffNode
endfunction


function [index]=TaylorDev_VisitExp(node, index)
	// Exponential node.
	// We only support scalar exponentiation.

	global SCIAD_EXPR SCIAD_VISITOR
	predec = SCIAD_VISITOR($).listPredec(index)
	sz = size(Expr_GetNode(SCIAD_EXPR($), predec))
	if (or(sz ~= [1,1])) then
		error("TaylorDev : non-scalar exponentiation (node " + index + ") is not supported.")
	end  // if
	coeffPred = SCIAD_VISITOR($).coeffList(predec) * log(node.base)

	coeffNode = zeros(coeffPred)
	coeffNode(1) = exp(coeffPred(1))
	for k = 1:SCIAD_VISITOR($).order do
		for i = 0:(k-1) do
			coeffNode(k+1) = coeffNode(k+1) + (k - i) * coeffNode(i+1) * coeffPred(k-i+1)
		end  // for
		coeffNode(k+1) = coeffNode(k+1) / k
	end  // for

	SCIAD_VISITOR($).coeffList(index) = coeffNode
endfunction


function [index]=TaylorDev_VisitElemExp(node, index)
	// Element-wise exponential node.

	global SCIAD_EXPR SCIAD_VISITOR
	predec = SCIAD_VISITOR($).listPredec(index)
	sz = size(Expr_GetNode(SCIAD_EXPR($), predec))
	coeffPred = SCIAD_VISITOR($).coeffList(predec)

	coeffNode = zeros(coeffPred)
	if (and(sz == [1,1])) then
		// Operand is scalar.
		coeffPred = coeffPred .* log(node.base)
		coeffNode(1) = exp(coeffPred(1))
		for k = 1:SCIAD_VISITOR($).order do
			for i = 0:(k-1) do
				coeffNode(k+1) = coeffNode(k+1) + (k-i) .* coeffNode(i+1) .* coeffPred(k-i+1)
			end  // for
			coeffNode(k+1) = coeffNode(k+1) ./ k
		end  // for
	elseif (sz(1) == 1) then
		// Operand is a row vector.
		coeffPred = coeffPred .* (ones(SCIAD_VISITOR($).order + 1, 1) * log(node.base))
		coeffNode(1,:) = exp(coeffPred(1,:))
		for k = 1:SCIAD_VISITOR($).order do
			for i = 0:(k-1) do
				coeffNode(k+1,:) = coeffNode(k+1,:) + (k-i) .* coeffNode(i+1,:) .* coeffPred(k-i+1,:)
			end  // for
			coeffNode(k+1,:) = coeffNode(k+1,:) ./ k
		end  // for
	elseif (sz(2) == 1) then
		// Operand is a column vector.
		coeffPred = coeffPred .* (log(node.base) * ones(1, SCIAD_VISITOR($).order + 1))
		coeffNode(:,1) = exp(coeffPred(:,1))
		for k = 1:SCIAD_VISITOR($).order do
			for i = 0:(k-1) do
				coeffNode(:,k+1) = coeffNode(:,k+1) + (k-i) .* coeffNode(:,i+1) .* coeffPred(:,k-i+1)
			end  // for
			coeffNode(:,k+1) = coeffNode(:,k+1) ./ k
		end  // for
	else
		// Operand is a matrix.
		logbase = log(node.base)
		for i = 0:SCIAD_VISITOR($).order do
			coeffPred(:,:,i+1) = coeffPred(:,:,i+1) .* logbase
		end  // for
		coeffNode(:,:,1) = exp(coeffPred(:,:,1))
		for k = 1:SCIAD_VISITOR($).order do
			for i = 0:(k-1) do
				coeffNode(:,:,k+1) = coeffNode(:,:,k+1) + (k-i) .* coeffNode(:,:,i+1) .* coeffPred(:,:,k-i+1)
			end  // for
			coeffNode(:,:,k+1) = coeffNode(:,:,k+1) ./ k
		end  // for
	end  // if

	SCIAD_VISITOR($).coeffList(index) = coeffNode
endfunction


function [index]=TaylorDev_VisitLog(node, index)
	// (Element-wise) natural logarithm node.

	global SCIAD_EXPR SCIAD_VISITOR
	predec = SCIAD_VISITOR($).listPredec(index)
	sz = size(Expr_GetNode(SCIAD_EXPR($), predec))
	coeffPred = SCIAD_VISITOR($).coeffList(predec)

	coeffNode = zeros(coeffPred)
	if (and(sz == [1,1])) then
		// Operand is scalar.
		if (coeffPred(1) <= 0) then
			error("TaylorDev : scalar operand of logarithm node " + index + " has null/negative 0-order term.")
		end  // if
		coeffNode(1) = log(coeffPred(1))
		for k = 1:SCIAD_VISITOR($).order do
			s = 0
			for i = 1:(k-1) do
				s = s + i .* coeffNode(i+1) .* coeffPred(k-i+1)
			end  // for
			coeffNode(k+1) = (coeffPred(k+1) - s ./ k) ./ coeffPred(1)
		end  // for
	elseif (sz(1) == 1) then
		// Operand is a row vector.
		if (or(coeffPred(1,:) <= 0)) then
			error("TaylorDev : row vector operand of logarithm node " + index + " has null/negative 0-order term.")
		end  // if
		coeffNode(1,:) = log(coeffPred(1,:))
		for k = 1:SCIAD_VISITOR($).order do
			s = zeros(1, sz(2))
			for i = 1:(k-1) do
				s = s + i .* coeffNode(i+1,:) .* coeffPred(k-i+1,:)
			end  // for
			coeffNode(k+1,:) = (coeffPred(k+1,:) - s ./ k) ./ coeffPred(1,:)
		end  // for
	elseif (sz(2) == 1) then
		// Operand is a column vector.
		if (or(coeffPred(:,1) <= 0)) then
			error("TaylorDev : column vector operand of logarithm node " + index + " has null/negative 0-order term.")
		end  // if
		coeffNode(:,1) = log(coeffPred(:,1))
		for k = 1:SCIAD_VISITOR($).order do
			s = zeros(sz(1), 1)
			for i = 1:(k-1) do
				s = s + i .* coeffNode(:,i+1) .* coeffPred(:,k-i+1)
			end  // for
			coeffNode(:,k+1) = (coeffPred(:,k+1) - s ./ k) ./ coeffPred(:,1)
		end  // for
	else
		// Operand is a matrix.
		if (or(coeffPred(:,:,1) <= 0)) then
			error("TaylorDev : matrix vector operand of logarithm node " + index + " has null/negative 0-order term.")
		end  // if
		coeffNode(:,:,1) = log(coeffPred(:,:,1))
		for k = 1:SCIAD_VISITOR($).order do
			s = zeros(sz(1), sz(2))
			for i = 1:(k-1) do
				s = s + i .* coeffNode(:,:,i+1) .* coeffPred(:,:,k-i+1)
			end  // for
			coeffNode(:,:,k+1) = (coeffPred(:,:,k+1) - s ./ k) ./ coeffPred(:,:,1)
		end  // for
	end  // if

	SCIAD_VISITOR($).coeffList(index) = coeffNode
endfunction


function [index]=TaylorDev_VisitMonome(node, index)
	// Monome node.
	// Only scalar monomes are supported.

	global SCIAD_EXPR SCIAD_VISITOR
	predec = SCIAD_VISITOR($).listPredec(index)
	sz = size(Expr_GetNode(SCIAD_EXPR($), predec))
	if (or(sz ~= [1,1])) then
		error("TaylorDev : monome node " + index + " has unsupported non-scalar operand.")
	end  // if
	coeffPred = SCIAD_VISITOR($).coeffList(predec)
	if (coeffPred(1) == 0) then
		error("TaylorDev : operand of monome node " + index + " has null 0-order term.")
	end  // if

	coeffNode = zeros(coeffPred)
	coeffNode(1) = coeffPred(1) ^ node.exponent
	for k = 1:SCIAD_VISITOR($).order do
		s = 0
		for i = 0:(k-1) do
			s = s + (node.exponent * (k - i) - i) * coeffNode(i+1) * coeffPred(k - i + 1)
		end  // for
		coeffNode(k+1) = s / (k * coeffPred(1))
	end  // for

	SCIAD_VISITOR($).coeffList(index) = coeffNode
endfunction


function [index]=TaylorDev_VisitElemMonome(node, index)
	// Element-wise monome node.

	global SCIAD_EXPR SCIAD_VISITOR
	predec = SCIAD_VISITOR($).listPredec(index)
	sz = size(Expr_GetNode(SCIAD_EXPR($), predec))
	coeffPred = SCIAD_VISITOR($).coeffList(predec)

	coeffNode = zeros(coeffPred)
	if (and(sz == [1,1])) then
		// Operand is scalar.
		if (coeffPred(1) == 0) then
			error("TaylorDev : scalar operand of elem-monome node " + index + " has null 0-order term.")
		end  // if
		coeffNode(1) = coeffPred(1) .^ node.exponent
		for k = 1:SCIAD_VISITOR($).order do
			s = 0
			for i = 0:(k-1) do
				s = s + (node.exponent .* (k - i) - i) .* coeffNode(i+1) .* coeffPred(k-i+1)
			end  // for
			coeffNode(k+1) = s ./ (k .* coeffPred(1))
		end  // for
	elseif (sz(1) == 1) then
		// Operand is a row vector.
		if (or(coeffPred(1,:) == 0)) then
			error("TaylorDev : row vector operand of elem-monome node " + index + " has null 0-order term.")
		end  // if
		coeffNode(1,:) = coeffPred(1,:) .^ node.exponent
		for k = 1:SCIAD_VISITOR($).order do
			s = zeros(1, sz(2))
			for i = 0:(k-1) do
				s = s + (node.exponent .* (k - i) - i) .* coeffNode(i+1,:) .* coeffPred(k-i+1,:)
			end  // for
			coeffNode(k+1,:) = s ./ (k .* coeffPred(1,:))
		end  // for
	elseif (sz(2) == 1) then
		// Operand is a column vector.
		if (or(coeffPred(:,1) == 0)) then
			error("TaylorDev : column vector operand of elem-monome node " + index + " has null 0-order term.")
		end  // if
		coeffNode(:,1) = coeffPred(:,1) .^ node.exponent
		for k = 1:SCIAD_VISITOR($).order do
			s = zeros(sz(1), 1)
			for i = 0:(k-1) do
				s = s + (node.exponent .* (k - i) - i) .* coeffNode(:,i+1) .* coeffPred(:,k-i+1)
			end  // for
			coeffNode(:,k+1) = s ./ (k .* coeffPred(:,1))
		end  // for
	else
		// Operand is a matrix.
		if (or(coeffPred(:,:,1) == 0)) then
			error("TaylorDev : column vector operand of elem-monome node " + index + " has null 0-order term.")
		end  // if
		coeffNode(:,:,1) = coeffPred(:,:,1) .^ node.exponent
		for k = 1:SCIAD_VISITOR($).order do
			s = zeros(sz(1), sz(2))
			for i = 0:(k-1) do
				s = s + (node.exponent .* (k - i) - i) .* coeffNode(:,:,i+1) .* coeffPred(:,:,k-i+1)
			end  // for
			coeffNode(:,:,k+1) = s ./ (k .* coeffPred(:,:,1))
		end  // for
	end  // if

	SCIAD_VISITOR($).coeffList(index) = coeffNode
endfunction


function [index]=TaylorDev_VisitSin(node, index)
	// (Element-wise) sine node.
	// Both sine and cosine coefficients must be computed for this to work.

	global SCIAD_EXPR SCIAD_VISITOR
	predec = SCIAD_VISITOR($).listPredec(index)
	coeffPred = SCIAD_VISITOR($).coeffList(predec)
	[coeffSin,coeffCos] = TaylorDev_ComputeCoeffSinCos(SCIAD_VISITOR($).order, Expr_GetNode(SCIAD_EXPR($), predec), coeffPred)
	SCIAD_VISITOR($).coeffList(index) = coeffSin
endfunction


function [index]=TaylorDev_VisitCos(node, index)
	// (Element-wise) cosine node.
	// Both sine and cosine coefficients must be computed for this to work.

	global SCIAD_EXPR SCIAD_VISITOR
	predec = SCIAD_VISITOR($).listPredec(index)
	coeffPred = SCIAD_VISITOR($).coeffList(predec)
	[coeffSin,coeffCos] = TaylorDev_ComputeCoeffSinCos(SCIAD_VISITOR($).order, Expr_GetNode(SCIAD_EXPR($), predec), coeffPred)
	SCIAD_VISITOR($).coeffList(index) = coeffCos
endfunction


function [coeffSin,coeffCos]=TaylorDev_ComputeCoeffSinCos(order, nodeArg, coeffArg)
	// Computes Taylor coefficients for sine and cosine of given argument node, knowing the Taylor coefficients of this argument.

	sz = size(nodeArg)
	coeffSin = zeros(coeffArg)
	coeffCos = zeros(coeffArg)
	if (and(sz == [1,1])) then
		// Argument is a scalar.
		coeffSin(1) = sin(coeffArg(1))
		coeffCos(1) = cos(coeffArg(1))
		for k = 1:order do
			ssin = 0
			scos = 0
			for i = 0:(k-1) do
				ssin = ssin + (k - i) .* coeffCos(i+1) .* coeffArg(k-i+1)
				scos = scos + (k - i) .* coeffSin(i+1) .* coeffArg(k-i+1)
			end  // for
			coeffSin(k+1) = ssin ./ k
			coeffCos(k+1) = -scos ./ k
		end  // for
	elseif (sz(1) == 1) then
		// Argument is a row vector.
		coeffSin(1,:) = sin(coeffArg(1,:))
		coeffCos(1,:) = cos(coeffArg(1,:))
		for k = 1:order do
			ssin = zeros(1, sz(2))
			scos = zeros(1, sz(2))
			for i = 0:(k-1) do
				ssin = ssin + (k - i) .* coeffCos(i+1,:) .* coeffArg(k-i+1,:)
				scos = scos + (k - i) .* coeffSin(i+1,:) .* coeffArg(k-i+1,:)
			end  // for
			coeffSin(k+1,:) = ssin ./ k
			coeffCos(k+1,:) = -scos ./ k
		end  // for
	elseif (sz(2) == 1) then
		// Argument is a column vector.
		coeffSin(:,1) = sin(coeffArg(:,1))
		coeffCos(:,1) = cos(coeffArg(:,1))
		for k = 1:order do
			ssin = zeros(sz(1), 1)
			scos = zeros(sz(1), 1)
			for i = 0:(k-1) do
				ssin = ssin + (k - i) .* coeffCos(:,i+1) .* coeffArg(:,k-i+1)
				scos = scos + (k - i) .* coeffSin(:,i+1) .* coeffArg(:,k-i+1)
			end  // for
			coeffSin(:,k+1) = ssin ./ k
			coeffCos(:,k+1) = -scos ./ k
		end  // for
	else
		// Argument is a matrix.
		coeffSin(:,:,1) = sin(coeffArg(:,:,1))
		coeffCos(:,:,1) = cos(coeffArg(:,:,1))
		for k = 1:order do
			ssin = zeros(sz(1), sz(2))
			scos = zeros(sz(1), sz(2))
			for i = 0:(k-1) do
				ssin = ssin + (k - i) .* coeffCos(:,:,i+1) .* coeffArg(:,:,k-i+1)
				scos = scos + (k - i) .* coeffSin(:,:,i+1) .* coeffArg(:,:,k-i+1)
			end  // for
			coeffSin(:,:,k+1) = ssin ./ k
			coeffCos(:,:,k+1) = -scos ./ k
		end  // for
	end  // if
endfunction


function [index]=TaylorDev_VisitAtan(node, index)
	// (Element-wise) arc-tangent node.

	global SCIAD_EXPR SCIAD_VISITOR
	predec = SCIAD_VISITOR($).listPredec(index)
	sz = size(Expr_GetNode(SCIAD_EXPR($), predec))
	coeffPred = SCIAD_VISITOR($).coeffList(predec)
	coeffNode = zeros(coeffPred)
	coeff_1pPred2 = zeros(coeffPred)

	if (and(sz == [1,1])) then
		// Argument is a scalar.
		coeffNode(1) = atan(coeffPred(1))
		coeff_1pPred2 = 1 + coeffPred(1) .^ 2
		for k = 1:SCIAD_VISITOR($).order do
			s = 0
			if (modulo(k, 2) == 0) then
				for i = 0:((k-2)/2) do
					s = s + coeffPred(i+1) .* coeffPred(k-i+1)
				end  // for
				coeff_1pPred2(k+1) = 2 .* s + coeffPred(k/2 + 1) .^ 2
			else
				for i = 0:((k-1)/2) do
					s = s + coeffPred(i+1) .* coeffPred(k-i+1)
				end  // for
				coeff_1pPred2(k+1) = 2 .* s
			end  // if

			s = 0
			for i = 1:(k-1) do
				s = s + i .* coeffNode(i+1) .* coeff_1pPred2(k-i+1)
			end  // for
			coeffNode(k+1) = (coeffPred(k+1) - s ./ k) ./ coeff_1pPred2(1)
		end  // for
	elseif (sz(1) == 1) then
		// Argument is a row vector.
		coeffNode(1,:) = atan(coeffPred(1,:))
		coeff_1pPred2 = 1 + coeffPred(1,:) .^ 2
		for k = 1:SCIAD_VISITOR($).order do
			s = zeros(1, sz(2))
			if (modulo(k, 2) == 0) then
				for i = 0:((k-2)/2) do
					s = s + coeffPred(i+1,:) .* coeffPred(k-i+1,:)
				end  // for
				coeff_1pPred2(k+1,:) = 2 .* s + coeffPred(k/2 + 1,:) .^ 2
			else
				for i = 0:((k-1)/2) do
					s = s + coeffPred(i+1,:) .* coeffPred(k-i+1,:)
				end  // for
				coeff_1pPred2(k+1,:) = 2 .* s
			end  // if

			s = 0
			for i = 1:(k-1) do
				s = s + i .* coeffNode(i+1,:) .* coeff_1pPred2(k-i+1,:)
			end  // for
			coeffNode(k+1,:) = (coeffPred(k+1,:) - s ./ k) ./ coeff_1pPred2(1,:)
		end  // for
	elseif (sz(2) == 1) then
		// Argument is a column vector.
		coeffNode(:,1) = atan(coeffPred(:,1))
		coeff_1pPred2 = 1 + coeffPred(:,1) .^ 2
		for k = 1:SCIAD_VISITOR($).order do
			s = zeros(sz(1), 1)
			if (modulo(k, 2) == 0) then
				for i = 0:((k-2)/2) do
					s = s + coeffPred(:,i+1) .* coeffPred(:,k-i+1)
				end  // for
				coeff_1pPred2(:,k+1) = 2 .* s + coeffPred(:,k/2 + 1) .^ 2
			else
				for i = 0:((k-1)/2) do
					s = s + coeffPred(:,i+1) .* coeffPred(:,k-i+1)
				end  // for
				coeff_1pPred2(:,k+1) = 2 .* s
			end  // if

			s = 0
			for i = 1:(k-1) do
				s = s + i .* coeffNode(:,i+1) .* coeff_1pPred2(:,k-i+1)
			end  // for
			coeffNode(:,k+1) = (coeffPred(:,k+1) - s ./ k) ./ coeff_1pPred2(:,1)
		end  // for
	else
		// Argument is a matrix.
		coeffNode(:,:,1) = atan(coeffPred(:,:,1))
		coeff_1pPred2 = 1 + coeffPred(:,:,1) .^ 2
		for k = 1:SCIAD_VISITOR($).order do
			s = zeros(sz(1), sz(2))
			if (modulo(k, 2) == 0) then
				for i = 0:((k-2)/2) do
					s = s + coeffPred(:,:,i+1) .* coeffPred(:,:,k-i+1)
				end  // for
				coeff_1pPred2(:,:,k+1) = 2 .* s + coeffPred(:,:,k/2 + 1) .^ 2
			else
				for i = 0:((k-1)/2) do
					s = s + coeffPred(:,:,i+1) .* coeffPred(:,:,k-i+1)
				end  // for
				coeff_1pPred2(:,:,k+1) = 2 .* s
			end  // if

			s = 0
			for i = 1:(k-1) do
				s = s + i .* coeffNode(:,:,i+1) .* coeff_1pPred2(:,:,k-i+1)
			end  // for
			coeffNode(:,:,k+1) = (coeffPred(:,:,k+1) - s ./ k) ./ coeff_1pPred2(:,:,1)
		end  // for
	end  // if

	SCIAD_VISITOR($).coeffList(index) = coeffNode
endfunction


function [index]=TaylorDev_VisitAbs(node, index)
	// Unsupported as of yet.
	global SCIAD_EXPR SCIAD_VISITOR
	error("TaylorDev : absolute value node not supported yet.")
endfunction


function [index]=TaylorDev_VisitStub(node, index)
	error("TaylorDev : not supposed to find a stub here!")
endfunction


// End of module.

